/*
 * Copyright (c) 1994, 1996 Hellmuth Michaelis. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *---------------------------------------------------------------------------*
 *
 *	bisdnd - BSD ISDN management daemon
 *      ===================================
 *
 *	Last Edit-Date: [Fri Jul  5 12:09:07 1996]
 *
 *	-hm	made error checking more robust
 *	-hm	check how much controllers configured
 *	-hm	diffs from Gary for new rates implementation
 *	-hm	fix whitespace skips for on/off
 *	-hm	cleanup error handling
 *	-hm	new telephone number handling
 *	-hm	idle time cleanup
 *	-hm	setting the controllers isdn protocol type
 *	-hm	telno wildcards
 *	-hm	timeout/idletime fixes
 *	-hm	adding unitlength and earlyhangup keyword
 *	-hm	adding support for NetBSD from Arne
 *	-hm	calledbackwait variable
 *	-hm	bugfix patch from Greg Lehey, no more than 2 entries possible
 *	-hm	bugfix, if no facility configured don't reconfigure it to 0!
 *	-hm	bugfix, dumps core if logging to file configured
 *	-hm	regular expression log message matching
 *	-hm	removing loglevel configuration
 *
 *----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#ifdef __NetBSD__
#include <sys/ioctl.h>
#endif /* __NetBSD__ */

#include <sys/param.h>
#include <sys/time.h>

#include "defs.h"
#include "config.h"

static int optbool(char *string, int *var, int entry, int line);
static int optint(char *string, unsigned int *var, int entry, int line);
static int optstr(char *string, char *dest, int entry, int line);
static int optnumstr(char *string, char *dest, int entry, int line, int wildallow);
static char *optcstr(char *string, int entry, int line);
static int check_oneconfig(int config_entries, unsigned int flags);
static int check_system(unsigned int flags);
static int check_config(void);
static int handle_eflag(unsigned int *flags, unsigned int flag, char *string, int entry, int line);
static int handle_sflag(unsigned int *flags, unsigned int flag, char *string, int line);
static int init_subsys(int config_entries);

static char error[256];		/* error string */

#ifdef CONF_DEBUG_MAIN
int main()
{
	int ret;
	ret = readconf(CONFIG_FILE_DEF);
	if(ret > 0)
		fprintf(stderr, "readconf returns [%d] = [%s]\n", ret, sys_errlist[ret]);
	else
		fprintf(stderr, "readconf returns [%d]\n", ret);
	return(ret);
}
#endif

/*---------------------------------------------------------------------------*
 *	get configuration
 *---------------------------------------------------------------------------*/
int
readconf(char *filename)
{
	char buffer[MAXPATHLEN + 1];		/* gp buffer */
	FILE *fp;
	unsigned int flags = 0;
	unsigned int sflags = 0;	
	int line = 0;
	int mode = -1;
	int got_system = 0;
	
	error[0] = '\0';
	config_error = NULL;			/* init error ptr */

	numcards = 0;
	
	config_entries = 0;

	if((fp = fopen(filename, "r")) == NULL)
	{
		sprintf(error, "error open %s: %s", filename, sys_errlist[errno]);
		config_error = error;
		return(ERROR);
	}

	while((fgets(buffer, MAXPATHLEN, fp)) != NULL)
	{
		line++;

/* comments */
		if(buffer[0] == '#'  || buffer[0] == ' ' ||
		   buffer[0] == '\t' || buffer[0] == '\n')
		{
			continue;
		}

/* system configuration */

		if(!strncmp(buffer, CF_SYSTEM, strlen(CF_SYSTEM)))
		{
			if(got_system == 0)
			{
				mode = MODE_SYSTEM;
				sflags = FL_SYSTEM;
				got_system = 1;
			}
			else
			{				
				sprintf(error, "more than one 'system' keyword in line %d!", line);
				goto config_error;
			}
		}

/* entry separator */

		else if(!strncmp(buffer, CF_ENTRY, strlen(CF_ENTRY)))
		{
			mode = MODE_ENTRY;

			if(flags & FL_ENTRY)
			{
				if((check_oneconfig(config_entries, flags)) == ERROR)
					goto config_error;
				config_entries++;
				flags = FL_ENTRY;
			}
			else
			{
				flags |= FL_ENTRY;
			}
		}

/* strings */

		else if((!strncmp(buffer, CF_NAME, strlen(CF_NAME))) && (mode == MODE_ENTRY))
		{
			if((handle_eflag(&flags, FL_NAME, CF_NAME, config_entries, line)) == ERROR)
				goto config_error;
			if((optstr(buffer, &conf[config_entries].name[0], config_entries, line)) == ERROR)
				goto config_error;
		}
		else if((!strncmp(buffer, CF_INTERFACE, strlen(CF_INTERFACE))) && (mode == MODE_ENTRY))
		{
			if((handle_eflag(&flags, FL_INTERFACE, CF_INTERFACE, config_entries, line)) == ERROR)
				goto config_error;
			if((optstr(buffer, &conf[config_entries].interface[0], config_entries, line)) == ERROR)
				goto config_error;

			if(!strncmp(&conf[config_entries].interface[0], CVAL_IPI, strlen(CVAL_IPI)))
			{
				conf[config_entries].atyp = ATYP_IP;
				conf[config_entries].appl = (int)strtol(&conf[config_entries].interface[strlen(CVAL_IPI)], (char **)NULL, 10);
				conf[config_entries].service = SERV_DATA;
			}
#ifdef NOTDEF
			else if(!strncmp(&conf[config_entries].interface[0], CVAL_TTY, strlen(CVAL_TTY)))
			{
				conf[config_entries].atyp = ATYP_TTY;
				conf[config_entries].service = SERV_DATA;
			}
#endif
			else if(!strncmp(&conf[config_entries].interface[0], CVAL_ANSW, strlen(CVAL_ANSW)))
			{
				conf[config_entries].atyp = ATYP_TEL;
				conf[config_entries].service = SERV_PHONE;
				conf[config_entries].telaction = ACT_ANSWER;
			}
			else if(!strncmp(&conf[config_entries].interface[0], CVAL_PLAY, strlen(CVAL_PLAY)))
			{
				conf[config_entries].atyp = ATYP_TEL;
				conf[config_entries].service = SERV_PHONE;
				conf[config_entries].telaction = ACT_PLAYBK;
			}
			else
			{
				sprintf(error, "unknown parameter '%s' for keyword '%s' in entry number %d line %d!",
					&conf[config_entries].interface[0],
					CF_INTERFACE, config_entries, line);
				goto config_error;
			}

		}

		else if((!strncmp(buffer, CF_RATESFILE, strlen(CF_RATESFILE))) && (mode == MODE_SYSTEM))
		{
			if((handle_sflag(&sflags, FL_RATESFILE, CF_RATESFILE, line)) == ERROR)
				goto config_error;
			if((optstr(buffer, &ratesfile[0], 0, line)) == ERROR)
				goto config_error;
		}
		else if((!strncmp(buffer, CF_ACCTFILE, strlen(CF_ACCTFILE))) && (mode == MODE_SYSTEM))
		{
			if((handle_sflag(&sflags, FL_ACCTFILE, CF_ACCTFILE, line)) == ERROR)
				goto config_error;
			if((optstr(buffer, &acctfile[0], 0, line)) == ERROR)
				goto config_error;
		}
		else if((!strncmp(buffer, CF_ANSWPROG, strlen(CF_ANSWPROG))) && (mode == MODE_SYSTEM))
		{
			if((handle_sflag(&sflags, FL_ANSWPROG, CF_ANSWPROG, line)) == ERROR)
				goto config_error;
			if((optstr(buffer, &answerprog[0], 0, line)) == ERROR)
				goto config_error;
		}
		else if((!strncmp(buffer, CF_PLAYPROG, strlen(CF_PLAYPROG))) && (mode == MODE_SYSTEM))
		{
			if((handle_sflag(&sflags, FL_PLAYPROG, CF_PLAYPROG, line)) == ERROR)
				goto config_error;
			if((optstr(buffer, &playprog[0], 0, line)) == ERROR)
				goto config_error;
		}
		else if((!strncmp(buffer, CF_REGEXPR, strlen(CF_REGEXPR))) && (mode == MODE_SYSTEM))
		{
			static int regecount = 0;
			char ab[16];
			
			sprintf(ab, "%s%d", CF_REGEXPR, regecount);

			if(regecount >= MAX_RE)
			{
				sprintf(error, "error for %s in line %d: allowed max is %d!", ab, line, MAX_RE);
				goto config_error;
			}
			if(regecount == 0)
			{
				handle_sflag(&sflags, FL_REGEXPR, CF_REGEXPR, line);
			}

			if(!strncmp(buffer, ab, strlen(ab)))
			{
				char pattern[256];	/* XXX */
				int i;

				if((optstr(buffer, &pattern[0], 0, line)) == ERROR)
					goto config_error;

				if((i = regcomp(&(rarr[regecount].re), pattern,
					REG_EXTENDED|REG_NOSUB)) != 0)
				{
					char buf[256];
					regerror(i, &(rarr[regecount].re), buf, sizeof(buf));
					sprintf(error, "regcomp error for %s in line %d: [%s]", ab, line, buf);
					goto config_error;
				}
				else
				{
					char *p;
					if((p = malloc(strlen(pattern) + 1)) == NULL)
					{
						sprintf(error, "malloc error for %s in line %d", ab, line);
						goto config_error;
					}
					strcpy(p, pattern);
					rarr[regecount].re_expr = p;
					rarr[regecount].re_flg = 1;
				}
			}
			else
			{
				sprintf(error, "invalid, duplicate or not consecutive 'regexpr' in line %d", line);
				goto config_error;
			}
			regecount++;
		}

		else if((!strncmp(buffer, CF_REGPROG, strlen(CF_REGPROG))) && (mode == MODE_SYSTEM))
		{
			static int regpcount = 0;
			char ab[16];
			
			sprintf(ab, "%s%d", CF_REGPROG, regpcount);
			
			if(regpcount >= MAX_RE)
			{
				sprintf(error, "error for %s in line %d: allowed max is %d!", ab, line, MAX_RE);
				goto config_error;
			}
			
			if(regpcount == 0)
			{
				handle_sflag(&sflags, FL_REGPROG, CF_REGPROG, line);
			}

			if(!strncmp(buffer, ab, strlen(ab)))
			{
				char pattern[256];	/* XXX */
				char *p;

				if((optstr(buffer, &pattern[0], 0, line)) == ERROR)
					goto config_error;

				if((p = malloc(strlen(pattern) + 1)) == NULL)
				{
					sprintf(error, "malloc error for %s in line %d", ab, line);
					goto config_error;
				}
				strcpy(p, pattern);
				rarr[regpcount].re_prog = p;
			}
			else
			{
				sprintf(error, "invalid, duplicate or not consecutive 'regprog' in line %d", line);
				goto config_error;
			}
			regpcount++;			
		}
		
/* string choice variables */

		else if((!strncmp(buffer, CF_ISDNTYPE, strlen(CF_ISDNTYPE))) && (mode == MODE_ENTRY))
		{
			char *p;
			if((handle_eflag(&flags, FL_ISDNTYPE, CF_ISDNTYPE, config_entries, line)) == ERROR)
			{
				goto config_error;
			}
			if((p = optcstr(buffer, config_entries, line)) == NULL)
			{
				sprintf(error, "syntax error for keyword '%s' in entry number %d line %d,\nentry=%s",
					CF_ISDNTYPE, config_entries, line, buffer);
				goto config_error;
			}
			if(!strncmp(p, CVAL_DSS1, strlen(CVAL_DSS1)))
			{
				conf[config_entries].isdntype = ISDNDSS1;
			}
			else if(!strncmp(p, CVAL_1TR6, strlen(CVAL_1TR6)))
			{
				conf[config_entries].isdntype = ISDN1TR6;
			}
			else
			{
				sprintf(error, "unknown parameter '%s' for keyword '%s' in entry number %d line %d!",
					p, CF_ISDNTYPE, config_entries, line);
				goto config_error;
			}
		}
		else if((!strncmp(buffer, CF_REACTION, strlen(CF_REACTION))) && (mode == MODE_ENTRY))
		{
			char *p;
			if((handle_eflag(&flags, FL_REACTION, CF_REACTION, config_entries, line)) == ERROR)
			{
				goto config_error;
			}
			if((p = optcstr(buffer, config_entries, line)) == NULL)
			{
				sprintf(error, "syntax error for keyword '%s' in entry number %d line %d,\nentry=%s",
					CF_REACTION, config_entries, line, buffer);
				goto config_error;
			}
			if(!strncmp(p, CVAL_ACCEPT, strlen(CVAL_ACCEPT)))
			{
				conf[config_entries].reaction = REACT_ACCEPT;
			}
			else if(!strncmp(p, CVAL_REJECT, strlen(CVAL_REJECT)))
			{
				conf[config_entries].reaction = REACT_REJECT;
			}
			else if(!strncmp(p, CVAL_CALLBACK, strlen(CVAL_CALLBACK)))
			{
				conf[config_entries].reaction = REACT_CALLBACK;
			}
			else if(!strncmp(p, CVAL_CALLEDBACK, strlen(CVAL_CALLEDBACK)))
			{
				conf[config_entries].reaction = REACT_CALLEDBACK;
			}
			else if(!strncmp(p, CVAL_ALERTBASE, strlen(CVAL_ALERTBASE)))
			{
				conf[config_entries].reaction = REACT_ALERTBASE;
			}
			else
			{
				sprintf(error, "unknown parameter '%s' for keyword '%s' in entry %d line %d!",
					p, CF_REACTION, config_entries, line);
				goto config_error;
			}
		}
		else if((!strncmp(buffer, CF_PROTOCOL, strlen(CF_PROTOCOL))) && (mode == MODE_ENTRY))
		{
			char *p;
			if((handle_eflag(&flags, FL_PROTOCOL, CF_PROTOCOL, config_entries, line)) == ERROR)
			{
				goto config_error;
			}
			if((p = optcstr(buffer, config_entries, line)) == NULL)
			{
				sprintf(error, "syntax error for keyword '%s' in entry number %d line %d,\nentry=%s",
					CF_PROTOCOL, config_entries, line, buffer);
				goto config_error;
			}
			if(!strncmp(p, CVAL_RAW, strlen(CVAL_RAW)))
			{
				conf[config_entries].protocol = PROT_RAW;
				conf[config_entries].service &= 0x0f;
			}
			else if(!strncmp(p, CVAL_UI, strlen(CVAL_UI)))
			{
				conf[config_entries].protocol = PROT_UI;
				conf[config_entries].service |= 0x10;
			}
			else if(!strncmp(p, CVAL_X75, strlen(CVAL_X75)))
			{
				conf[config_entries].protocol = PROT_X75;
				conf[config_entries].service |= 0x20;
			}
			else if(!strncmp(p, CVAL_T70, strlen(CVAL_T70)))
			{
				conf[config_entries].protocol = PROT_T70;
				conf[config_entries].service |= 0x120;
			}
			else if(!strncmp(p, CVAL_T90, strlen(CVAL_T90)))
			{
				conf[config_entries].protocol = PROT_T90;
				conf[config_entries].service |= 0x320;
			}
			else if(!strncmp(p, CVAL_ISO, strlen(CVAL_ISO)))
			{
				conf[config_entries].protocol = PROT_ISO;
				conf[config_entries].service |= 0x220;
			}
			else
			{
				sprintf(error, "unknown parameter '%s' for keyword '%s' in entry %d line %d!",
					p, CF_PROTOCOL, config_entries, line);
				goto config_error;
			}
		}
		
/* numeric strings */

		else if((!strncmp(buffer, CF_TELNLOC_LDO, strlen(CF_TELNLOC_LDO))) && (mode == MODE_ENTRY))
		{
			if((handle_eflag(&flags, FL_TELNLOC_LDO, CF_TELNLOC_LDO, config_entries, line)) == ERROR)
				goto config_error;
			if((optnumstr(buffer, conf[config_entries].telnloc_ldo, config_entries, line, NO_WILD)) == ERROR)
				goto config_error;
		}
		else if((!strncmp(buffer, CF_TELNREM_LDO, strlen(CF_TELNREM_LDO))) && (mode == MODE_ENTRY))
		{
			if((handle_eflag(&flags, FL_TELNREM_LDO, CF_TELNREM_LDO, config_entries, line)) == ERROR)
				goto config_error;
			if((optnumstr(buffer, conf[config_entries].telnrem_ldo, config_entries, line, NO_WILD)) == ERROR)
				goto config_error;
		}
		else if((!strncmp(buffer, CF_TELNLOC_RDI, strlen(CF_TELNLOC_RDI))) && (mode == MODE_ENTRY))
		{
			if((handle_eflag(&flags, FL_TELNLOC_RDI, CF_TELNLOC_RDI, config_entries, line)) == ERROR)
				goto config_error;
			if((optnumstr(buffer, conf[config_entries].telnloc_rdi, config_entries, line, NO_WILD)) == ERROR)
				goto config_error;
		}
		else if((!strncmp(buffer, CF_TELNREM_RDI, strlen(CF_TELNREM_RDI))) && (mode == MODE_ENTRY))
		{
			if((handle_eflag(&flags, FL_TELNREM_RDI, CF_TELNREM_RDI, config_entries, line)) == ERROR)
				goto config_error;
			if((optnumstr(buffer, conf[config_entries].telnrem_rdi, config_entries, line, ALLOW_WILD)) == ERROR)
				goto config_error;
		}

/* booleans */
		else if((!strncmp(buffer, CF_UNITLENDYN, strlen(CF_UNITLENDYN))) && (mode == MODE_ENTRY))
		{
			if((handle_eflag(&flags, FL_UNITLENDYN, CF_UNITLENDYN, config_entries, line)) == ERROR)
				goto config_error;
			if((optbool(buffer, &conf[config_entries].unitlength_dyn, config_entries, line)) == ERROR)
				goto config_error;
		}

		else if((!strncmp(buffer, CF_USEACCTFILE, strlen(CF_USEACCTFILE))) && (mode == MODE_SYSTEM))
		{
			if((handle_sflag(&sflags, FL_USEACCTFILE, CF_USEACCTFILE, line)) == ERROR)
				goto config_error;
			if((optbool(buffer, &useacctfile, 0, line)) == ERROR)
				goto config_error;
		}

		else if((!strncmp(buffer, CF_ACCTALL, strlen(CF_ACCTALL))) && (mode == MODE_SYSTEM))
		{
			if((handle_sflag(&sflags, FL_ACCTALL, CF_ACCTALL, line)) == ERROR)
				goto config_error;
			if((optbool(buffer, &acct_all, 0, line)) == ERROR)
				goto config_error;
		}
		
/* integers */
		else if((!strncmp(buffer, CF_CONTROLLER, strlen(CF_CONTROLLER))) && (mode == MODE_ENTRY))
		{
			if((handle_eflag(&flags, FL_CONTROLLER, CF_CONTROLLER, config_entries, line)) == ERROR)
				goto config_error;
			if((optint(buffer, &conf[config_entries].controller, config_entries, line)) == ERROR)
				goto config_error;
			/* remember last controller */
			if(conf[config_entries].controller > numcards)
				numcards = conf[config_entries].controller;
		}
		else if((!strncmp(buffer, CF_DIALRETRIES, strlen(CF_DIALRETRIES))) && (mode == MODE_ENTRY))
		{
			if((handle_eflag(&flags, FL_DIALRETRIES, CF_DIALRETRIES, config_entries, line)) == ERROR)
				goto config_error;
			if((optint(buffer, &conf[config_entries].dialretries, config_entries, line)) == ERROR)
				goto config_error;
		}
		else if((!strncmp(buffer, CF_RECOVERYTIME, strlen(CF_RECOVERYTIME))) && (mode == MODE_ENTRY))
		{
			if((handle_eflag(&flags, FL_RECOVERYTIME, CF_RECOVERYTIME, config_entries, line)) == ERROR)
				goto config_error;
			if((optint(buffer, &conf[config_entries].recoverytime, config_entries, line)) == ERROR)
				goto config_error;
		}
		else if((!strncmp(buffer, CF_CALLBACKWAIT, strlen(CF_CALLBACKWAIT))) && (mode == MODE_ENTRY))
		{
			if((handle_eflag(&flags, FL_CALLBACKWAIT, CF_CALLBACKWAIT, config_entries, line)) == ERROR)
				goto config_error;
			if((optint(buffer, &conf[config_entries].callbackwait, config_entries, line)) == ERROR)
				goto config_error;
		}
		else if((!strncmp(buffer, CF_IDLETIMEIN, strlen(CF_IDLETIMEIN))) && (mode == MODE_ENTRY))
		{
			if((handle_eflag(&flags, FL_IDLETIMEIN, CF_IDLETIMEIN, config_entries, line)) == ERROR)
				goto config_error;
			if((optint(buffer, &conf[config_entries].idletime_in, config_entries, line)) == ERROR)
				goto config_error;
		}
		else if((!strncmp(buffer, CF_IDLETIMEOUT, strlen(CF_IDLETIMEOUT))) && (mode == MODE_ENTRY))
		{
			if((handle_eflag(&flags, FL_IDLETIMEOUT, CF_IDLETIMEOUT, config_entries, line)) == ERROR)
				goto config_error;
			if((optint(buffer, &conf[config_entries].idletime_out, config_entries, line)) == ERROR)
				goto config_error;
		}
		else if((!strncmp(buffer, CF_BLOCKSIZE, strlen(CF_BLOCKSIZE))) && (mode == MODE_ENTRY))
		{
			if((handle_eflag(&flags, FL_BLOCKSIZE, CF_BLOCKSIZE, config_entries, line)) == ERROR)
				goto config_error;
			if((optint(buffer, &conf[config_entries].blocksize, config_entries, line)) == ERROR)
				goto config_error;
		}
		else if((!strncmp(buffer, CF_RATETYPE, strlen(CF_RATETYPE))) && (mode == MODE_ENTRY))
		{
			if((handle_eflag(&flags, FL_RATETYPE, CF_RATETYPE, config_entries, line)) == ERROR)
				goto config_error;
			if((optint(buffer, &conf[config_entries].ratetyp, config_entries, line)) == ERROR)
				goto config_error;
		}
		else if((!strncmp(buffer, CF_UNITLENGTH, strlen(CF_UNITLENGTH))) && (mode == MODE_ENTRY))
		{
			if((handle_eflag(&flags, FL_UNITLENGTH, CF_UNITLENGTH, config_entries, line)) == ERROR)
				goto config_error;
			if((optint(buffer, &conf[config_entries].unitlength, config_entries, line)) == ERROR)
				goto config_error;
		}
		else if((!strncmp(buffer, CF_EARLYHANGUP, strlen(CF_EARLYHANGUP))) && (mode == MODE_ENTRY))
		{
			if((handle_eflag(&flags, FL_EARLYHANGUP, CF_EARLYHANGUP, config_entries, line)) == ERROR)
				goto config_error;
			if((optint(buffer, &conf[config_entries].earlyhangup, config_entries, line)) == ERROR)
				goto config_error;
		}
		else if((!strncmp(buffer, CF_CALLEDBACKWAIT, strlen(CF_CALLEDBACKWAIT))) && (mode == MODE_ENTRY))
		{
			if((handle_eflag(&flags, FL_CALLEDBACKWAIT, CF_CALLEDBACKWAIT, config_entries, line)) == ERROR)
				goto config_error;
			if((optint(buffer, &conf[config_entries].calledbackwait, config_entries, line)) == ERROR)
				goto config_error;
		}

/* unknown */
		else
		{
			char *ptr;
			
			if((ptr = (char *) index(&buffer[0], '=')) != NULL)
			{
				*ptr = '\0';
				sprintf(error, "unknown keyword [%s] in entry number %d line %d",
					buffer, config_entries, line);
			}
			else if((ptr = (char *) index(&buffer[0], '\n')) != NULL)
			{
				*ptr = '\0';
				sprintf(error, "unknown keyword in entry number %d line %d,\n  entry=[%s]",
					config_entries, line, buffer);
			}
			else
			{
				sprintf(error, "unknown keyword in entry number %d line %d",
					config_entries, line);
			}				
			goto config_error;
		}
	}
  
	if((check_oneconfig(config_entries, flags)) == ERROR)
		goto config_error;

	if((check_system(sflags)) == ERROR)
		goto config_error;

	fclose(fp);		
	return(check_config());

config_error:
	fclose(fp);
	config_error = error;
	return(ERROR);
}

/*---------------------------------------------------------------------------*
 *	handle checking and setting of entry config flags
 *---------------------------------------------------------------------------*/
static int
handle_eflag(unsigned int *flags, unsigned int flag, char *string, int entry, int line)
{
	if(*flags & flag)
	{
		sprintf(error, "duplicate entry for keyword '%s' in entry number %d line %d",
			string, entry, line);
		return(ERROR);
	}
	*flags |= flag;
	return(GOOD);
}

/*---------------------------------------------------------------------------*
 *	handle checking and setting of system config flags
 *---------------------------------------------------------------------------*/
static int
handle_sflag(unsigned int *flags, unsigned int flag, char *string, int line)
{
	if(*flags & flag)
	{
		sprintf(error, "duplicate entry for keyword '%s' in system section, line %d",
			string, line);
		return(ERROR);
	}
	*flags |= flag;
	return(GOOD);
}

/*---------------------------------------------------------------------------*
 *	process on/off strings
 *---------------------------------------------------------------------------*/
static int
optbool(char *string, int *var, int entry, int line)
{
	char *ptr;

	if ((ptr = (char *) index(string, '=')) == NULL)
	{
		sprintf(error, "no '=' after keyword in entry number %d line %d",
			entry, line);
		return(ERROR);
	}

	ptr++;

	while(isspace(*ptr))
		ptr++;

	if (!strncmp(ptr, CVAL_ON, strlen(CVAL_ON)))
	{
		*var = 1;
	}
	else if (!strncmp(ptr, CVAL_OFF, strlen(CVAL_OFF)))
	{
		*var = 0;
	}
	else
	{
		sprintf(error, "value not 'on' or 'off' in entry number %d line %d",
			entry, line);
		return(ERROR);
	}
	return(GOOD);
}

/*---------------------------------------------------------------------------*
 *	process integer values
 *---------------------------------------------------------------------------*/
static int
optint(char *string, unsigned *var, int entry, int line)
{
	unsigned char *ptr;

	if ((ptr = (unsigned char *) index(string, '=')) == NULL)
	{
		sprintf(error, "no '=' after keyword in entry number %d line %d",
			entry, line);
		return(ERROR);
	}

	ptr++;

	while(isspace(*ptr))
		ptr++;

	errno = 0;

	*var = (int)strtol(ptr, (char **)NULL, 10);

	if(errno == ERANGE)
	{
		sprintf(error, "value under- or overflow in entry number %d line %d",
			entry, line);
		return(ERROR);
	}
	return(GOOD);
}

/*---------------------------------------------------------------------------*
 *	process strings
 *---------------------------------------------------------------------------*/
static int
optstr(char *string, char dest[], int entry, int line)
{
	char *ptr, *bptr;

	if ((ptr = (char *) index(string, '=')) == NULL)
	{
		sprintf(error, "no '=' after keyword in entry number %d line %d",
			entry, line);
		return(ERROR);
	}

	ptr++;

	while(isspace(*ptr))
		ptr++;

	bptr = dest;

	while (*ptr && (*ptr != '#') && (isprint(*ptr)) && (!isspace(*ptr)))
		*bptr++ = *ptr++;

	*bptr = '\0';
	return(GOOD);
}

/*---------------------------------------------------------------------------*
 *	process numeric strings
 *---------------------------------------------------------------------------*/
static int
optnumstr(char *string, char dest[], int entry, int line, int wildallow)
{
	char *ptr, *bptr;

	if ((ptr = (char *) index(string, '=')) == NULL)
	{
		sprintf(error, "no '=' after keyword in entry number %d line %d",
			entry, line);
		return(ERROR);
	}

	ptr++;

	while(isspace(*ptr))	/* eat spaces */
		ptr++;

	bptr = dest;

	if((wildallow == ALLOW_WILD) && (*ptr == '*'))	/* wildcard ? */
	{
		*bptr++ = *ptr++;
	}
	else			/* digits */
	{
		while (*ptr && (*ptr != '#') && (isdigit(*ptr)) && (!isspace(*ptr)))
			*bptr++ = *ptr++;
	}

	*bptr = '\0';
	return(GOOD);
}

/*---------------------------------------------------------------------------*
 *	partly process choice strings
 *---------------------------------------------------------------------------*/
static char *
optcstr(char *string, int entry, int line)
{
	static char cstr[80];	
	char *ptr, *bptr;

	if ((ptr = (char *) index(string, '=')) == NULL)
	{
		sprintf(error, "no '=' after keyword in entry number %d line %d",
			entry, line);
		return(NULL);
	}

	ptr++;

	while(isspace(*ptr))
		ptr++;

	bptr = &cstr[0];

	while (*ptr && (*ptr != '#') && (isprint(*ptr)) && (!isspace(*ptr)))
		*bptr++ = *ptr++;

	*bptr = '\0';

	return(&cstr[0]);
}

/*---------------------------------------------------------------------------*
 *	validate one configuration entry
 *---------------------------------------------------------------------------*/
static int
check_oneconfig(int config_entries, unsigned int flags)
{
	int f;
	int buf = 0;
	int newbuf = 0;	
	int ret = GOOD;
	char device[64];	

	if(flags & FL_NAME)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: name = %s", config_entries, &conf[config_entries].name[0]);
#endif		
	}
	else
	{
		sprintf(error, "ERROR, entry %d: name missing", config_entries);
		ret = ERROR;
	}
	
	if(flags & FL_INTERFACE)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: interface = %s", config_entries, &conf[config_entries].interface[0]);
#endif		
	}
	else
	{
		sprintf(error, "ERROR, entry %d: interface missing", config_entries);
		ret = ERROR;
	}
	
	if(flags & FL_ISDNTYPE)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: isdntype = %d", config_entries, conf[config_entries].isdntype);
#endif		
	}
	else
	{
		sprintf(error, "ERROR, entry %d: isdntype missing", config_entries);
		ret = ERROR;
	}
	
	if(flags & FL_REACTION)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: reaction = %d", config_entries, conf[config_entries].reaction);
#endif		
	}
	else
	{
		sprintf(error, "ERROR, entry %d: reaction missing", config_entries);
		ret = ERROR;
	}
	
	if(flags & FL_PROTOCOL)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: protocol = %d", config_entries, conf[config_entries].protocol);
#endif		
	}
	else
	{
		sprintf(error, "ERROR, entry %d: protocol missing", config_entries);
		ret = ERROR;
	}
	
	if(flags & FL_TELNLOC_LDO)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: telnloc-ldo = %s", config_entries, conf[config_entries].telnloc_ldo);
#endif		
	}
	else
	{
		if(conf[config_entries].service != SERV_PHONE)
		{
			sprintf(error, "ERROR, entry %d: telnloc-ldo missing", config_entries);
			ret = ERROR;
		}
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "WARNING, no telnloc-ldo in entry %d (service=phone)", config_entries);
#endif		
	}
	
	if(flags & FL_TELNREM_LDO)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: telnrem-ldo = %s", config_entries, conf[config_entries].telnrem_ldo);
#endif		
	}
	else
	{
		if(conf[config_entries].service != SERV_PHONE)
		{
			sprintf(error, "ERROR, entry %d: telnrem-ldo missing", config_entries);
			ret = ERROR;
		}
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "WARNING, no telnrem-ldo in entry %d (service=phone)", config_entries);
#endif		
	}
	
	if(flags & FL_TELNLOC_RDI)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: telnloc-rdi = %s", config_entries, conf[config_entries].telnloc_rdi);
#endif		
	}
	else
	{
		sprintf(error, "ERROR, entry %d: telnloc-rdi missing", config_entries);
		ret = ERROR;
	}
	
	if(flags & FL_TELNREM_RDI)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: telnrem-rdi = %s", config_entries, conf[config_entries].telnrem_rdi);
#endif		
	}
	else
	{
		sprintf(error, "ERROR, entry %d: telnrem-rdi missing", config_entries);
		ret = ERROR;
	}
	
	if(flags & FL_CONTROLLER)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: controller = %d", config_entries, conf[config_entries].controller);
#endif		
	}
	else
	{
		sprintf(error, "ERROR, entry %d: controller missing", config_entries);
		ret = ERROR;
	}
	
	if(flags & FL_DIALRETRIES)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: dialretries = %d", config_entries, conf[config_entries].dialretries);
#endif		
	}
	else
	{
		conf[config_entries].dialretries = DIAL_DEFAULT;
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "WARNING, no dialretries in entry %d, using %d", config_entries, conf[config_entries].dialretries);
#endif		
	}
	
	if(flags & FL_RECOVERYTIME)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: recoverytime = %d", config_entries, conf[config_entries].recoverytime);
#endif		
	}
	else
	{
		conf[config_entries].recoverytime = RECV_DEFAULT;
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "WARNING, no recoverytime in entry %d, using %d", config_entries, conf[config_entries].recoverytime);
#endif		
	}
	
	if(flags & FL_CALLBACKWAIT)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: callbackwait time = %d", config_entries, conf[config_entries].callbackwait);
#endif		
	}
	else
	{
		conf[config_entries].callbackwait = CBW_DEFAULT;
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "WARNING, no callbackwait time in entry %d, using %d", config_entries, conf[config_entries].callbackwait);
#endif		
	}
	
	if(flags & FL_IDLETIMEIN)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: idletime-in = %d", config_entries, conf[config_entries].idletime_in);
#endif		
	}
	else
	{
		conf[config_entries].idletime_in = IDLEIN_DEFAULT;
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "WARNING, no idletime-in in entry %d, using %d", config_entries, conf[config_entries].idletime_in);
#endif		
	}
	
	if(flags & FL_IDLETIMEOUT)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: idletime-out = %d", config_entries, conf[config_entries].idletime_out);
#endif		
	}
	else
	{
		conf[config_entries].idletime_out = IDLEOUT_DEFAULT;
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "WARNING, no idletime-out in entry %d, using %d", config_entries, conf[config_entries].idletime_out);
#endif		
	}
	
	if(flags & FL_BLOCKSIZE)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: blocksize = %d", config_entries, conf[config_entries].blocksize);
#endif		
	}
	else
	{
		conf[config_entries].blocksize = BLOCKSZ_MAX;
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "WARNING, no blocksize in entry %d, using %d", config_entries, conf[config_entries].blocksize);
#endif		
	}
	
	if(flags & FL_UNITLENDYN)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: unitlength-dyn = %d", config_entries, conf[config_entries].unitlength_dyn);
#endif		
	}
	else
	{
		conf[config_entries].unitlength_dyn = ULDYN_OFF;
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "WARNING, no unitlength-dyn in entry %d, using %d", config_entries, conf[config_entries].unitlength_dyn);
#endif		
	}

	if(flags & FL_CALLEDBACKWAIT)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: calledbackwait max time = %d", config_entries, conf[config_entries].calledbackwait);
#endif		
	}
	else
	{
		conf[config_entries].calledbackwait = CDBW_DEFAULT;
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "WARNING, no calledbackwait max time in entry %d, using %d", config_entries, conf[config_entries].calledbackwait);
#endif		
	}
	
	if(flags & FL_RATETYPE)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: ratetyp = %d", config_entries, conf[config_entries].ratetyp);
#endif		
	}
	else
	{
		conf[config_entries].ratetyp = INVALID_RATE;
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "WARNING, no rate type in entry %d, using %d", config_entries, conf[config_entries].ratetyp);
#endif		
	}

	if(flags & FL_UNITLENGTH)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: unitlength = %d", config_entries, conf[config_entries].unitlength);
#endif		
	}
	else
	{
		conf[config_entries].unitlength = -1;
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "WARNING, no unitlength in entry %d", config_entries);
#endif		
	}

	if(flags & FL_EARLYHANGUP)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "entry %d: earlyhangup = %d", config_entries, conf[config_entries].earlyhangup);
#endif		
	}
	else
	{
		conf[config_entries].earlyhangup = EARLYHANGUP_DEFAULT;
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "WARNING, no earlyhangup in entry %d, using %d", config_entries, conf[config_entries].earlyhangup);
#endif		
	}

	/* checks done, now do initialization for one config entry */
	
	if((init_subsys(config_entries) == ERROR))
		ret = ERROR;

/* -------------------- configure timeout method ----------------- */

/* unit length time specified on commandline = HIGHEST PRIORITY */

	if(got_unitlen)
	{
		conf[config_entries].unit_length_typ = ULTYP_CMDL;
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "unit length source = commandline value (%d secs)", unit_length);
#endif
	}	

/* use dynamic unit length calculation ? */

	else if(flags & FL_UNITLENDYN)
	{
		conf[config_entries].unit_length_typ = ULTYP_DYN;
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "unit length source = dynamically calculated");
#endif
	}

/* use unit length from ratesfile ? */

	else if(got_rate && (flags & FL_RATETYPE))
	{
		conf[config_entries].unit_length_typ = ULTYP_RATE;
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "unit length source = rate file entry %d", conf[config_entries].ratetyp);
#endif
	}

/* use a fixed unit length config file ? */

	else if(flags & FL_UNITLENGTH)
	{
		conf[config_entries].unit_length_typ = ULTYP_CONF;
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "unit length source = config file unitlength (%d secs)", conf[config_entries].unitlength);
#endif
	}

/* last resort, use a default timeout */

	else
	{
		conf[config_entries].unit_length_typ = ULTYP_NONE;
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "Line idle hangup timeout method = unconfigured (default %d secs)", UNITLENGTH_DEFAULT);
#endif
	}

/* -------------------- configure controller isdn type ----------------- */

	sprintf(device, "%s%d", ICARDDEVICE, conf[config_entries].controller);

	if((f = open(device, O_RDWR)) < 0)
	{
		sprintf(error, "error open %s: %s", device, sys_errlist[errno]);
		config_error = error;
		return(ERROR);
	}

	if (ioctl(f, ISDN_GET, &buf) < 0)
	{
		sprintf(error, "ioctl ISDN_GET error on %s: %s", device, sys_errlist[errno]);
		config_error = error;
		return(ERROR);
	}
	
	switch(buf)
	{
		case ISDNDSS1:
			if(conf[config_entries].isdntype == ISDN1TR6)
				newbuf = ISDN1TR6;
			break;

		case ISDN1TR6:
			if(conf[config_entries].isdntype == ISDNDSS1)
				newbuf = ISDNDSS1;
			break;

		default:
			sprintf(error, "ioctl ISDN_GET returned unknown ISDN type from %s", device);
			config_error = error;
			return(ERROR);
			break;
	}

#if DEBUG
	if(do_debug & DEBUGL_NORM)
	{
		log(LL_DBG, "device %s currently set to ISDN protocol %s%s",
			device,
			buf == ISDNDSS1 ? "DSS1" : "1TR6",
			buf == conf[config_entries].isdntype ? ", OK!" : "");
	}
#endif

	if(buf != conf[config_entries].isdntype)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "setting ISDN protocol for device %s to %s", device, conf[config_entries].isdntype == ISDNDSS1 ? "DSS1" : "1TR6");
#endif
		if (ioctl(f, ISDN_SET, &newbuf) < 0)
		{
			sprintf(error, "ioctl ISDN_SET error on %s: %s", device, sys_errlist[errno]);
			config_error = error;
			return(ERROR);
		}
	}

	close(f);	
		
	return(ret);
}

/*---------------------------------------------------------------------------*
 *	validate all configuration entries
 *---------------------------------------------------------------------------*/
static int
check_config(void)
{
	return(GOOD);
}

/*---------------------------------------------------------------------------*
 *	initialize ip/tty/tel subsystem
 *---------------------------------------------------------------------------*/
static int
init_subsys(int config_entries)
{

	switch(conf[config_entries].atyp)
	{
		case ATYP_IP:
			config_ip(conf[config_entries].appl, &conf[config_entries]);
			break;
#ifdef NOTDEF			
		case ATYP_TTY:
			config_tty(conf[config_entries].appl, &conf[config_entries]);
                        break;
#endif
		case ATYP_TEL:
			config_tel(conf[config_entries].appl, &conf[config_entries]);
			break;
	}
	return(GOOD);
}

/*---------------------------------------------------------------------------*
 *	validate system configuration section
 *---------------------------------------------------------------------------*/
static int
check_system(unsigned int flags)
{
	if(flags & FL_RATESFILE)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "system: user specific ratesfile = %s", &ratesfile[0]);
#endif		
	}
	else
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "system: system default ratesfile = %s", &ratesfile[0]);
#endif		
	}

	if(flags & FL_ACCTFILE)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "system: user specific acctfile = %s", &acctfile[0]);
#endif		
	}
	else
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "system: system default acctfile = %s", &acctfile[0]);
#endif		
	}
		
	if(flags & FL_USEACCTFILE)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "system: user specific useacctfile = %d", useacctfile);
#endif		
	}
	else
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "system: system default useacctfile = %d", useacctfile);
#endif		
	}

	if(flags & FL_ACCTALL)
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "system: write accounting info for all connections");
#endif		
	}
	else
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "system: write accounting info just for charged connections");
#endif		
	}

	if(flags & FL_REGEXPR)
	{
		int i;

		for(i = 0; i < MAX_RE; i++)
		{
			if(rarr[i].re_flg)
			{
				char path[MAXPATHLEN+1];

#if DEBUG
				if(do_debug & DEBUGL_NORM)
				{
					log(LL_DBG, "system: regexp slot %d: expr=[%s] prog=[%s]",
						i, rarr[i].re_expr, rarr[i].re_prog);
				}
#endif
				sprintf(path, "%s/%s", ETCPATH, rarr[i].re_prog);
				
				if((access(path, X_OK|F_OK)) == -1)
				{
					sprintf(error, "regexpr program %s, access failed: %s", path, strerror(errno));
					config_error = error;
					return(ERROR);
				}
			}
			else
			{
				break;
			}
		}
	}
	else
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "system: no regular expressions defined");
#endif		
	}

	if(flags & FL_PLAYPROG)
	{
		char path[MAXPATHLEN+1];

		sprintf(path, "%s/%s", ETCPATH, playprog);
		
		if((access(path, X_OK|F_OK)) == -1)
		{
#if DEBUG
			if(do_debug & DEBUGL_NORM)
				log(LL_DBG, "system: WARNING, cannot access user defined playprog = %s", playprog);
#endif		
		}
		else
		{
#if DEBUG
			if(do_debug & DEBUGL_NORM)
				log(LL_DBG, "system: user defined playprog = %s", playprog);
#endif		
		}
	}
	else
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "system: system default playprog = %s", playprog);
#endif		
	}

	if(flags & FL_ANSWPROG)
	{
		char path[MAXPATHLEN+1];

		sprintf(path, "%s/%s", ETCPATH, answerprog);
		
		if((access(path, X_OK|F_OK)) == -1)
		{
#if DEBUG
			if(do_debug & DEBUGL_NORM)
				log(LL_DBG, "system: WARNING, cannot access user defined answerprog = %s", answerprog);
#endif		
		}
		else
		{
#if DEBUG
			if(do_debug & DEBUGL_NORM)
				log(LL_DBG, "system: user defined answerprog = %s", answerprog);
#endif
		}
	}
	else
	{
#if DEBUG
		if(do_debug & DEBUGL_NORM)
			log(LL_DBG, "system: system default answerprog = %s", answerprog);
#endif		
	}

	
	return(GOOD);
}

/*--------------------------------- EOF ------------------------------------*/
