/*
 * Copyright (c) 1994, 1996 Hellmuth Michaelis. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *---------------------------------------------------------------------------*
 *
 *	bisdnd - BSD ISDN management daemon
 *      ===================================
 *
 *	Last Edit-Date: [Sat Jul  6 14:05:53 1996]
 *
 *	-hm	short hold implementation
 *	-gj	add support for multiple rates
 *	-hm	logging via syslogd implemented
 *	-hm	log type implemented
 *	-hm	find_appl() return flags
 *	-hm	callback states, more states, debugging
 *	-hm	new telephone number handling
 *	-hm	converting to message structures
 *	-hm	separate charging and accounting file
 *	-hm	configurable syslog level and facility
 *	-hm	calledbackwait variable
 *	-hm	regular expression log message matching
 *	-hm	new config file default location
 *	-hm	logging with syslog levels
 *	-hm	moving pathnames to pathnames.h
 *
 *----------------------------------------------------------------------------*/

#include <curses.h>
#include <syslog.h>
#include <regex.h>
#include <sys/param.h>
#include <machine/bisdn_ioctl.h>

#include "pathnames.h"
#include "configure.h"

#define VERSION	0		/* version number	*/
#define REL	97		/* release number	*/

#define GOOD	0
#define	ERROR	(-1)
#define WARNING (-2)

#define INVALID (-1)

#if DEBUG                 /* debugging flags to specify what to display */
#define DEBUGL_NORM	0x01	/* normal/usual debug messages		*/
#define DEBUGL_RATES	0x02	/* messages related to rates		*/
#define DEBUGL_TIME	0x04	/* messages related to timing/retry	*/
#define DEBUGL_STATE	0x08	/* messages related to states changes	*/
#endif

/* telephony program names */

#define PLAY_PROG_DEF	"play"		/* program name for playing */
#define ANSW_PROG_DEF	"answ"		/* program name for answering */

#define UMASK		022		/* file creation perm mask */

#define NCARDS		4		/* max # of cards (not supported ) */

#define NTTY		(NITTY + NOTTY)

/* window dimensions */

#define UPPER_B		2		/* upper window start  */

/* vertical stdscr positions */

#define UPPER_HEAD	0		/* upper header */
#define LOWER_HEAD	10		/* lower header */

/* horizontal positions for upper window */

#define H_CNTL		0		/* controller */
#define H_CHAN		(H_CNTL+4)	/* channel */
#define H_TELN		(H_CHAN+3)	/* telno */
#define H_SERV		(H_TELN+21)	/* service (4) */
#define	H_APPL		(H_SERV+10)	/* application (6) */
#define H_IO		(H_APPL+10)	/* input/output (3) */
#define H_OUT		(H_IO+6)	/* bytes out (8) */
#define H_IN		(H_OUT+10)	/* bytes in (8) */
#define H_UNITS		(H_IN+10)	/* charging units */

/* logfile generation */

#define LOGBUF		80

/* configuration entry structure */

#define MAX_LINELENGTH	1024
#define MAX_CONFIGLINES	16

struct confdata {
	char interface[16];	/* interface name 			*/
	int  atyp;		/* type: ip/tty/telephony		*/
	int  appl;		/* application id			*/
	char name[16];		/* name for accounting purposes 	*/
	int  controller;	/* controller for outgoing calls 	*/

	int  isdntype;		/* which isdn protocol to use 		*/

	char telnloc_ldo[TELNO_MAX];	/* local number for dialing out	*/
	char telnrem_ldo[TELNO_MAX];	/* remote number for dialing out*/
	char telnloc_rdi[TELNO_MAX];	/* local number for dialing in	*/
	char telnrem_rdi[TELNO_MAX];	/* remote number for dialing in	*/	

	int  reaction;		/* accept/reject/callback 		*/
#define REACT_ACCEPT		0
#define REACT_REJECT		1
#define REACT_CALLBACK		2
#define REACT_CALLEDBACK	3
#define REACT_ALERTBASE		4

	int  service;		/* service 				*/
#define	SERV_PHONE	1
#define SERV_DATA	2
#define SERV_CEPT	3

	int  protocol;		/* protocol to use 			*/
#define PROT_RAW	0
#define PROT_UI		1
#define PROT_X75	2
#define PROT_T70	3
#define PROT_T90	4
#define PROT_ISO	5

	short	telaction;	/* telephony: answer/playback		*/
#define ACT_ANSWER	0		/* answer call 			*/
#define ACT_PLAYBK	1		/* play back 			*/

	int  dialretries;	/* no of times to try to dial		*/
#define DIAL_DEFAULT	3

	int  recoverytime;	/* recovery time before dial retry	*/
#define RECV_DEFAULT	1

	int  callbackwait;	/* time to wait before calling back	*/
#define CBW_DEFAULT	1

	int  idletime_in;	/* max idle time (incoming calls)	*/
#define IDLEIN_DEFAULT	0
	int  idletime_out;	/* max idle time (outgoing calls)	*/	
#define IDLEOUT_DEFAULT	60

	int  unitlength_dyn;	/* use dynamic unit length calculation	*/
#define ULDYN_OFF	0
#define ULDYN_ON	1

	int  blocksize;		/* ISDN block size			*/
#define BLOCKSZ_MAX	2048

	int  unit_length_typ;	/* where we get the unit length from	*/
#define ULTYP_NONE	0		/* nowhere specified		*/
#define ULTYP_CMDL	1		/* specified on commandline	*/
#define   ULTYP_CMDLMIN	5		/* minimum value from cmdl	*/
#define   ULTYP_CMDLMAX	3600		/* minimum value from cmdl	*/
#define ULTYP_CONF	2		/* get it from config file	*/
#define ULTYP_RATE	3		/* get it dynamic from ratesfile*/
#define ULTYP_DYN	4		/* dynamic calculated unit lenght */

	int  unitlength;		/* length of a charging unit	*/
#define UNITLENGTH_DEFAULT	60	/* last resort unit length	*/

	int  earlyhangup;		/* time in seconds to hangup 	*/
					/* before the next expected	*/
					/* charging unit		*/
#define EARLYHANGUP_DEFAULT	5
					
	int ratetyp;		/* type of rate	*/
#define NO_RATE		(NRATES+1)
#define INVALID_RATE	(-1)

	int  calledbackwait;	/* max time to wait for being called back */
#define CDBW_DEFAULT	10

/******************** dynamically used entries **************************/

	int aoc_last;		/* dyn idle: last AOCD timestamp	*/
	int aoc_now;		/*           current AOCD timestamp	*/
	int aoc_diff;		/*           current unit length	*/
	int aoc_lastdiff;	/*           last charge unit length	*/
	int aoc_valid;		/*           time diff is valid flag	*/
#define AOC_INVALID	0		/* aoc_diff is NOT valid	*/
#define AOC_VALID	1		/* aoc_diff is valid		*/

	time_t	lasttime;	/* last time dialed			*/
	int	curtry;		/* current number of tries to dial	*/

	int	state;		/* state of connection 			*/
#define IDLE		0		/* connection is idle		*/
#define	DIALING		1		/* dialing out 			*/
#define AWCB_WAITDISC	2		/* dialing, await disconnect and callback */
#define AWAITCALLBK	3		/* waiting for callback from remote */
#define CB_WAITDISC	4		/* got call, wait for disconnect*/
					/*  before calling back remote	*/
#define CONNECTED	5		/* connected with remote	*/

	int	charge;		/* amount of charge in units		*/

	char	telno_remote[TELNO_MAX];/* incoming telephone number	*/

	time_t	connect_time;	/* the time a connection was made	*/

	int	inbytes;	/* number of bytes in			*/
	int	outbytes;	/* number of bytes out			*/

	int	channel;	/* channel used for connection		*/
};

#define NDAYS	7	/* number of days */
#define NRATES	4	/* number of rate structures supported */

/* struct for rates - each day has one or more */

struct rates
{
	int start_hr;	/* hour at which this rate starts, e.g. 12 */
	int start_min;	/* minute of start ... */
	int end_hr;	/* hour at which this rate ends, e.g. 19 */
	int end_min;	/* minute of end ... */
	int rate;	/* how long can I telephone at this price, seconds */
	struct rates *next;	/* a day can have multiple rates */
};

struct answ
{
	short an;
	short cn;
	struct confdata *conf;
	char dst_telno[TELNO_MAX];
	char src_telno[TELNO_MAX];
};

struct alert
{
	short cnt;
	u_short id;
	time_t lastact;
	accept_t *accept;
};

/* find_appl return values/flags (if not GOOD or ERROR) */

#define FA_GOOD		0	/* everything fits, ok to connect	*/
#define FA_ERROR	(-1)	/* unknown (severe) error		*/
#define FA_NOTME	(-2)	/* not my number			*/
#define FA_CARD		(-3)	/* invalid card number			*/
#define FA_ATYP		(-4)	/* invalid application type		*/
#define FA_INV_DEST	0x01	/* my number, but invalid destination	*/
#define FA_INV_SERV	0x02	/* my number, but invalid service	*/
#define FA_INV_CTRL	0x04	/* my number, but invalid controller	*/

/* tables: numbers to text for display */

#define NO_APPL		4
#define SERVICEMASK	0x03
#define NO_SERVICES	4

#define REGPROG_DEF	"program"	/* default program to use */

struct rarr {
	int re_flg;		/* valid entry flag */
	char *re_expr;		/* plain text expression */
	regex_t re;		/* compiled expression */
	char *re_prog;		/* the program to be executed */
};

/* the internal identifiers for bisdn log levels. CAUTION: this has to stay */
/* in sync with the loglevel to text and sysloglevel table in log.c !! */

enum logids
{
	LL_ERR,		/* information messages			*/	
	LL_INF,		/* information messages			*/
	LL_SYS,		/* messages about daemon status etc.	*/
	LL_ACT,		/* accounting messages			*/
	LL_NAC,		/* negative accept messages		*/
	LL_ACP,		/* call accepted messages		*/
	LL_NDL,		/* negative dial messages		*/
	LL_DIA,		/* dialing ok messages			*/
	LL_NCN,		/* negative connection messages		*/
	LL_CON,		/* connection messages			*/
	LL_DIS,		/* disconnection messages		*/
	LL_CRG,		/* charging messages			*/
	LL_ALR		/* alert messages			*/
#if DEBUG
	,LL_DBG		/* debug messages			*/
#endif
};

#ifdef MAIN

int do_fork = 1;		/* fork bisdnd, default yes */
int do_fullscreen = 0;		/* use fullscreen display, default no */

#if DEBUG
int do_debug = 0;		/* no debugging messages by default */
#endif

int do_rdev = 0;		/* flag, switch to another output device */
int do_ttytype = 0;		/* flag, use another termcap entry */
char *rdev = "";		/* new device string */
char *ttype = "";		/* new termcap entry name string */

time_t unit_length;		/* calculated length of a unit */
int got_rate = 0;		/* flag, ratesfile found */
int got_unitlen = 0;		/* flag, user specified the length of a unit */

int curses_ready = 0;		/* flag, curses display is initialized */

int uselogfile = 0;		/* flag, use logfile instead of syslog */
int useacctfile = 0;		/* flag, write accounting information into file */

int logfacility = LOG_LOCAL0;	/* the facility used for syslog output */

FILE *acctfp = NULL;		/* the accounting file pointer */

char *configfile = CONFIG_FILE_DEF;		/* configuration filename */

char logfile[MAXPATHLEN] = LOG_FILE_DEF;	/* log filename */
char ratesfile[MAXPATHLEN] = RATESFILE;		/* rates filename */
char acctfile[MAXPATHLEN] = ACCTFILE;		/* accounting  filename */
char answerprog[MAXPATHLEN] = ANSW_PROG_DEF;	/* play program name */
char playprog[MAXPATHLEN] = PLAY_PROG_DEF;	/* answer program name */

WINDOW *upper_w;		/* curses upper window pointer */
WINDOW *lower_w;		/* curses lower window pointer */

/* table to map application id and services to strings for display */
char *appl_tab[NO_APPL] = {"IP", "Login", "Answer", "Play"};
char *serv_tab[NO_SERVICES] = {"???", "Tel", "Data", "CEPT"};

int config_entries = 0;		/* number of entries in config array	*/
int numcards = 0;		/* number of cards configured		*/
int acct_all = 1;		/* all connections into accounting file */

char *config_error = NULL;	/* in case of error: error message	*/
char *rate_error = NULL;	/* in case of error: error message	*/

struct confdata conf[MAX_CONFIGLINES];	/* the configuration structure */

struct answ answ[NITEL];	/* the answer structures */
struct rates *rates[NRATES][NDAYS];	/* the rates structure */
struct alert alert[NALERT];		/* the alert structure */

#if DEBUG
char *statetab[] = {		/* map states to names for debugging */
	"IDLE",
	"DIALING",
	"AWCB_WAITDISC",
	"AWAITCALLBK",
	"CB_WAITDISC",
	"CONNECTED",
	NULL
};
#endif

struct rarr rarr[MAX_RE]; /* regexpr & progs table for log pattern matching */

#else /* !MAIN */

extern time_t unit_length;
extern int got_rate;
extern int got_unitlen;
extern int do_fork;
extern int do_fullscreen;

#if DEBUG
extern int do_debug;
#endif

extern int do_rdev;
extern int do_ttytype;
extern char *rdev;
extern char *ttype;
extern int curses_ready;

extern int uselogfile;
extern int useacctfile;

extern int loglevel;
extern int logfacility;

extern FILE *acctfp;

extern char *configfile;

extern char logfile[];
extern char ratesfile[];
extern char acctfile[];
extern char answerprog[];
extern char playprog[];

extern WINDOW *upper_w;
extern WINDOW *lower_w;
extern char *appl_tab[];
extern char *serv_tab[];

extern int config_entries;
extern int numcards;
extern int acct_all;

extern char *config_error;
extern char *rate_error;

extern struct confdata conf[];

extern struct answ answ[];
extern struct rates *rates[NRATES][NDAYS]; /* gcc demands the sizes */

extern struct alert alert[];

#if DEBUG
extern char *statetab[];
#endif

extern struct rarr rarr[MAX_RE];

#endif /* MAIN */

extern void cold_init_ip ( void );
extern void cold_init_tel ( void );
extern int config_ip ( int appl_no, struct confdata *confdata_p );
extern int config_tel ( int appl_no, struct confdata *cp );
extern int connected ( accept_t **accept_pp, int isdnfd, int atyp, int appl_no, int chan, int active );
extern accept_t * connect_ip ( int appl_no, int channel, int active);
extern accept_t * connect_tel ( int appl_no, int channel, int active);
extern void disconnected ( int isdnfd, int atyp, int appl_no );
extern void disconnect_ip ( int appl_no );
extern void disconnect_tel ( int appl_no );
extern void do_quit ( int exitval );
extern int find_appl ( accept_t **accept_pp, int chan, int service, int alert_id, char *dst_telno, char *src_telno );
extern void finish_log ( void );
extern struct confdata *getcp ( int atyp, int appl );
extern struct confdata *getcpbycc( int controller, int channel );
extern char *getlogdatetime( void );
extern dial_t * getno_ip ( int appl_no );
extern dial_t * getno_tel ( int appl_no );
extern int get_current_rate ( struct confdata *cp, int logit );
extern int get_telno ( dial_t **dial_pp, int atyp, int appl_no );
extern void hangup( int isdnfd, struct confdata *cp );
extern void handle_recovery ( int isdnfd );
extern void info_1tr6 ( int isdnfd, u_char *s );
extern void info_dss1 ( int isdnfd, u_char *s );
extern void init_log ( void );
extern int ip_fill_accept ( accept_t **accept_pp, int channel, struct confdata *confdata_p );
extern void log ( int what, const char *fmt, ... );
extern int main ( int argc, char **argv );
extern void msg_accept ( int isdnfd, unsigned char *rbuf );
extern void msg_connect ( int isdnfd, unsigned char *rbuf );
extern void msg_dial ( int isdnfd, unsigned char *rbuf );
extern void msg_disconnect ( int isdnfd, unsigned char *rbuf );
extern void msg_info ( int isdnfd, unsigned char *rbuf );
extern void msg_ipacct ( int isdnfd, unsigned char *rbuf );
extern int readconf ( char *filename );
extern int readrates ( char *filename );
extern int tel_fill_accept ( accept_t **accept_pp, int channel, struct confdata *confdata_p, char *dst_telno, char *src_telno );

/* EOF */
