/*
 * Copyright (c) 1994, 1996 Hellmuth Michaelis. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *---------------------------------------------------------------------------*
 *
 *	bisdnd - BSD ISDN management daemon
 *      ===================================
 *
 *	Last Edit-Date: [Fri Jul  5 12:12:51 1996]
 *
 *	-hm	implemented logging
 *	-hm	implementing logging via syslogd
 *	-hm	no newlines in the log strings anymore
 *	-hm	logid recorded with log messages
 *	-hm	background debug logging
 *	-hm	runtime choice of logfile or syslog logging
 *	-hm	configurable syslog facility and level
 *	-hm	regular expression log message matching
 *	-hm	logging with syslog levels
 *
 *----------------------------------------------------------------------------*/

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <errno.h>
#include <syslog.h>
#include <unistd.h>
#include <regex.h>

#include "defs.h"

#define LOGBUFLEN 256

static FILE *logfp = NULL;

static void check_reg(char *logstring);
static void exec_prog(char *prog, char *argument);

struct logtab {
	char *text;
	int pri;
};

static struct logtab logtab[] = {
	{"ERR",	LOG_ERR},	/* fatal errors */
	{"INF",	LOG_INFO},	/* information messages */
	{"SYS",	LOG_NOTICE},	/* messages about daemon status etc. */
	{"ACT",	LOG_INFO},	/* accounting messages */
	{"NAC",	LOG_NOTICE},	/* negative accept messages */
	{"ACP",	LOG_INFO},	/* call accepted messages */
	{"NDL",	LOG_NOTICE},	/* negative dial messages */
	{"DIA",	LOG_INFO},	/* dialing ok messages */
	{"NCN",	LOG_NOTICE},	/* negative connection messages */
	{"CON",	LOG_INFO},	/* connection messages */
	{"DIS",	LOG_INFO},	/* disconnection messages */
	{"CRG",	LOG_INFO},	/* charging messages */
	{"ALR",	LOG_NOTICE},	/* alert messages */
#if DEBUG
	{"DBG", LOG_DEBUG}	/* debug messages */
#endif	
};

/*---------------------------------------------------------------------------*
 *	initialize logging
 *---------------------------------------------------------------------------*/
void
init_log(void)
{
	int i;

	if(uselogfile)
	{
		if((logfp = fopen(logfile, "a")) == NULL)
		{
			fprintf(stderr, "ERROR, cannot open logfile %s: %s\n",
				logfile, strerror(errno));
			exit(1);
		}
	
		/* set unbuffered operation */
	
		setvbuf(logfp, (char *)NULL, _IONBF, 0);
	}
	else
	{
#if DEBUG
		if(do_debug > 0 && do_fork == 0 && do_fullscreen == 0)
			(void)openlog("bisdnd",
				LOG_PID|LOG_CONS|LOG_NDELAY|LOG_PERROR,
				logfacility);
		else
#endif
		(void)openlog("bisdnd", LOG_PID|LOG_CONS|LOG_NDELAY,
				logfacility);
	}

	/* initialize the regexp array */

	for(i = 0; i < MAX_RE; i++)
	{
		char *p;
		char buf[64];

		sprintf(buf, "%s%d", REGPROG_DEF, i);

		rarr[i].re_flg = 0;

		if((p = malloc(strlen(buf) + 1)) == NULL)
		{
			log(LL_DBG, "init_log: malloc failed: %s", strerror(errno));
			do_quit(1);
		}

		strcpy(p, buf);

		rarr[i].re_prog = p;
	}
}

/*---------------------------------------------------------------------------*
 *	finish logging
 *---------------------------------------------------------------------------*/
void
finish_log(void)
{
	if(uselogfile)
	{
		fflush(logfp);
		fclose(logfp);
	}
	else
	{
		(void)closelog();
	}
}

/*---------------------------------------------------------------------------*
 *	place entry into logfile
 *---------------------------------------------------------------------------*/
void
log(int what, const char *fmt, ...)
{
	char buffer[LOGBUFLEN];
	va_list ap;

	va_start(ap, fmt);
	
	vsnprintf(buffer, LOGBUFLEN-1, fmt, ap);
	
	if(uselogfile)
	{
		register char *dp;
	
		dp = getlogdatetime();	/* get time string ptr */
	
		if(do_fullscreen && curses_ready)
		{
			wprintw(lower_w, "%s %s %-.*s\n", dp, logtab[what].text, COLS-22, buffer);
			wrefresh(lower_w);
		}
		fprintf(logfp, "%s %s %s\n", dp, logtab[what].text, buffer);
	}
	else
	{
		register char *s = buffer;
		
		if(do_fullscreen && curses_ready)	/* put log on screen ? */
		{
			wprintw(lower_w, "%s %s %-.*s\n", getlogdatetime(), logtab[what].text, COLS-22, buffer);
			wrefresh(lower_w);
		}

		/* strip leading spaces from syslog output */
		
		while(*s == ' ')
			s++;
			
		syslog(logtab[what].pri, "%s %s", logtab[what].text, s);
	}

	va_end(ap);

#if DEBUG
	if(what != LL_DBG) /* don't check debug logs, endless loop !!! */
#endif
		check_reg(buffer);
}

/*---------------------------------------------------------------------------*
 *	return ptr to static area containing date/time
 *---------------------------------------------------------------------------*/
char *
getlogdatetime()
{
	static char logdatetime[41];
	time_t tim;
	register struct tm *tp;
	
	tim = time(NULL);
	tp = localtime(&tim);
	strftime(logdatetime,40,"%d.%m.%y %H:%M:%S",tp);
	return(logdatetime);
}

/*---------------------------------------------------------------------------*
 *	check for a match in the regexp array
 *---------------------------------------------------------------------------*/
static void
check_reg(char *logstring)
{
	register int i;

	for(i = 0; i < MAX_RE; i++)
	{
		if(rarr[i].re_flg && (!regexec(&(rarr[i].re), logstring, (size_t) 0, NULL, 0)))
		{
#if DEBUG
			if(do_debug & DEBUGL_NORM)
				log(LL_DBG, "              check_reg: match, exec=[%s] arg=[%s]", rarr[i].re_prog, logstring);
#endif				
			exec_prog(rarr[i].re_prog, logstring);
			break;
		}
	}
}

/*---------------------------------------------------------------------------*
 *	execute prog as a subprocess and pass an argumentstring
 *---------------------------------------------------------------------------*/
static void
exec_prog(char *prog, char *argument)
{
	char path[MAXPATHLEN+1];
	
	switch (fork())
	{
		case -1:		/* error */
			log(LL_SYS, "ERROR, exec_prog/fork: %s", strerror(errno));
			return;
		case 0:			/* child */
			break;
		default:		/* parent */
			return;
	}

	/* this is the child now */

	sprintf(path, "%s/%s", ETCPATH, prog);

	if((execlp(path, prog, argument, (char *) 0 )) < 0)
	{
		log(LL_SYS, "CHILD-ERROR, exec_prog/execlp: %s", strerror(errno));
		_exit(127);
	}
}		

/* EOF */
