/*
 * Copyright (c) 1994, 1996 Hellmuth Michaelis. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *---------------------------------------------------------------------------*
 *
 *	bisdnd - BSD ISDN management daemon
 *      ===================================
 *
 *	Last Edit-Date: [Fri Jul  5 12:14:50 1996]
 *
 *	-hm	read() error checking
 *	-hm	passive callback handling (i'm being called back)
 *	-hm	alarm time command line parm
 *	-hm	ISDN_DONTACCEPT ioctl
 *	-hm	retry/timeout mechanism implementation and debugging
 *	-hm	timing call gate to timeout/retry functions
 *	-hm	added ip accounting info processing
 *	-hm	integrated advice of charge decoding Q.931/Q.932
 *	-hm	new ip accounting format
 *	-hm	info 'i' messages now carry the isdn protocol type
 *	-hm	adding info_1tr6()
 *	-hm	adding "s:" in front of every bisdnd-originated log msg
 *	-hm	1TR6 charging information decoding
 *	-hm	made handle_charge more generally
 *	-hm	new states for callback operation
 *	-hm	timeout/idletime fixes
 *	-hm	adding support for NetBSD from Arne
 *	-hm	logging to syslog or logfile, new config system keyword
 *	-hm	implemented menu driven commandmode in fullscreen mode
 *	-hm	implement timeout for menu display
 *	-hm	patch from Andreas for the daemon pid syslog message
 *	-hm	logfile and facility selection from commandline
 *	-hm	regular expression log message matching
 *
 *----------------------------------------------------------------------------*/

#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>

#ifdef __NetBSD__
#include <sys/ioctl.h>
#endif /* __NetBSD__ */

#include <sys/stat.h>
#include <sys/wait.h>

#define MAIN
#include "defs.h"

#include "config.h"

static void catchsig();
static void catchsighup();
static void catchsigio();
static void catchsigchld();
static void daemonize (void );
static void usage ( void );
static void do_rescan( void );
static void check_pid( void );
static void write_pid( void );
static void init_screen( void );
static void do_menu( void );

static int isdnfd;

/*---------------------------------------------------------------------------*
 *	program entry
 *---------------------------------------------------------------------------*/
int
main(int argc, char **argv)
{
	int i;

#if DEBUG
	while ((i = getopt(argc, argv, "c:d:fFlL:r:s:t:u:?")) != EOF)
#else
	while ((i = getopt(argc, argv, "c:fFlL:r:s:t:u:?")) != EOF)
#endif
	{
		switch (i)
		{
			case 'c':
				configfile = optarg;
				break;
#if DEBUG
			case 'd':
				if(isdigit(*optarg))
				{
					do_debug = strtoul(optarg, NULL, 10);
				}
				else
				{
					fprintf(stderr, "Error: option -d requires a numeric argument!\n");
					usage();
				}
				break;
#endif
			case 'f':
				do_fullscreen = 1;
				do_fork = 0;			
				break;

			case 'F':
				do_fork = 0;
				break;

			case 'l':
				uselogfile = 1;
				break;

			case 'L':
				strcpy(logfile, optarg);
				break;

			case 'r':
				rdev = optarg;
				do_rdev = 1;
				break;

			case 's':
				if(isdigit(*optarg))
				{
					int facility;
					logfacility = strtoul(optarg, NULL, 10);
					facility = logfacility << 3;

					if((facility < LOG_KERN) ||
					   (facility > LOG_FTP && facility < LOG_LOCAL0) ||
					   (facility > LOG_LOCAL7))
					{
						fprintf(stderr, "Error, option -s has invalid logging facility %d", logfacility);
						usage();
					}
					logfacility = facility;
				}
				else
				{
					fprintf(stderr, "Error: option -s requires a numeric argument!\n");
					usage();
				}
				break;

			case 't':
				ttype = optarg;
				do_ttytype = 1;
				break;

			case 'u':
				if(isdigit(*optarg))
				{
					unit_length = strtoul(optarg, NULL, 10);
					if(unit_length < ULTYP_CMDLMIN)
						unit_length = ULTYP_CMDLMIN;
					else if(unit_length > ULTYP_CMDLMAX)
						unit_length = ULTYP_CMDLMAX;
					got_unitlen = 1;
				}
				else
				{
					fprintf(stderr, "Error: option -T requires a numeric argument!\n");
					usage();
				}
				break;

			case '?':
			default:
				usage();
				break;
		}
	}

	init_log();	/* initialize the logging subsystem */

	check_pid();	/* check if we are already running */
	
	if(do_fork || (do_fullscreen && do_rdev))
		daemonize();	/* become a daemon */

	umask(UMASK);	/* set our umask ... */

	/* write our pid file to protect us against multiple invocations */
	
	write_pid();

	/* set signal handlers */
	
	(void) signal(SIGHUP, catchsighup);
	(void) signal(SIGCHLD, catchsigchld);
	
	(void) signal(SIGTERM, catchsig);
	(void) signal(SIGKILL, catchsig);
	(void) signal(SIGINT, catchsig);
	(void) signal(SIGQUIT, catchsig);

	/* open isdn device */
	
	if((isdnfd = open(ISDNDEVICE, O_RDWR)) < 0)
	{
		log(LL_ERR, "cannot open %s: %s", ISDNDEVICE, strerror(errno));
		exit(1);
	}

	/* handle the rates stuff */
	
	if((i = readrates(ratesfile)) == ERROR)
	{
		if(rate_error != NULL)
			log(LL_ERR, "%s", rate_error);
		exit(1);
	}

	if(i == GOOD)
	{
		got_rate = 1;	/* flag, ratefile read and ok */
	}
	else
	{
		if(rate_error != NULL)
			log(LL_SYS, "%s", rate_error);
	}

	/* cold init subsystems */
	
	cold_init_ip();		/* cold init ip subsystem */
#ifdef NOTDEF
	cold_init_tty();	/* cold init tty subsystem */
#endif
	cold_init_tel();	/* cold init phone subsystem */

	/* read runtime configuration file and configure ourselves */
	
	if((readconf(configfile)) == ERROR)
	{
		if(config_error != NULL)
			log(LL_ERR, "%s", config_error);
		else
			log(LL_ERR, "re-readconf: unknown error!");
		exit(1);
	}

	/* if writing accounting info, open file, set unbuffered */
	
	if(useacctfile)
	{
		if((acctfp = fopen(acctfile, "a")) == NULL)
		{
			log(LL_ERR, "ERROR, can't open acctfile %s for writing, terminating!", acctfile);
			exit(1);
		}
		setvbuf(acctfp, (char *)NULL, _IONBF, 0);		
	}

	/* in case fullscreendisplay, initialize */

	if(do_fullscreen)
	{
		init_screen();
	}

	/* reset the disdn softc message header buffer for each card */

	i = 0;
	
	if(ioctl(isdnfd, ISDN_DISDNRESET, &i) < 0)
		log(LL_SYS, "              ioctl ISDN_DISDNRESET failed: %s", strerror(errno));
  		
	/* go into loop */
	
 	log(LL_SYS, "              daemon started (pid = %d)", getpid());
 
	for(;;)
	{
		unsigned char rbuf[BLOCKSZ_MAX];
		int len;
		
		if((len = read(isdnfd, rbuf, BLOCKSZ_MAX)) > 0)
		{
			switch (*rbuf)
			{
				case MSG_ACCEPT:
					msg_accept(isdnfd, rbuf);
					break;
					
				case MSG_DIAL:
					msg_dial(isdnfd, rbuf);
					break;

				case MSG_INFO:
					msg_info(isdnfd, rbuf);
					break;
					
				case MSG_IPACCT:
					msg_ipacct(isdnfd, rbuf);
					break;

				case MSG_CONNECT:
					msg_connect(isdnfd, rbuf);
					break;

				case MSG_DISCONN:
					msg_disconnect(isdnfd, rbuf);
					break;
					
				default:
					break;
			}
		}
		else
		{
			if((len == -1) && (errno != EAGAIN))			
				log(LL_SYS, "              ERROR, read error on isdn device, errno = %d!", errno);
		}

		/* handle timeout and recovery */		

		handle_recovery(isdnfd);
	}
}

/*---------------------------------------------------------------------------*
 *	exit program
 *---------------------------------------------------------------------------*/
void
do_quit(int exitval)
{
	if(exitval)
		log(LL_ERR, "              daemon terminated");
	else
		log(LL_SYS, "              daemon terminated");	

	unlink(PIDFILE);

	if(do_fullscreen && curses_ready)
	{
		mvcur(0, COLS-1, LINES-1, 0);
		endwin();
	}

	finish_log();

	exit(exitval);
}

/*---------------------------------------------------------------------------*
 *	rescan config file
 *---------------------------------------------------------------------------*/
static void
do_rescan(void)
{
	log(LL_SYS, "              daemon config file rescan");

	cold_init_ip();
#ifdef NOTDEF
	cold_init_tty();
#endif
	cold_init_tel();

	if((readconf(configfile)) != GOOD)
	{
		if(config_error != NULL)
			log(LL_ERR, "re-readconf: %s", config_error);
		else
			log(LL_ERR, "re-readconf: unknown error!");		
		exit(1);
	}
}

/*---------------------------------------------------------------------------*
 *	SIGQUIT signal handler
 *---------------------------------------------------------------------------*/
static void
catchsig()
{
	do_quit(0);
}

/*---------------------------------------------------------------------------*
 *	SIGHUP signal handler
 *---------------------------------------------------------------------------*/
static void
catchsighup()
{
	do_rescan();
}

/*---------------------------------------------------------------------------*
 *	SIGIO signal handler
 *---------------------------------------------------------------------------*/
static void
catchsigio()
{
	fd_set set;
	struct timeval timeout;

	/* get characters until no more are pending from keyboard */
	
	for(;;)
	{
		FD_ZERO(&set);
		FD_SET(STDIN_FILENO, &set);
		timeout.tv_sec = 0;
		timeout.tv_usec = 0;

		if((select(STDIN_FILENO + 1, &set, NULL, NULL, &timeout)) > 0)
		{
			int ch = getch();
		
			switch(ch)
			{
				case 0x0c:	/* control L */
					wrefresh(curscr);
					break;
		
				case '\n':
				case '\r':
					do_menu();
					break;
			}
		}
		else
		{
			break;
		}
	}
}

/*---------------------------------------------------------------------------*
 *	SIGCHLD signal handler
 *---------------------------------------------------------------------------*/
static void
catchsigchld()
{
	int retstat;
	pid_t pid;
	
	if((pid = waitpid(-1, &retstat, WNOHANG)) <= 0)
	{
		log(LL_SYS, "              ERROR, waitpid: %s", strerror(errno));
	}
	else
	{
		if(WIFEXITED(retstat))
		{
			log(LL_SYS, "              normal child (pid=%d) termination, exitstat = %d",
				pid, WEXITSTATUS(retstat));
		}
		else if(WIFSIGNALED(retstat))
		{
			if(WCOREDUMP(retstat))
				log(LL_SYS, "              child (pid=%d) termination due to signal %d (coredump)",
					pid, WTERMSIG(retstat));
			else
				log(LL_SYS, "              child (pid=%d) termination due to signal %d",
					pid, WTERMSIG(retstat));
		}
	}
}
 
/*---------------------------------------------------------------------------*
 *	check if another instance of us is already running
 *---------------------------------------------------------------------------*/
static void
check_pid(void)
{
	FILE *fp;
	
	/* check if another lock-file already exists */

	if((fp = fopen(PIDFILE, "r")) != NULL)
	{
		/* lockfile found, check */
		
		int oldpid;

		/* read pid from file */
		
		if((fscanf(fp, "%d", &oldpid)) != 1)
		{
			log(LL_ERR, "ERROR, reading pid from lockfile failed, terminating!");
			exit(1);
		}

		/* check if process got from file is still alive */
		
		if((kill( oldpid, 0)) != 0)
		{
			/* process does not exist */

			/* close file */

			fclose(fp);

			log(LL_SYS, "              removing old lock-file %s", PIDFILE);

			/* remove file */
			
			unlink(PIDFILE);
		}
		else
		{
			/* process is still alive */
			
			log(LL_ERR, "ERROR, another daemon is already running, pid = %d, terminating!", oldpid);
			exit(1);
		}
	}
}

/*---------------------------------------------------------------------------*
 *	establish and init process lock file
 *---------------------------------------------------------------------------*/
static void
write_pid(void)
{
	FILE *fp;
	
	/* write my pid into lock-file */
	
	if((fp = fopen(PIDFILE, "w")) == NULL)
	{
		log(LL_SYS, "ERROR, can't open lockfile for writing, terminating");
		do_quit(1);
	}

	if((fprintf(fp, "%d", (int)getpid())) == EOF)
	{
		log(LL_SYS, "ERROR, can't write pid to lockfile, terminating");
		do_quit(1);
	}

	fclose(fp);
}

/*---------------------------------------------------------------------------*
 *	init curses fullscreen display
 *---------------------------------------------------------------------------*/
static void
init_screen(void)
{
	char buffer[512];
	int uheight, lheight;
	int i;
	int ret;

	initscr();			/* curses init */
	
	if((COLS < 80) || (LINES < 24))
	{
		log(LL_SYS, "ERROR, minimal screensize must be 80x24, is %dx%d, terminating!",COLS, LINES);
		do_quit(1);
	}		

	noecho();
	raw();

	uheight = ( numcards + 1) * 2;	/* cards * b-channels */
	lheight = LINES - uheight - 4;	/* rest of display */
	
	if((upper_w = newwin(uheight, COLS, UPPER_B, 0)) == NULL)
	{
		log(LL_ERR, "ERROR, curses init upper window, terminating!");
		exit(1);
	}

	if((lower_w = newwin(lheight, COLS, UPPER_B+uheight+2, 0)) == NULL)
	{
		log(LL_ERR, "ERROR, curses init lower window, LINES = %d, lheight = %d, uheight = %d, terminating!", LINES, lheight, uheight);
		exit(1);
	}
	
	scrollok(lower_w, 1);

	sprintf(buffer, "----- ISDN Channel Status ----------------------- bisdnd %d.%d [pid %d] -", VERSION, REL, (int)getpid());

	while(strlen(buffer) < COLS)
		strcat(buffer, "-");	

	move(0, 0);
	standout();
	addstr(buffer);
	standend();
	
	move(1, 0);
	addstr("Ctl Ch Name/Number          Service   Applic    Dir   Out-Bytes In-Bytes  Charge");

	sprintf(buffer, "----- ISDN Logfile Display ---------------------------------------------------");
	while(strlen(buffer) < COLS)
		strcat(buffer, "-");	

	move(uheight+2, 0);
	standout();
	addstr(buffer);
	standend();

	move(uheight+3, 0);
	addstr("Date     Time     Typ Information   Description");	

	refresh();

	for(i = 0; i <= (numcards * 2); i+=2 )
	{
		mvwprintw(upper_w, i,   H_CNTL, "C%d  B1 ", i+1);
		mvwprintw(upper_w, i+1, H_CNTL, "    B2 ");
	}
	wrefresh(upper_w);

	wmove(lower_w, 0, 0);
	wrefresh(lower_w);

	(void) signal(SIGIO, catchsigio);

	ret = fcntl(0, F_GETFL, 0);
	ret |= O_ASYNC;

	if((fcntl(0, F_SETFL, ret)) < 0)
	{
		endwin();
		log(LL_ERR, "ERROR, fcntl F_SETFL for stdin: %s", strerror(errno));
		exit(1);
	}

	curses_ready = 1;
}

/*---------------------------------------------------------------------------*
 *	become a daemon
 *---------------------------------------------------------------------------*/
static void
daemonize(void)
{
	int fd;

	switch (fork())
	{
		case -1:		/* error */
			log(LL_ERR, "ERROR, daemonize/fork: %s", strerror(errno));
			exit(1);
		case 0:			/* child */
			break;
		default:		/* parent */
			exit(0);
	}

	/* new session / no control tty */

	if(setsid() == -1)
	{
		log(LL_ERR, "ERROR, setsid returns: %s", strerror(errno));
		exit(1);
	}

	/* go away from mounted dir */
	
	(void)chdir("/");

	/* move i/o to another device ? */
	
	if(do_fullscreen && do_rdev)
	{
		char *tp;
		
		if((fd = open(rdev, O_RDWR, 0)) != -1)
		{
			if(!isatty(fd))
			{
				log(LL_ERR, "ERROR, device %s is not a tty!", rdev);
				exit(1);
			}
			if((dup2(fd, STDIN_FILENO)) == -1)
			{
				log(LL_ERR, "ERROR, dup2 stdin: %s", strerror(errno));
				exit(1);
			}				
			if((dup2(fd, STDOUT_FILENO)) == -1)
			{
				log(LL_ERR, "ERROR, dup2 stdout: %s", strerror(errno));
				exit(1);
			}				
			if((dup2(fd, STDERR_FILENO)) == -1)
			{
				log(LL_ERR, "ERROR, dup2 stderr: %s", strerror(errno));
				exit(1);
			}				
		}
		else
		{
			log(LL_ERR, "ERROR, cannot open redirected device: %s", strerror(errno));
			exit(1);
		}
			
		if(fd > 2)
		{
			if((close(fd)) == -1)
			{
				log(LL_ERR, "ERROR, close in daemonize: %s", strerror(errno));
				exit(1);
			}				
		}

		/* curses output && fork NEEDS controlling tty */
		
		if((ioctl(STDIN_FILENO, TIOCSCTTY, (char *)NULL)) < 0)
		{
			log(LL_ERR, "ERROR, cannot setup tty as controlling terminal: %s", strerror(errno));
			exit(1);
		}

		/* in case there is no environment ... */

		if(((tp = getenv("TERM")) == NULL) || (*tp == '\0'))
		{
			if(do_ttytype == 0)
			{
				log(LL_ERR, "ERROR, no environment variable TERM found and -t not specified!");
				exit(1);
			}

			if((setenv("TERM", ttype, 1)) != 0)
			{
				log(LL_ERR, "ERROR, setenv TERM=%s failed: %s", ttype, strerror(errno));
				exit(1);
			}
		}
	}
}

/*---------------------------------------------------------------------------*
 *	usage display and exit
 *---------------------------------------------------------------------------*/
static void
usage(void)
{
	fprintf(stderr, "\n");
	fprintf(stderr, "bisdnd - bsd isdn manager daemon, version %d.%d, compiled %s %s\n", VERSION, REL, __DATE__, __TIME__);
#if DEBUG
	fprintf(stderr, "  usage: bisdnd [-c file] [-d level] [-F]\n");
#else
	fprintf(stderr, "  usage: bisdnd [-c file] [-F]\n");
#endif
	fprintf(stderr, "                [-f [-r dev] [-t termtype]] [-u time]\n");
	fprintf(stderr, "                [-l] [-L file] [-s facility]\n");	
	fprintf(stderr, "    -c <filename> configuration file name (def: %s)\n", CONFIG_FILE_DEF);
#if DEBUG
	fprintf(stderr, "    -d <level>    set debugging flag bits:\n");
	fprintf(stderr, "                  normal flag bit = 0x%x\n", DEBUGL_NORM);
	fprintf(stderr, "                  rates  flag bit = 0x%x\n", DEBUGL_RATES);
	fprintf(stderr, "                  time   flag bit = 0x%x\n", DEBUGL_TIME);
	fprintf(stderr, "                  states flag bit = 0x%x\n", DEBUGL_STATE);	
#endif
	fprintf(stderr, "    -f            fullscreen status display\n");
	fprintf(stderr, "    -F            do not become a daemon process\n");
	fprintf(stderr, "    -l            use a logfile instead of syslog\n");
	fprintf(stderr, "    -L <file>     use file instead of %s for logging\n", LOG_FILE_DEF);
	fprintf(stderr, "    -r <device>   redirect output to other device    (for -f)\n");
	fprintf(stderr, "    -s <facility> use facility instead of %d for syslog logging\n", LOG_LOCAL0 >> 3);
	fprintf(stderr, "    -t <termtype> terminal type of redirected screen (for -f)\n");
	fprintf(stderr, "    -u <time>     length of a charging unit in seconds\n");
	fprintf(stderr, "\n");
	exit(1);
}

/*---------------------------------------------------------------------------*
 *	curses menu for fullscreen command mode
 *---------------------------------------------------------------------------*/
static void
do_menu(void)
{

#define WMENU_LEN 	39
#define WMENU_HGT 	9
#define WMENU_TITLE 	"Command"
#define WMENU_POSLN	10
#define WMENU_POSCO	5
#define WMITEMS 	5
#define WMTIMEOUT	5

	static char *menu[WMITEMS] =
	{
		"1 - refresh Display",
		"2 - Rescan the configuration file",
		"3 - Quit, terminate program",
		"4 - hangup connection on channel B1",
		"5 - hangup connection on channel B2"		
	};

	WINDOW *menu_w;
	int c;
	int mpos;
	int ret;
	fd_set set;
	struct timeval timeout;
	struct confdata *cp;
	
	/* get current flag settings */

	ret = fcntl(0, F_GETFL, 0);
	ret &= ~O_ASYNC;

	/* switch off async i/o */
	
	if((fcntl(0, F_SETFL, ret)) < 0)
	{
		log(LL_SYS, "              ERROR, fcntl F_SETFL for stdin: %s", strerror(errno));
		return;
	}

	/* create a new window in the lower screen area */
	
	if((menu_w = newwin(WMENU_HGT, WMENU_LEN, WMENU_POSLN, WMENU_POSCO )) == NULL)
	{
		log(LL_SYS, "              ERROR, curses init menu window!");
		return;
	}

	/* create a border around the window */
	
	box(menu_w, '|', '-');

	/* add a title */
	
	wstandout(menu_w);
	mvwaddstr(menu_w, 0, (WMENU_LEN / 2) - (strlen(WMENU_TITLE) / 2), WMENU_TITLE);
	wstandend(menu_w);	

	/* fill the windo with the menu options */
	
	for(mpos=0; mpos <= (WMITEMS-1); mpos++)
		mvwaddstr(menu_w, mpos + 2, 2, menu[mpos]);

	/* highlight the first menu option */
	
	mpos = 0;
	wstandout(menu_w);
	mvwaddstr(menu_w, mpos + 2, 2, menu[mpos]);
	wstandend(menu_w);

	/* input loop */
	
	for(;;)
	{
		wrefresh(menu_w);

		FD_ZERO(&set);
		FD_SET(STDIN_FILENO, &set);
		timeout.tv_sec = WMTIMEOUT;
		timeout.tv_usec = 0;

		/* if no char is available within timeout, exit menu*/
		
		if((select(STDIN_FILENO + 1, &set, NULL, NULL, &timeout)) <= 0)
			goto mexit;
		
		c = wgetch(menu_w);

		switch(c)
		{
			case ' ':
			case '\t':	/* hilite next option */
				mvwaddstr(menu_w, mpos + 2, 2, menu[mpos]);
				mpos++;
				if(mpos >= WMITEMS)
					mpos = 0;
				wstandout(menu_w);
				mvwaddstr(menu_w, mpos + 2, 2, menu[mpos]);
				wstandend(menu_w);
				break;

			case '1':
			case 'D':
			case 'd':
				wrefresh(curscr);
				goto mexit;

			case '2':
			case 'R':
			case 'r':
				do_rescan();
				goto mexit;

			case '3':
			case 'Q':
			case 'q':
				do_quit(0);
				goto mexit;

			case '4':
				if((cp = getcpbycc(0, 0)) != NULL)
					hangup(isdnfd, cp);
				goto mexit;

			case '5':
				if((cp = getcpbycc(0, 1)) != NULL)
					hangup(isdnfd, cp);
				goto mexit;
				
			case '\n':
			case '\r':	/* exec highlighted option */
				switch(mpos)
				{
					case 0:
						wrefresh(curscr);
						break;
					case 1:
						do_rescan();
						break;
					case 2:
						do_quit(0);
						break;
					case 3:
						if((cp = getcpbycc(0, 0)) != NULL)
							hangup(isdnfd, cp);
						break;
					case 4:
						if((cp = getcpbycc(0, 0)) != NULL)
							hangup(isdnfd, cp);
						break;
				}
				goto mexit;
				break;
		
			default:
				goto mexit;
				break;
		}
	}

mexit:
	/* delete the menu window */

	delwin(menu_w);

	/* re-display the original lower window contents */
	
	touchwin(lower_w);
	wrefresh(lower_w);

	/* get current flags */
	
	ret = fcntl(0, F_GETFL, 0);
	ret |= O_ASYNC;

	/* re-activate async i/o */
	
	if((fcntl(0, F_SETFL, ret)) < 0)
		log(LL_SYS, "              ERROR, fcntl F_SETFL for stdin: %s", strerror(errno));
	
	return;
}

/* EOF */
