/*
 * Copyright (c) 1995, 1996 Hellmuth Michaelis. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *---------------------------------------------------------------------------*
 *
 *	bisdnd - BSD ISDN management daemon
 *      ===================================
 *
 *	Last Edit-Date: [Sat Jul  6 14:02:30 1996]
 *
 *	-hm	message processing implementation
 *	-hm	get_telno argument passing
 *	-hm	converting to message structs
 *	-hm	ISDN_HANGUP argument is now a structure
 *	-hm	local/remote disconnect indication
 *	-hm	fullscreen channel position fix
 *	-hm	avoid multiple dialouts when state != IDLE in msg_dial()
 *	-hm	adding support for NetBSD from Arne
 *	-hm	accounting format changed to include connect seconds
 *	-hm	resetting charge
 *	-hm	allow any connection to be logged in accounting file
 *
 *----------------------------------------------------------------------------*/

#include <string.h>
#include <errno.h>

#ifdef __NetBSD__
#include <sys/ioctl.h>
#endif /* __NetBSD__ */

#include <sys/time.h>
#include <sys/types.h>

#include "defs.h"

#include "pcause.h"

/*---------------------------------------------------------------------------*
 *	process accept message
 *---------------------------------------------------------------------------*/
void msg_accept(int isdnfd, unsigned char *rbuf)
{
	char *dst_telno;
	char *src_telno;
	int channel;				/* channel */
	int service;
	int alert_id;
	int appl_no;
	int appl_typ;
	accept_t *accept_p;
	struct confdata *confdata_p;
	int ret;
	message_accept_t *mp = (message_accept_t *)rbuf;
	
	appl_typ = mp->header.appl_type;
	appl_no = mp->header.appl_no;
	channel = mp->header.channel;
	service = mp->header.service;
	
	alert_id = mp->alert_id;
	dst_telno = mp->dst_telno;
	src_telno = mp->src_telno;

	if((ret = find_appl(&accept_p, channel, service, alert_id,
			dst_telno, src_telno)) != FA_GOOD)
	{
		if(ioctl(isdnfd, ISDN_DONTACCEPT, &channel) < 0)
			log(LL_SYS, "              ioctl ISDN_DONTACCEPT failed: %s", strerror(errno));

		switch(ret)
		{
			case FA_ERROR:
				log(LL_NAC, "T%d N%d C%d/%d S%d unknown find_appl() error (dst:%s src:%s)",
					appl_typ, appl_no, channel>>1, channel&1, service, dst_telno, src_telno);
				break;
#if DEBUG
			case FA_NOTME:
				if(do_debug & DEBUGL_NORM)
					log(LL_DBG, "T- N- C%d/%d S%d not my number (dst:%s src:%s)",
						channel>>1, channel&1, service, dst_telno, src_telno);
				break;
#endif
			case FA_CARD:
				log(LL_NAC, "T%d N%d C%d/%d S%d invalid controller no (dst:%s src:%s)",
					appl_typ, appl_no, channel>>1, channel&1, service, dst_telno, src_telno);
				break;
			case FA_ATYP:
				log(LL_NAC, "T%d N%d C%d/%d S%d invalid application type (dst:%s src:%s)",
					appl_typ, appl_no, channel>>1, channel&1, service, dst_telno, src_telno);
				break;
			default:
				if(ret & FA_INV_DEST)
					log(LL_NAC, "T%d N%d C%d/%d S%d unknown caller telno (dst:%s src:%s)",
						appl_typ, appl_no, channel>>1, channel&1, service, dst_telno, src_telno);
				if(ret & FA_INV_SERV)
					log(LL_NAC, "T%d N%d C%d/%d S%d unknown service (dst:%s src:%s)",
						appl_typ, appl_no, channel>>1, channel&1, service, dst_telno, src_telno);
				if(ret & FA_INV_CTRL)
					log(LL_NAC, "T%d N%d C%d/%d S%d invalid controller (dst:%s src:%s)",
						appl_typ, appl_no, channel>>1, channel&1, service, dst_telno, src_telno);
				if(!(   (ret & FA_INV_DEST) ||
					(ret & FA_INV_SERV) ||
					(ret & FA_INV_CTRL) ))
				{
					log(LL_NAC, "T%d N%d C%d/%d S%d UNKNOWN ERROR NO % from find_appl (dst:%s src:%s)",
						appl_typ, appl_no, channel>>1, channel&1, service, ret, dst_telno, src_telno);
				}
		}
		return;
	}

	confdata_p = getcp(accept_p->atyp, accept_p->appl);

	if(confdata_p->reaction == REACT_CALLBACK)
	{
		dial_t *dial_p;

		confdata_p->charge = 0;	

		if((get_telno(&dial_p, accept_p->atyp, accept_p->appl)) == GOOD)
		{
			/* hangup & callback */

			hangup_t hangup;

			hangup.appl_type = accept_p->atyp;
			hangup.appl_no = accept_p->appl;
			hangup.channel = channel;
			hangup.cause = 0;

			if(ioctl(isdnfd, ISDN_HANGUP, &hangup) < 0)
				log(LL_SYS, "              ioctl ISDN_HANGUP failed: %s", strerror(errno));

			log(LL_INF, "T%d N%d C%d/%d S%d incoming callback (dst:%s src:%s)",
				accept_p->atyp, accept_p->appl, channel>>1, channel&1, service,
				dst_telno, src_telno);

			/* state = wait for disconnect before calling back */

#if DEBUG
			if(do_debug & DEBUGL_STATE)
				log(LL_DBG, "              STATE change [msg_accept] %s --> CB_WAITDISC", statetab[confdata_p->state]);
#endif
		
			confdata_p->state = CB_WAITDISC;

			confdata_p->curtry = 0;

			/* callback done in handle_recovery() */
		}
		return;
	}

	accept_p->ctrl = channel;
	appl_no = accept_p->appl;
	appl_typ = accept_p->atyp;
	
	if(ioctl(isdnfd, ISDN_ACCEPT, accept_p) < 0)
	{
		log(LL_SYS, "              ioctl ISDN_ACCEPT failed: %s", strerror(errno));
		return;
	}

	log(LL_ACP, "T%d N%d C%d/%d S%d incoming call (dst:%s src:%s)",
		appl_typ, appl_no, channel>>1, channel&1 ,service,
		dst_telno, src_telno);		
}

/*---------------------------------------------------------------------------*
 *	process dial message
 *---------------------------------------------------------------------------*/
void msg_dial(int isdnfd, unsigned char *rbuf)
{
	int appl_no;
	int appl_typ;
	dial_t *dial_p;
	struct confdata *confdata_p;
	int dodial = 0;
	message_dial_t *mp = (message_dial_t *)rbuf;
	
	appl_typ = mp->header.appl_type;
	appl_no = mp->header.appl_no;

	if((get_telno(&dial_p, appl_typ, appl_no)) == ERROR)
	{
		hangup_t hangup;

		hangup.appl_type = appl_typ;
		hangup.appl_no = appl_no;
		hangup.channel = 0xff;
		hangup.cause = 0;

		if(ioctl(isdnfd, ISDN_HANGUP, &hangup) < 0)
			log(LL_SYS, "              ioctl ISDN_HANGUP failed: %s", strerror(errno));

		log(LL_NDL, "T%d N%d C-/- S%d dial get_telno failed",
			appl_typ, appl_no, mp->header.service & SERVICEMASK);
		return;
	}

	confdata_p = getcp(appl_typ, appl_no);

	/* already active, dialing etc .. ? */
	
	if(confdata_p->state != IDLE)
	{
#if DEBUG
		log(LL_DBG, "T%d N%d C%d/- S%d dial req while !IDLE %s",
			appl_typ, appl_no,
			confdata_p->controller,
			mp->header.service & SERVICEMASK,
			dial_p->dst_telno);
#endif
		return;
	}
	
	dial_p->ctrl = confdata_p->controller;
	confdata_p->charge = 0;
	confdata_p->connect_time = 0;	
	
	if(confdata_p->reaction == REACT_CALLEDBACK)	/* i will be called back */
	{
		dodial = 0;

		if(confdata_p->state == IDLE)
		{
			confdata_p->lasttime = time(NULL);
			confdata_p->curtry = 0;
			dodial = 1;

			log(LL_INF, "T%d N%d C%d/- S%d callback dial %s",
				appl_typ, appl_no,
				confdata_p->controller,
				mp->header.service & SERVICEMASK,
				dial_p->dst_telno);
#if DEBUG
			if(do_debug & DEBUGL_STATE)
				log(LL_DBG, "              STATE change [msg_dial] %s --> AWCB_WAITDISC", statetab[confdata_p->state]);
#endif

			confdata_p->state = AWCB_WAITDISC;
		}
		else
		{
			if(time(NULL) > (confdata_p->lasttime + confdata_p->callbackwait))
			{
				log(LL_INF, "T%d N%d C%d/- S%d callback dial retry %s",
					appl_typ, appl_no, 
					confdata_p->controller,
					mp->header.service & SERVICEMASK,
					dial_p->dst_telno);
				confdata_p->lasttime = time(NULL);
				dodial = 1;
			}
			else
			{
				log(LL_INF, "T%d N%d C%d/- S%s callback retry time not reached %s",
					appl_typ, appl_no,
					confdata_p->controller,
					mp->header.service & SERVICEMASK,
					dial_p->dst_telno);
			}
		}
	}
	else
	{
		if(confdata_p->state == IDLE)
		{
			confdata_p->curtry = 0;
			confdata_p->lasttime = time(NULL);
#if DEBUG
			if(do_debug & DEBUGL_STATE)
				log(LL_DBG, "              STATE change [msg_dial] %s --> DIALING", statetab[confdata_p->state]);
#endif
			confdata_p->state = DIALING;
			dodial = 1;
		}
		else
		{
			if(time(NULL) > (confdata_p->lasttime + confdata_p->recoverytime))
			{
				log(LL_INF, "T%d N%d C%d/- S%d dialing %s",
					appl_typ, appl_no,
					confdata_p->controller,
					mp->header.service & SERVICEMASK,
					dial_p->dst_telno);
				confdata_p->lasttime = time(NULL);
				dodial = 1;
			}
			else
			{
				log(LL_INF, "T%d N%d C%d/- S%d dial retry time not reached %s",
					appl_typ, appl_no,
					confdata_p->controller,
					mp->header.service & SERVICEMASK,
					dial_p->dst_telno);
			}
		}
	}
	
	if(dodial)
	{
		confdata_p->curtry++;

		log(LL_DIA, "T%d N%d C%d/- S%d dialed %s",
			appl_typ, appl_no,
			confdata_p->controller,
			mp->header.service & SERVICEMASK,
			dial_p->dst_telno);
		
		if (ioctl(isdnfd, ISDN_DIAL, dial_p) < 0)
		{
			log(LL_SYS, "              ioctl ISDN_DIAL failed: %s", strerror(errno));
			return;
		}
	}
}

/*---------------------------------------------------------------------------*
 *	process info message
 *---------------------------------------------------------------------------*/
void msg_info(int isdnfd, unsigned char *rbuf)
{
	int appl_no;
	int appl_typ;
	int channel;
	int service;
	int prot;
	message_info_t *mp = (message_info_t *)rbuf;
	
	appl_typ = mp->header.appl_type;
	appl_no = mp->header.appl_no;
	channel = mp->header.channel;
	service = mp->header.service;

	prot = mp->protocol;

	if((appl_no >= 0x7f) || (channel >= 0x7f))
	{
#if DEBUG > 1
		if(appl_no >= 0x7f)
		{
			if(do_debug & DEBUGL_NORM)
				log(LL_DBG, "              ERROR, msg_info got appl_no=0x%x, appl_typ=%d", appl_no, appl_typ);
		}
		if(channel >= 0x7f)
		{
			if(do_debug & DEBUGL_NORM)
				log(LL_DBG, "              ERROR, msg_info got channel=0x%x, appl_typ=%d", channel, appl_typ);
		}
#endif		
		return;
	}
		
	switch(prot)
	{
		case ISDNDSS1:
			info_dss1(isdnfd, rbuf);
			break;
		case ISDN1TR6:
			info_1tr6(isdnfd, rbuf);
			break;
		default:
			log(LL_INF, "T%d N%d C%d/%d S%d unknown protocol=%d, %s",
				appl_typ, appl_no, channel>>1, channel&1, service, prot, mp->info);
			break;
	}
}

/*---------------------------------------------------------------------------*
 *	process ip accounting message
 *---------------------------------------------------------------------------*/
void msg_ipacct(int isdnfd, unsigned char *rbuf)
{
	int appl_no;
	int appl_typ;
	int service;
	int channel;
	message_ipacct_t *mp = (message_ipacct_t *)rbuf;
	
	appl_typ = mp->header.appl_type;
	appl_no = mp->header.appl_no;
	channel = mp->header.channel;
	service = mp->header.service;

	if(mp->flag == IPIACCT_FIN)
 	{
		log(LL_ACT, "T%d N%d C%d/%d S%d IP-i/o b=%d/%d c=%d/%d p=%d/%d",
			appl_typ, appl_no, channel>>1, channel&1, service,
			mp->input_bytes,
			mp->output_bytes,
			mp->compr_input_bytes,
			mp->compr_output_bytes,
			mp->input_packets,
			mp->output_packets);
		if(useacctfile)
		{
			struct confdata *confdata_p;
			confdata_p = getcp(appl_typ, appl_no);
			confdata_p->inbytes = mp->input_bytes;
			confdata_p->outbytes = mp->output_bytes;
		}
	}
	else if(do_fullscreen && (mp->flag == IPIACCT_MID))
	{
		mvwprintw(upper_w, ((channel>>1)+(channel&1)), H_OUT, "%-8d", mp->output_bytes);
		mvwprintw(upper_w, ((channel>>1)+(channel&1)), H_IN, "%-8d", mp->input_bytes);
		wrefresh(upper_w);
	}
}

/*---------------------------------------------------------------------------*
 *	process connect message
 *---------------------------------------------------------------------------*/
void msg_connect(int isdnfd, unsigned char *rbuf)
{
	accept_t *accept_p;
	struct confdata *confdata_p;
	int active;
	int ret;
	int appl_no;
	int appl_typ;
	int service;
	int channel;
	message_connect_t *mp = (message_connect_t *)rbuf;
	
	appl_typ = mp->header.appl_type;
	appl_no = mp->header.appl_no;
	channel = mp->header.channel;
	service = mp->header.service;

	active = mp->active_flag;

	confdata_p = getcp(appl_typ, appl_no);
	
	ret = connected(&accept_p, isdnfd, appl_typ, appl_no, channel, active);

	if(active)	/* active = 0 -> got CONNECT_ACKNOWLEDGE	*/
	{		/* active = 1 -> got CONNECT			*/

		if(ret == ERROR)
		{
			hangup_t hangup;

			hangup.appl_type = appl_typ;
			hangup.appl_no = appl_no;
			hangup.channel = channel;
			hangup.cause = 0;

			if(ioctl(isdnfd, ISDN_HANGUP, &hangup) < 0)
				log(LL_SYS, "              ioctl ISDN_HANGUP failed: %s", strerror(errno));

			log(LL_NCN, "T%d N%d C%d/%d S%d connect failed",
				appl_typ, appl_no, channel>>1, channel&1, service);
#if DEBUG
			if(do_debug & DEBUGL_STATE)
				log(LL_DBG, "              STATE change [msg_dial] %s --> IDLE (ERROR1)", statetab[confdata_p->state]);
#endif
			confdata_p->state = IDLE;
			return;
		}

		accept_p->ctrl = channel;

		if (ioctl(isdnfd, ISDN_ACCEPT, accept_p) < 0)
		{
			log(LL_SYS, "              ioctl ISDN_ACCEPT failed: %s", strerror(errno));
#if DEBUG
			if(do_debug & DEBUGL_STATE)
				log(LL_DBG, "              STATE change [msg_dial] %s --> IDLE (ERROR2)", statetab[confdata_p->state]);
#endif
			confdata_p->state = IDLE;
			return;
		}
	}

#if DEBUG
	if(ret == ERROR)
		log(LL_DBG, "              ERROR, connected() returned ERROR, active = 0!");
#endif

	confdata_p->aoc_now = time(NULL);
	confdata_p->aoc_last = 0;
	confdata_p->aoc_diff = 0;
	confdata_p->aoc_lastdiff = -1;
	confdata_p->aoc_valid = AOC_INVALID;
	confdata_p->charge = 0;
	confdata_p->connect_time = confdata_p->aoc_now;
	confdata_p->channel = channel & 1;
	confdata_p->inbytes = INVALID;
	confdata_p->outbytes = INVALID;	

	log(LL_CON, "T%d N%d C%d/%d S%d connected [%s]",
		appl_typ, appl_no, channel>>1, channel&1, service, confdata_p->name);

	if(do_fullscreen)
	{
		char buffer[64];
		int a = appl_typ;

		/* get application type */

		if(a == ATYP_TEL)
		{
			struct answ *r = &answ[appl_no];
			a += r->conf->telaction;
		}

		/* get a remote telephone number */

		if(confdata_p->telnrem_rdi[0] == '*')
			sprintf(buffer, "%s/%s", confdata_p->name, confdata_p->telno_remote);
		else
			sprintf(buffer, "%s/%s", confdata_p->name, confdata_p->telnrem_rdi);
		buffer[H_SERV - H_TELN - 1] = '\0';

		mvwaddstr(upper_w, ((channel>>1)+(channel&1)), H_TELN, buffer);
		mvwaddstr(upper_w, ((channel>>1)+(channel&1)), H_SERV, serv_tab[(confdata_p->service) & SERVICEMASK]);
		mvwaddstr(upper_w, ((channel>>1)+(channel&1)), H_APPL, appl_tab[a]);

		if(confdata_p->state == DIALING)
			mvwaddstr(upper_w, ((channel>>1)+(channel&1)), H_IO, "Out");
		else
			mvwaddstr(upper_w, ((channel>>1)+(channel&1)), H_IO, "In");

		mvwaddstr(upper_w, ((channel>>1)+(channel&1)), H_UNITS, "-");
		wrefresh(upper_w);
	}
#if DEBUG
	if(do_debug & DEBUGL_STATE)
		log(LL_DBG, "              STATE change [msg_connect] %s --> CONNECTED", statetab[confdata_p->state]);
#endif
	confdata_p->state = CONNECTED;
}

/*---------------------------------------------------------------------------*
 *	process disconnect message
 *---------------------------------------------------------------------------*/
void msg_disconnect(int isdnfd, unsigned char *rbuf)
{
	int protocol;
	int cause;
	struct confdata *p;

	int appl_no;
	int appl_typ;
	int service;
	int channel;
	message_disconnect_t *mp = (message_disconnect_t *)rbuf;
	
	appl_typ = mp->header.appl_type;
	appl_no = mp->header.appl_no;
	channel = mp->header.channel;
	service = mp->header.service;

	protocol = mp->protocol;
	cause = mp->cause;

	disconnected(isdnfd, appl_typ, appl_no);

	if(cause == 0 && protocol == ISDNUNKN)	/* this is a hack XXX */
		protocol = ISDNDSS1;

	if(channel < 0x7f)	/* normal disconnect */
	{
		log(LL_DIS, "T%d N%d C%d/%d S%d disconnect (%s), cause=%s",
			appl_typ, appl_no, channel>>1, channel&1, service,
			mp->localflag == DISC_LOCAL ? "local" : "remote",
			print_error(protocol, cause));
	}
	else			/* calledback disconnect */
	{
		log(LL_DIS, "T%d N%d C-/- S%d disconnect (%s), cause=%s",
			appl_typ, appl_no, service,
			mp->localflag == DISC_LOCAL ? "local" : "remote",
			print_error(protocol, cause));
	}		

	p = getcp(appl_typ, appl_no);

	switch(p->state)
	{
		case AWCB_WAITDISC:		/* trigger callback ok,	    */
#if DEBUG
		if(do_debug & DEBUGL_STATE)
			log(LL_DBG, "              STATE change [msg_disconnect] %s --> AWAITCALLBACK", statetab[p->state]);
#endif
			p->state = AWAITCALLBK; /* now wait for callback    */
			break;

		case CB_WAITDISC:
#if DEBUG
		if(do_debug & DEBUGL_STATE)
			log(LL_DBG, "              STATE change [msg_disconnect] %s --> DIALING", statetab[p->state]);
#endif
			p->state = DIALING;	/* did wait for disconnect  */
			break;			/* got it, now dial	    */
	
		default:			/* anything else ...	    */
#if DEBUG
		if(do_debug & DEBUGL_STATE)
			log(LL_DBG, "              STATE change [msg_disconnect] %s --> IDLE", statetab[p->state]);
#endif
			p->state = IDLE;
			break;
	}

	if(p->charge || acct_all)
	{
		if(p->charge)
		{
			log(LL_CRG, "T%d N%d C%d/%d S%d charging, units = %d",
				appl_typ, appl_no, channel>>1, channel&1,
				service, p->charge);
		}

		if(useacctfile && (p->connect_time != 0))
		{
			int con_secs;
			char logdatetime[41];
			struct tm *tp;

			con_secs = difftime(time(NULL), p->connect_time);
			
			tp = localtime(&p->connect_time);
			strftime(logdatetime,40,"%d.%m.%y %H:%M:%S",tp);

			if(p->inbytes != INVALID && p->outbytes != INVALID)
			{
				fprintf(acctfp, "%s - %s %s %d (%d) (%d/%d)\n",
					logdatetime, getlogdatetime(),
					p->name, p->charge, con_secs,
					p->inbytes, p->outbytes);
			}
			else
			{
				fprintf(acctfp, "%s - %s %s %d (%d)\n",
					logdatetime, getlogdatetime(),
					p->name, p->charge, con_secs);
			}
		}
		p->charge = 0;
	}
	
	if(do_fullscreen)
	{
		wmove(upper_w, ((channel>>1)+(channel&1)), H_TELN);
		wclrtoeol(upper_w);
		wrefresh(upper_w);
	}
}

/* EOF */
