/*
 * Copyright (c) 1995, 1996 Hellmuth Michaelis. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *---------------------------------------------------------------------------*
 *
 *	bisdnd - BSD ISDN management daemon
 *      ===================================
 *
 *	Last Edit-Date: [Fri Jul  5 12:06:32 1996]
 *
 *	-hm	message processing utility routines
 *	-hm	isdn_number_t, no subaddresses anymore
 *	-hm	phone numbers as strings
 *	-hm	debugging phone number handling
 *	-hm	error handling
 *	-hm	cleanup, remove more static's
 *	-hm	removed getno
 *	-hm	converting to message structs
 *	-hm	earlyhangup made configurable
 *	-hm	fixing fullscreen channel position
 *	-hm	adding support for NetBSD from Arne
 *	-hm	interactive hangup support routines
 *	-hm	improving retry & recovery
 *
 *----------------------------------------------------------------------------*/

#include <string.h>
#include <errno.h>

#ifdef __NetBSD__
#include <sys/ioctl.h>
#endif /* __NetBSD__ */

#include <sys/time.h>

#include <machine/bisdn_isdn.h>

#include "defs.h"

static void handle_charge(int isdnfd, struct confdata *cp, int units,
		time_t now, int an, int channel);
static int decode_q932_aoc(unsigned char *);

/*---------------------------------------------------------------------------*
 *	return pointer to configuration entry
 *---------------------------------------------------------------------------*/
struct confdata *getcp(int atyp, int appl)
{
	register int i;

	for(i=0; i <= config_entries; i++)
	{		
		if((conf[i].atyp == atyp) && (conf[i].appl == appl))
			return(&conf[i]);
	}
	log(LL_SYS, "getcp cannot find config entry for typ=%d, appl=%d", atyp, appl);
	do_quit(99);
	return(NULL);
}

/*---------------------------------------------------------------------------*
 *	return pointer to configuration entry by controller & channel
 *---------------------------------------------------------------------------*/
struct confdata *getcpbycc(int controller, int channel)
{
	register int i;

	for(i=0; i <= config_entries; i++)
	{		
		if((conf[i].controller == controller)   &&
		   (conf[i].channel == channel)		&&
		   (conf[i].state == CONNECTED))
		{
			return(&conf[i]);
		}
	}
	return(NULL);
}

/*---------------------------------------------------------------------------*
 *	interactive hangup a connection (called from do_menu())
 *---------------------------------------------------------------------------*/
void hangup(int isdnfd, struct confdata *cp)
{
	hangup_t hangup;

	hangup.appl_type = cp->atyp;
	hangup.appl_no = cp->appl;
	hangup.channel = cp->channel;
	hangup.cause = 0;

	if(ioctl(isdnfd, ISDN_HANGUP, &hangup) < 0)
		log(LL_SYS, "              ioctl ISDN_HANGUP failed: %s", strerror(errno));

	log(LL_INF, "T%d N%d C%d/%d S%d local user initiated disconnect",
		cp->atyp, cp->appl, cp->controller, cp->channel,
		cp->service & SERVICEMASK);
}

/*---------------------------------------------------------------------------*
 *	process the charging units
 *---------------------------------------------------------------------------*/
static void handle_charge(int isdnfd, struct confdata *cp, int units,
	time_t now, int an, int channel)
{
	cp->charge = units;

	if(cp->aoc_last == 0)
	{
		cp->aoc_last = now;
	}
	else if(cp->aoc_now == 0)
	{
		cp->aoc_now = now;
	}
	else
	{
		cp->aoc_last = cp->aoc_now;
		cp->aoc_now = now;
		cp->aoc_diff = cp->aoc_now - cp->aoc_last;
		cp->aoc_valid = AOC_VALID;
	}
	
	if(do_fullscreen)
	{
		mvwprintw(upper_w, ((channel>>1)+(channel&1)), H_UNITS, "%d", units);
		wrefresh(upper_w);
	}

	if((cp->unit_length_typ == ULTYP_DYN) && (cp->aoc_valid == AOC_VALID))
	{
		if(cp->aoc_diff != cp->aoc_lastdiff)
		{
			timo_t newto;

			newto.timeout = (cp->aoc_diff) - (cp->earlyhangup);
			newto.appl = an;
#if DEBUG
			if(do_debug && cp->aoc_valid)
				log(LL_DBG, "              charging unit length = %d seconds", cp->aoc_diff);
#endif				
			cp->aoc_lastdiff = cp->aoc_diff;

			if((ioctl(isdnfd, ISDN_STIMEOUT, &newto)) < 0)
				log(LL_SYS, "              ioctl ISDN_STIMEOUT failed: %s", strerror(errno));
		}
	}
#if DEBUG
	else
	{
		if(do_debug && cp->aoc_valid)
			log(LL_DBG, "              charging unit length = %d seconds", cp->aoc_diff);
	}		
#endif
}

/*---------------------------------------------------------------------------*
 *	display INFO messages for the Q.931/DSS1 protocol
 *---------------------------------------------------------------------------*/
void info_dss1(int isdnfd, u_char *s)
{
	int appl_no;
	int appl_typ;
	int channel;
	int service;
	time_t time_now;
	struct confdata *cp;
	char logbuf[1024];
	int i, j;
	message_info_t *mp = (message_info_t *)s;
	unsigned char *ip;
	
	appl_typ = mp->header.appl_type;
	appl_no = mp->header.appl_no;
	channel = mp->header.channel;
	service = mp->header.service;

	ip = &mp->info[0];
	
	switch(*ip)
	{
		case IEI_DATETIME:	/* date/time */
			for(i = 2, j = 0; i <= *(ip+1)+1; i++, j += 3)
				sprintf(&logbuf[j], " %02d", ip[i]);

			log(LL_INF, "T%d N%d C%d/%d S%d date/time:%s",
				appl_typ, appl_no,
				channel>>1, channel&1,
				service, logbuf);
			break;
			
		case IEI_FACILITY:	/* facility */
			time_now = time(NULL);

			cp = getcp(appl_typ, appl_no);

			i = decode_q932_aoc(s);

			if(i != -1)
				handle_charge(isdnfd, cp, i, time_now, appl_no, channel);
			break;
			
		default:
			for(i = 0, j = 0; i < (mp->length-1); i++, j += 5)
				sprintf(&logbuf[j], " 0x%02x", mp->info[i]);
			log(LL_INF, "T%d N%d C%d/%d S%d unknown Q.931 info:%s",
				appl_typ, appl_no, channel>>1, channel&1, service, logbuf);
			break;
	}
}

/*---------------------------------------------------------------------------*
 *	display INFO messages for the 1TR6 protocol
 *---------------------------------------------------------------------------*/
void info_1tr6(int isdnfd, u_char *s)
{
	int i, j, k, sum;
	time_t time_now;
	struct confdata *cp;
	char logbuf[1024];
	char *p;
	int appl_no;
	int appl_typ;
	int channel;
	int service;
	message_info_t *mp = (message_info_t *)s;
	unsigned char *ip;
	
	appl_typ = mp->header.appl_type;
	appl_no = mp->header.appl_no;
	channel = mp->header.channel;
	service = mp->header.service;

	ip = &mp->info[0];
	
	switch(*ip)
	{
		case IEI_DATE:		/* date/time */
			p = logbuf;				

			for(i = 2; i <= *(ip+1)+1; i++)
				*p++ = ip[i];

			*p = '\0';

			log(LL_INF, "T%d N%d C%d/%d S%d date/time: %s",
				appl_typ, appl_no, channel>>1, channel&1, service, logbuf);
			break;
			
		case IEI_CHARGINFO:	/* charging information */
			time_now = time(NULL);

			cp = getcp(appl_typ, appl_no);

			sum = 0;
			k = 1;

			for(i = 3, j = *(ip+1) - 1; j > 0; i++, j--)
			{
				sum += (ip[j+2] - '0') * k;
				k = k * 10;
			}

#if DEBUG > 1
			if(do_debug & DEBUGL_NORM)
			{
				for(i = 0, j = 0; i < (mp->length-1); i++, j += 5)
					sprintf(&logbuf[j], " 0x%02x", mp->info[i]);
				log(LL_DBG, "T%d N%d C%d/%d S%d sum=%d - 1TR6 charge:%s",
					appl_typ, appl_no, channel>>1, channel&1, service, sum, logbuf);
			}
#endif

			handle_charge(isdnfd, cp, sum, time_now, appl_no, channel);
			break;
			
		default:
			for(i = 0, j = 0; i < (mp->length-1); i++, j += 5)
				sprintf(&logbuf[j], " 0x%02x", mp->info[i]);
			log(LL_INF, "T%d N%d C%d/%d S%d unknown 1TR6 info:%s",
				appl_typ, appl_no, channel>>1, channel&1, service, logbuf);
			break;
	}
}

/*---------------------------------------------------------------------------*
 *
 *	decode AOC facility messages
 *
 *	This was taken from the Linux ISDN package written by
 *	Matthias Urlichs. I can imagine what it does, but i don't
 *	understand it; anyway it seems to produce the right output.
 *
 *	Whoever understands the coding of parameters for facility
 *	messages, please tell me where i can find Q.956 or another
 *	explanation of things! (hm@hcs.de)
 *
 *	I tried to comment things that i might understand, please
 *	correct me if i'm wrong !!!
 *
 *---------------------------------------------------------------------------*/
static int decode_q932_aoc(unsigned char *s)
{
	int appl_no;
	int appl_typ;
	int channel;
	int service;
	int msg_len;
	int ret = -1;
	message_info_t *mp = (message_info_t *)s;
	unsigned char *msg_data;
	
	appl_typ = mp->header.appl_type;
	appl_no = mp->header.appl_no;
	channel = mp->header.channel;
	service = mp->header.service;

	msg_data = &mp->info[1];

	msg_len = *msg_data++;		

	/* check overall message length */
	
	if (msg_len == 0)
		return(ret);

	/* check facility ie protocol profile (Q.932 8.2.3) */
	
	if((*msg_data & 0x1F) != 0x11)	/* must be Remote Operation Protocol */
		return(ret);

	/* skip (whatever more?) data */
	
	while(msg_len > 0 && !(*msg_data & 0x80))
	{
		msg_data++;
		msg_len--;
	} 

	/* check length */
	
	if(msg_len < 2)
		return(ret);

	msg_data++;
	msg_len--;

	/* data == component type tag ?? (Q.932 IV.2.2) */
	
	if((*msg_data & 0xE0) != 0xA0)
		return(ret);

	/* data = invoke componet type tag (Q.932 IV.2.2) ? */
	
	if((*msg_data & 0x1F) == 1)
	{
		unsigned char nlen, ilen;
		int ident;

		msg_data++;
		msg_len--;

		/* sub-overall length check */
		
		if(msg_len < 1)
			return(ret);

		if(*msg_data & 0x80)	/* length format */
			return(ret);

		nlen = *msg_data++;
		msg_len--;

		if(msg_len < nlen)
			return(ret);

		msg_len -= nlen;

		if(nlen < 2)
			return(ret);

		/* 1st 0x02 tag (Q.932 III.2) */

		if(*msg_data != 0x02)
			return(ret);

		msg_data++; nlen--;

		if(*msg_data & 0x80)	/* length format */
			return(ret);

		/* 1st 0x02 tag length */
		
		ilen = *msg_data++;
		nlen--;

		if(ilen > nlen || ilen == 0)
			return(ret);

		nlen -= ilen;

		/* 1st 0x02 tag id */
		
		ident = 0;

		while(ilen > 0)
		{
			ident = (ident << 8) | (*msg_data++ & 0xFF);
			ilen--;
		}

		if(nlen < 2)
			return(ret);

		/* 2nd 0x02 tag */

		if(*msg_data != 0x02)
			return(ret);

		/* 2nd 0x02 tag length */

		msg_data++;
		nlen--;

		ilen = *msg_data++;
		nlen--;
		
		if(ilen > nlen || ilen == 0)
			return(ret);

		nlen -= ilen;

		/* 2nd 0x02 tag id */
		/* this seems to be the code for:	*/
		/* AOCD (currency) = 33 = 0x21		*/
		/* AOCD (units)    = 34 = 0x22		*/
		/* AOCE (currency) = 35 = 0x23		*/
		/* AOCD (units)    = 36 = 0x24		*/
		
		ident = 0;

		while(ilen > 0)
		{
			ident = (ident << 8) | (*msg_data++ & 0xFF);
			ilen--;
		}

		/* this is Matthias' personal secret ... ;-) */

/*-----------------------------------------------------------------------*/
/* CAUTION: do _NOT_ remove the line "int nlen = ilen" below because the */
/* compiler says its unused - the code does _NOT_ work any longer if you */
/* remove it !!!!!!!                                                     */
/*-----------------------------------------------------------------------*/

#define FOO1(s,a,b) 					\
		while(nlen > 1) {			\
			int ilen = msg_data[1];		\
			if(nlen < ilen+2) {		\
				return(ret);		\
			}				\
			nlen -= ilen+2;			\
			if((*msg_data & 0xFF) == (a)) {	\
				int nlen = ilen;	\
				msg_data += 2;		\
				b;			\
			} else {			\
				msg_data += ilen+2;	\
			}				\
		}
		


		switch(ident) {

		case 0x22: /* AOCDuring the call in charging units */

FOO1("1A",0x30,FOO1("1C",0xA1,FOO1("1D",0x30,FOO1("1E",0x02,({

				ident = 0;

				while(ilen > 0) {
					ident = (ident<<8) | *msg_data++;
					ilen--;
				}
				log(LL_INF, "T%d N%d C%d/%d S%d AOCD, units = %d",
					appl_typ, appl_no,
					channel>>1, channel&1,
					service, ident);

				ret = ident;
				})))))
			break;

		case 0x24: /* AOCEnd of call in charging units */

FOO1("2A",0x30,FOO1("2B",0x30,FOO1("2C",0xA1,FOO1("2D",0x30,FOO1("2E",0x02,({

				ident = 0;

				while(ilen > 0) {
					ident = (ident<<8) | *msg_data++;
					ilen--;
				}

				log(LL_INF, "T%d N%d C%d/%d S%d AOCE, units = %d",
					appl_typ, appl_no,
					channel>>1, channel&1,
					service, ident);

				ret = ident;
				}))))))
			break;
		default:
			break;
		}
	}
	return(ret);
}

/*---------------------------------------------------------------------------*
 *	timeout and recovery handling
 *---------------------------------------------------------------------------*/
void handle_recovery(int isdnfd)
{
	static int tv_first = 1;
	static struct timeval tv_last;
	struct timeval tv_now;
	struct confdata *p;
	int i;
	int dodial;
	
	/* there must be 1 sec minimum between calls to this section */

	if(tv_first)
	{
		gettimeofday(&tv_last, NULL);
		tv_first = 0;
	}

	gettimeofday(&tv_now, NULL);

	if((tv_now.tv_sec - tv_last.tv_sec) < 1)
	{

#if DEBUG
		if(do_debug & DEBUGL_TIME)
		{
			log(LL_DBG, "              time < 1 - last %ld:%ld now %ld:%ld",
				tv_last.tv_sec, tv_last.tv_usec,
				tv_now.tv_sec, tv_now.tv_usec);
		}
#endif				

		return;
	}
	else if((tv_now.tv_sec - tv_last.tv_sec) == 1)
	{
		if(((1000000 - tv_last.tv_usec) + tv_now.tv_usec) < 900000)
		{

#if DEBUG
			if(do_debug & DEBUGL_TIME)
			{
				log(LL_DBG, "              time < 900000us - last %ld:%ld now %ld:%ld",
					tv_last.tv_sec, tv_last.tv_usec,
					tv_now.tv_sec, tv_now.tv_usec);
			}
#endif

			return;
		}
	}

#if DEBUG
	if(do_debug & DEBUGL_TIME)
	{
		log(LL_DBG, "              time OK! - last %ld:%ld now %ld:%ld",
			tv_last.tv_sec, tv_last.tv_usec,
			tv_now.tv_sec, tv_now.tv_usec);
	}
#endif

	gettimeofday(&tv_last, NULL);
	 
	/* timeout and retry handling */		

	for(i=0; i <= config_entries; i++)
	{		
		dodial = 0;
		p = &conf[i];

		switch(p->state)
		{
			
			case IDLE:	/* connection is idle		*/
			case CONNECTED:	/* connected with remote	*/
				continue;
				break;

		/* dialing out to remote site */				
			case DIALING:
				if(time(NULL) > (p->lasttime + p->recoverytime))
					dodial = 1;
				break;
				
		/* dialing, await disconnect and callback */
			case AWCB_WAITDISC:
				if(time(NULL) > (p->lasttime + p->callbackwait))
					dodial = 1;
				break;

		/* waiting for callback from remote */
			case AWAITCALLBK:
				if(time(NULL) > (p->lasttime + p->calledbackwait))
					dodial = 1;
				break;

		/* got call, wait for disconnect before calling back remote */
			case CB_WAITDISC:
				continue;
				break;
		}

		if(p->curtry >= p->dialretries)
		{
			log(LL_NDL, "T%d N%d C%d/- S%d dial failed after %d retries",
				p->atyp, p->appl,  p->controller,
				p->service & SERVICEMASK, p->curtry);
#if DEBUG
			if(do_debug & DEBUGL_STATE)
				log(LL_DBG, "              STATE change [handle_recovery] %s --> IDLE", statetab[p->state]);
#endif
			p->state = IDLE;
			continue;
		}

		if(dodial)
		{
			dial_t *dial_p;

			p->curtry++;

			if((get_telno(&dial_p, p->atyp, p->appl)) == ERROR)
			{
				log(LL_NDL, "T%d N%d C%d/- S%d %s%s get_telno returned ERROR",
						p->atyp, p->appl,
						p->controller,
						p->service & SERVICEMASK,
						p->state == DIALING ? "dial" : "callback",
						p->curtry == 1 ? "" : " (retry)");
				continue;
			}

			p->lasttime = time(NULL);
			
			if((ioctl(isdnfd, ISDN_DIAL, dial_p)) < 0)
			{
				if(errno != EBUSY)
				{
					log(LL_NDL, "T%d N%d C%d/- S%d %s%s %s ioctl ISDN_DIAL error: %s",
						p->atyp, p->appl,
						p->controller,
						p->service & SERVICEMASK,
						p->state == DIALING ? "dial" : "callback",
						p->curtry == 1 ? "" : " (retry)",
						dial_p->dst_telno, strerror(errno));
				}
				continue;
			}

			log(LL_DIA, "T%d N%d C%d/- S%d %s%s %s",
					p->atyp, p->appl,
					p->controller,
					p->service & SERVICEMASK,
					p->state == DIALING ? "dial" : "callback",
					p->curtry == 1 ? "" : " (retry)", 
					dial_p->dst_telno);
		}
	}
}

/* EOF */
