/*
 * Copyright (c) 1994, 1996 Hellmuth Michaelis. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *---------------------------------------------------------------------------*
 *
 *	bisdnd - BSD ISDN management daemon
 *      ===================================
 *
 *	Last Edit-Date: [Fri Jul  5 12:07:03 1996]
 *
 *	-hm	rewriting find_appl to support error return value
 *	-hm	removed function local static pointers
 *	-hm	remote number wildcard now possible
 *
 *----------------------------------------------------------------------------*/

#include <string.h>

#include <sys/time.h>

#include "defs.h"

/*---------------------------------------------------------------------------*
 *	verify incoming connections:
 *	for given channel, service, alert, src- and dst-telno find
 *	the corresponding application in the config struct
 *---------------------------------------------------------------------------*/
int
find_appl(accept_t **accept_pp, int chan, int service, int alert_id,
              char *dst_telno, char *src_telno)
{
	struct confdata *confdata_p;
	accept_t *accept_p;
	int i;
	int card;
	int ret;
	
	card = chan >> 1;		/* 2 channels per card, compute card no */

	if(card > NCARDS)
		return(FA_CARD);

	if(alert_id)			/* are we alerting ? */
	{
		struct alert *alert_p = &(alert[alert_id & 0x0f]);

		if(!(accept_p = alert_p->accept))
			return(FA_ERROR);

		alert_p->lastact = time(NULL);

		if(alert_p->cnt--)
		{
			log(LL_ALR, "              alert:");
			accept_p->alert = alert_id;
		}
		else
		{
			alert_p->id = INVALID;
			alert_p->accept = NULL;
			accept_p->alert = 0;			
		}
		*accept_pp = accept_p;
		return(FA_GOOD);
	}

	ret = FA_GOOD;

	/* get config struct entry */
	
	for(i=0; i <= config_entries; i++)
	{
		confdata_p = &conf[i];		/* ptr to config entry */
		
		/* incoming number = my number ? */

		if(strncmp(confdata_p->telnloc_rdi, dst_telno, strlen(confdata_p->telnloc_rdi)))
			continue;

		/* remote number = configured remote number ? */
		
		if( (*(confdata_p->telnrem_rdi) != '*') &&
		    (strncmp(confdata_p->telnrem_rdi, src_telno, strlen(confdata_p->telnrem_rdi))) )
		{
			ret |= FA_INV_DEST;	/* invalid dest telno */
			continue;
		}

		/* incoming service = service for this if ? */
		
		if((confdata_p->service & 0x0f) != (service & 0x0f))
		{
			ret |= FA_INV_SERV;	/* invalid service */
			continue;
		}

		/* controller = configured controller ? */

		if(confdata_p->controller != card)
		{
			ret |= FA_INV_CTRL;	/* invalid controller */
			continue;
		}
		
		/* found entry, leave loop */
		ret = FA_GOOD;
		break;
	}

	if(i > config_entries)
	{
		if(ret != FA_GOOD)	  /* my number, but error ? */
			return(ret);
		else
			return(FA_NOTME); /* no matching entry for this call */
	}
	
	/* fill in accept structure */
	
	switch(confdata_p->atyp)
	{
		case ATYP_IP:
			ret = ip_fill_accept(&accept_p, chan, confdata_p);
			break;
#ifdef NOTDEF
		case ATYP_TTY:
			ret = tty_fill_accept(&accept_p, chan, confdata_p);
			break;
#endif
		case ATYP_TEL:
			ret = tel_fill_accept(&accept_p, chan, confdata_p,
					dst_telno, src_telno);
			break;

		default:
			ret = FA_ATYP;
			break;
	}

	if(ret != FA_GOOD)
		return(ret);
	
	/* reaction = alert ? */
	
	if(confdata_p->reaction >= REACT_ALERTBASE)
	{
		time_t actime;
		struct alert *alert_p = alert;

		log(LL_ALR, "              alert:");

		time(&actime);

		for(i=0; i < NALERT; i++, alert_p++)
		{
			if((alert_p->cnt <= 0) ||
			   ((actime - (alert_p->lastact)) > 10))
			{
				alert_p->cnt = confdata_p->reaction - 2;
				alert_p->accept = accept_p;
				alert_p->lastact = actime;
				accept_p->alert = (i | 0x80);
				*accept_pp = accept_p;
				return(FA_GOOD);
			}
		}
		return(FA_ERROR);
	}

	/* in any case save the number of the remote site */
	
	strcpy(confdata_p->telno_remote, src_telno);

	*accept_pp = accept_p;
	return(FA_GOOD);
}

/*---------------------------------------------------------------------------*
 *	application type specific connect message processing
 *---------------------------------------------------------------------------*/
int
connected(accept_t **accept_pp, int isdnfd, int atyp, int appl_no, int chan,
		int active)
{
	accept_t *accept_p;
	int ret;
	
	switch(atyp)
	{
		case ATYP_IP:
			accept_p = connect_ip(appl_no, chan, active);
			ret = GOOD;
			break;
#ifdef NOTDEF
		case ATYP_TTY:
			accept_p = connect_tty(isdnfd, appl_no, chan, active);
			ret = GOOD;
			break;
#endif
		case ATYP_TEL:
			accept_p = connect_tel(appl_no, chan, active);
			ret = GOOD;
			break;
		default:
			accept_p = NULL;
			ret = ERROR;
			break;
			
	}

	if(ret == GOOD)
		*accept_pp = accept_p;

	return(ret);
}

/*---------------------------------------------------------------------------*
 *	application type specific dial message processing
 *---------------------------------------------------------------------------*/
int
get_telno(dial_t **dial_pp, int atyp, int appl_no)
{
	dial_t *dial_p;
	int ret = ERROR;

	switch(atyp)
	{
		case ATYP_IP:
			if((dial_p = getno_ip(appl_no)) != NULL)
				ret = GOOD;
			break;
#ifdef NOTDEF
		case ATYP_TTY:
			if((dial_p = getno_tty(appl_no)) != NULL)
				ret = GOOD;
			break;
#endif
		case ATYP_TEL:
			if((dial_p = getno_tel(appl_no)) != NULL)
				ret = GOOD;
			break;

		default:
			dial_p = NULL;
			ret = ERROR;
			break;
	}

	if(ret == GOOD)
		*dial_pp = dial_p;

	return(ret);
}

/*---------------------------------------------------------------------------*
 *	application type specific disconnect message processing
 *---------------------------------------------------------------------------*/
void
disconnected(int isdnfd, int atyp, int appl_no)
{
	switch(atyp)
	{
		case ATYP_IP:
			disconnect_ip(appl_no);
			break;
#ifdef NOTDEF
		case ATYP_TTY:
			disconnect_tty(isdnfd, appl_no);
			break;
#endif
		case ATYP_TEL:
			disconnect_tel(appl_no);
			break;
	}
}

/* EOF */
