/*
 * Copyright (c) 1996 Hellmuth Michaelis.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *---------------------------------------------------------------------------*
 *
 *	bisdntrc_q931 - print Q.931 traces
 *	----------------------------------
 *
 *	last edit-date: [Fri Jul  5 12:20:13 1996]
 *
 *	-hm	splitting
 *	-hm	decoding ie's
 *	-hm	fixing length bug in p_bc()
 *	-hm	PROT_1TR6 -> PROT_DSS1 in p_cause()
 *
 *---------------------------------------------------------------------------*/

#include "bisdntrc.h"

static int p_cause(char *pbuf, unsigned char buf[]);
static int p_address(char *pbuf, unsigned char buf[]);
static int p_bc(char *pbuf, unsigned char buf[]);

/*---------------------------------------------------------------------------*
 *	decode DSS1/Q.931 ("Euro-ISDN") protocol
 *---------------------------------------------------------------------------*/
void
decode_q931(char *pbuf, int n, unsigned char *buf, int raw)
{
	int codeset = 0;
	int codelock = 0;
	int oldcodeset = 0;
	
	int pd;
	int len;
	int j;
	int i;

	if(n <= 0)
		return;

	*pbuf = '\0';
	
	if(raw)
	{

		for (i = 0; i < n; i += 16)
		{
			sprintf((pbuf+strlen(pbuf)),"Dump:%.3x  ", i);
			for (j = 0; j < 16; j++)
				if (i + j < n)
					sprintf((pbuf+strlen(pbuf)),"%02x ", buf[i + j]);
				else
					sprintf((pbuf+strlen(pbuf)),"   ");
			sprintf((pbuf+strlen(pbuf)),"      ");
			for (j = 0; j < 16 && i + j < n; j++)
				if (isprint(buf[i + j]))
					sprintf((pbuf+strlen(pbuf)),"%c", buf[i + j]);
				else
					sprintf((pbuf+strlen(pbuf)),".");
			sprintf((pbuf+strlen(pbuf)),"\n");
		}
	}

	i = 0;
		
	sprintf((pbuf+strlen(pbuf)), "Q931: ");

	/* protocol discriminator */

	pd = buf[i];

	if(pd >= 0x00 && pd <= 0x07)
		sprintf((pbuf+strlen(pbuf)), "pd=User-User (0x%02x), ",pd);
	else if(pd == 0x08)
		sprintf((pbuf+strlen(pbuf)), "pd=Q.931/I.451, ");
	else if(pd >= 0x10 && pd <= 0x3f)
		sprintf((pbuf+strlen(pbuf)), "pd=Other Layer 3 or X.25 (0x%02x), ",pd);
	else if(pd >= 0x40 && pd <= 0x4f)
		sprintf((pbuf+strlen(pbuf)), "pd=National Use (0x%02x), ",pd);
	else if(pd >= 0x50 && pd <= 0xfe)
		sprintf((pbuf+strlen(pbuf)), "pd=Other Layer 3 or X.25 (0x%02x), ",pd);
	else
		sprintf((pbuf+strlen(pbuf)), "pd=Reserved (0x%02x), ",pd);

	/* call reference */

	i++;

	len = buf[i] & 0x0f;

	switch(len)
	{
		case 0:
			sprintf((pbuf+strlen(pbuf)), "cr=Dummy, ");
			break;
		case 1:
			sprintf((pbuf+strlen(pbuf)), "cr=0x%02x %s, ", (buf[i+1] & 0x7f), (buf[i+1] & 0x80) ? "(from destination)" : "(from origination)");
			break;
		case 2:
			sprintf((pbuf+strlen(pbuf)), "cr=0x%02x 0x%02x %s, ", (buf[i+1] & 0x7f), (buf[i+2] & 0x7f), (buf[i+1] & 0x80) ? "(org)" : "(dst)");
			break;
	}

	i += (len+1);

	/* message type */	

	sprintf((pbuf+strlen(pbuf)), "message=");

	switch(buf[i])
	{
		/* escape to nationally specific message type */

		case 0x00:
			sprintf((pbuf+strlen(pbuf)), "ESCAPE: ");
			break;

		/* call establishment */

		case 0x01:
			sprintf((pbuf+strlen(pbuf)), "ALERTING: ");
			break;
		case 0x02:
			sprintf((pbuf+strlen(pbuf)), "CALL PROCEEDING: ");
			break;
		case 0x03:
			sprintf((pbuf+strlen(pbuf)), "PROGRESS: ");
			break;
		case 0x05:
			sprintf((pbuf+strlen(pbuf)), "SETUP: ");
			break;
		case 0x07:
			sprintf((pbuf+strlen(pbuf)), "CONNECT: ");
			break;
		case 0x0d:
			sprintf((pbuf+strlen(pbuf)), "SETUP ACKNOWLEDGE: ");
			break;
		case 0x0f:
			sprintf((pbuf+strlen(pbuf)), "CONNECT ACKNOWLEDGE: ");
			break;

		/* call information phase */

		case 0x20:
			sprintf((pbuf+strlen(pbuf)), "USER INFORMATION: ");
			break;
		case 0x21:
			sprintf((pbuf+strlen(pbuf)), "SUSPEND REJECT: ");
			break;
		case 0x22:
			sprintf((pbuf+strlen(pbuf)), "RESUME REJECT: ");
			break;
		case 0x24:
			sprintf((pbuf+strlen(pbuf)), "HOLD (Q.932): ");
			break;
		case 0x25:
			sprintf((pbuf+strlen(pbuf)), "SUSPEND: ");
			break;
		case 0x26:
			sprintf((pbuf+strlen(pbuf)), "RESUME: ");
			break;
		case 0x28:
			sprintf((pbuf+strlen(pbuf)), "HOLD ACKNOWLEDGE (Q.932): ");
			break;
		case 0x2d:
			sprintf((pbuf+strlen(pbuf)), "SUSPEND ACKNOWLEDGE: ");
			break;
		case 0x2e:
			sprintf((pbuf+strlen(pbuf)), "RESUME ACKNOWLEDGE: ");
			break;
		case 0x30:
			sprintf((pbuf+strlen(pbuf)), "HOLD REJECT (Q.932): ");
			break;
		case 0x31:
			sprintf((pbuf+strlen(pbuf)), "RETRIEVE (Q.932): ");
			break;
		case 0x32:
			sprintf((pbuf+strlen(pbuf)), "RETRIEVE ACKNOWLEDGE (Q.932): ");
			break;
		case 0x37:
			sprintf((pbuf+strlen(pbuf)), "RETRIEVE REJECT (Q.932): ");
			break;

		/* call clearing */
			
		case 0x40:
			sprintf((pbuf+strlen(pbuf)), "DETACH: ");
			break;
		case 0x45:
			sprintf((pbuf+strlen(pbuf)), "DISCONNECT: ");
			break;
		case 0x46:
			sprintf((pbuf+strlen(pbuf)), "RESTART: ");
			break;
		case 0x48:
			sprintf((pbuf+strlen(pbuf)), "DETACH ACKNOWLEDGE: ");
			break;
		case 0x4d:
			sprintf((pbuf+strlen(pbuf)), "RELEASE: ");
			break;
		case 0x4e:
			sprintf((pbuf+strlen(pbuf)), "RESTART ACKNOWLEDGE: ");
			break;
		case 0x5a:
			sprintf((pbuf+strlen(pbuf)), "RELEASE COMPLETE: ");
			break;
			
		/* misc messages */

		case 0x60:
			sprintf((pbuf+strlen(pbuf)), "SEGMENT: ");
			break;
		case 0x62:
			sprintf((pbuf+strlen(pbuf)), "FACILITY (Q.932): ");
			break;
		case 0x64:
			sprintf((pbuf+strlen(pbuf)), "REGISTER (Q.932): ");
			break;
		case 0x68:
			sprintf((pbuf+strlen(pbuf)), "CANCEL ACKNOWLEDGE: ");
			break;
		case 0x6a:
			sprintf((pbuf+strlen(pbuf)), "FACILITY ACKNOWLEDGE: ");
			break;
		case 0x6c:
			sprintf((pbuf+strlen(pbuf)), "REGISTER ACKNOWLEDGE: ");
			break;
		case 0x6e:
			sprintf((pbuf+strlen(pbuf)), "NOTIFY: ");
			break;
		case 0x70:
			sprintf((pbuf+strlen(pbuf)), "CANCEL REJECT: ");
			break;
		case 0x72:
			sprintf((pbuf+strlen(pbuf)), "FACILITY REJECT: ");
			break;
		case 0x74:
			sprintf((pbuf+strlen(pbuf)), "REGISTER REJECT: ");
			break;
		case 0x75:
			sprintf((pbuf+strlen(pbuf)), "STATUS ENQIRY: ");
			break;
		case 0x79:
			sprintf((pbuf+strlen(pbuf)), "CONGESTION CONTROL: ");
			break;
		case 0x7b:
			sprintf((pbuf+strlen(pbuf)), "INFORMATION: ");
			break;
		case 0x7d:
			sprintf((pbuf+strlen(pbuf)), "STATUS: ");
			break;
		default:
			sprintf((pbuf+strlen(pbuf)), "UNDEFINED, TYPE=0x%02x, ", buf[i]);
			break;
	}

	/* other information elements */

	i++;
	
	for (; i < n;)
	{
		sprintf((pbuf+strlen(pbuf)), "\n     ");
		
		if(buf[i] & 0x80)
		{
			/* single octett info element */

			switch(buf[i] & 0x70)
			{
				case 0x00:	/* reserved */
					sprintf((pbuf+strlen(pbuf)), "[reserved single octett info]");
					break;

				case 0x10:	/* shift */
					oldcodeset = codeset;
					codeset = buf[i] & 0x07;
					if(buf[i] & 0x08)
						codelock = 0;
					else
						codelock = 1;
					sprintf((pbuf+strlen(pbuf)), "[shift: codeset=%d lock=%d]", codeset, codelock);
					break;

				case 0x20:	/* more data */
					if(buf[i] & 0x01)
						sprintf((pbuf+strlen(pbuf)), "[sending complete]");
					else
						sprintf((pbuf+strlen(pbuf)), "[more data]");
					break;

				case 0x30:	/* congestion level */
					sprintf((pbuf+strlen(pbuf)), "[congestion level=");
					switch(buf[i] & 0x0f)
					{
						case 0x00:
							sprintf((pbuf+strlen(pbuf)), "rx-ready]");
							break;
						case 0x0f:
							sprintf((pbuf+strlen(pbuf)), "rx-not-ready]");
							break;
						default:
							sprintf((pbuf+strlen(pbuf)), "reserved (0x%02x)]", buf[i] & 0x0f);
							break;
					}
					break;
					
				case 0x50:	/* repeat ind */
					sprintf((pbuf+strlen(pbuf)), "[repeat indicator]");
					break;

				default:
					sprintf((pbuf+strlen(pbuf)), "[UNDEF SINGLE OCTET ELEMENT 0x%02x]", buf[i]);
					break;
			}

			i++;	/* next */

		}
		else
		{
			/* variable length info element */

			if(codeset == 0)
			{
				switch(buf[i])
				{
					case 0x00:
						sprintf((pbuf+strlen(pbuf)), "[segmented message: ");
						break;
					case 0x04:
						sprintf((pbuf+strlen(pbuf)), "[bearer capability: ");
						i += p_bc(pbuf, &buf[i]);
						goto next;
						break;
					case 0x08:
						sprintf((pbuf+strlen(pbuf)), "[cause: ");
						i += p_cause(pbuf, &buf[i]);
						goto next;
						break;
					case 0x0c:
						sprintf((pbuf+strlen(pbuf)), "[connected address (old): ");
						break;
					case 0x0d:
						sprintf((pbuf+strlen(pbuf)), "[extended facility (Q.932: )");
						break;
					case 0x10:
						sprintf((pbuf+strlen(pbuf)), "[call identity: ");
						break;
					case 0x14:
						sprintf((pbuf+strlen(pbuf)), "[call state: ");
						break;
					case 0x18:
						sprintf((pbuf+strlen(pbuf)), "[channel id: channel=");
						i++;
						len = buf[i];
						i++;
						switch(buf[i] & 0x03)
						{
							case 0:
								sprintf((pbuf+strlen(pbuf)), "no channel");
								break;
							case 1:
								sprintf((pbuf+strlen(pbuf)), "B-1");
								break;
							case 2:
								sprintf((pbuf+strlen(pbuf)), "B-2");
								break;
							case 3:
								sprintf((pbuf+strlen(pbuf)), "any channel");
								break;
						}
						if(buf[i] & 0x08)
							sprintf((pbuf+strlen(pbuf)), " (exclusive)]");
						else
							sprintf((pbuf+strlen(pbuf)), " (preferred)]");
						i++;
						goto next;
						break;
					case 0x19:
						sprintf((pbuf+strlen(pbuf)), "[data link connection id (Q.933): ");
						break;
					case 0x1c:
						sprintf((pbuf+strlen(pbuf)), "[facility (Q.932): ");
						break;
					case 0x1e:
						sprintf((pbuf+strlen(pbuf)), "[progress ind: ");
						break;
					case 0x20:
						sprintf((pbuf+strlen(pbuf)), "[network specific facilities: ");
						break;
					case 0x24:
						sprintf((pbuf+strlen(pbuf)), "[terminal capabilities: ");
						break;
					case 0x27:
						sprintf((pbuf+strlen(pbuf)), "[notification ind: ");
						break;
					case 0x28:
						sprintf((pbuf+strlen(pbuf)), "[display: ");
						i++;
						len = buf[i];
						i++;
						for(j = 0; j < len; j++)
						{
							sprintf((pbuf+strlen(pbuf)),"%c", buf[j+i]);
						}
						sprintf((pbuf+strlen(pbuf)),"]");
						i += j;
						goto next;
						break;
					case 0x29:
						sprintf((pbuf+strlen(pbuf)), "[date/time: ");
						i++;
						len = buf[i];
						i++;
						for(j = 0; j < len; j++)
						{
							sprintf((pbuf+strlen(pbuf)),"%.2d ", buf[j+i]);
						}
						sprintf((pbuf+strlen(pbuf)),"]");
						i += j;
						goto next;
						break;
					case 0x2c:
						sprintf((pbuf+strlen(pbuf)), "[keypad: ");
						break;
					case 0x30:
						sprintf((pbuf+strlen(pbuf)), "[keypad echo: ");
						break;
					case 0x32:
						sprintf((pbuf+strlen(pbuf)), "[information req (Q.932): ");
						break;
					case 0x34:
						sprintf((pbuf+strlen(pbuf)), "[signal: ");
						break;
					case 0x36:
						sprintf((pbuf+strlen(pbuf)), "[switchhook: ");
						break;
					case 0x38:
						sprintf((pbuf+strlen(pbuf)), "[feature activation (Q.932): ");
						break;
					case 0x39:
						sprintf((pbuf+strlen(pbuf)), "[feature ind (Q.932): ");
						break;
					case 0x3a:
						sprintf((pbuf+strlen(pbuf)), "[service profile id (Q.932): ");
						break;
					case 0x3b:
						sprintf((pbuf+strlen(pbuf)), "[endpoint id (Q.932): ");
						break;
					case 0x40:
						sprintf((pbuf+strlen(pbuf)), "[information rate: ");
						break;
					case 0x41:
						sprintf((pbuf+strlen(pbuf)), "[precedence level (Q.955): ");
						break;
					case 0x42:
						sprintf((pbuf+strlen(pbuf)), "[end-to-end transit delay: ");
						break;
					case 0x43:
						sprintf((pbuf+strlen(pbuf)), "[transit delay detection and indication: ");
						break;
					case 0x44:
						sprintf((pbuf+strlen(pbuf)), "[packet layer binary parameters: ");
						break;
					case 0x45:
						sprintf((pbuf+strlen(pbuf)), "[packet layer window size: ");
						break;
					case 0x46:
						sprintf((pbuf+strlen(pbuf)), "[packet size: ");
						break;
					case 0x47:
						sprintf((pbuf+strlen(pbuf)), "[closed user group: ");
						break;
					case 0x48:
						sprintf((pbuf+strlen(pbuf)), "[link layer core parameters (Q.933): ");
						break;
					case 0x49:
						sprintf((pbuf+strlen(pbuf)), "[link layer protocol parameters (Q.933): ");
						break;
					case 0x4a:
						sprintf((pbuf+strlen(pbuf)), "[reverse charging information: ");
						break;
					case 0x4c:
						sprintf((pbuf+strlen(pbuf)), "[connected number (Q.951): ");
						break;
					case 0x4d:
						sprintf((pbuf+strlen(pbuf)), "[connected subaddress (Q.951): ");
						break;
					case 0x50:
						sprintf((pbuf+strlen(pbuf)), "[X.213 priority (Q.933): ");
						break;
					case 0x51:
						sprintf((pbuf+strlen(pbuf)), "[report type (Q.933): ");
						break;
					case 0x53:
						sprintf((pbuf+strlen(pbuf)), "[link integrity verification (Q.933): ");
						break;
					case 0x57:
						sprintf((pbuf+strlen(pbuf)), "[PVC status (Q.933): ");
						break;
					case 0x6c:
						sprintf((pbuf+strlen(pbuf)), "[calling party number: ");
						i += p_address(pbuf, &buf[i]);
						goto next;
						break;
					case 0x6d:
						sprintf((pbuf+strlen(pbuf)), "[calling party subaddress: ");
						break;
					case 0x70:
						sprintf((pbuf+strlen(pbuf)), "[called party number: ");
						i += p_address(pbuf, &buf[i]);
						goto next;
						break;
					case 0x71:
						sprintf((pbuf+strlen(pbuf)), "[called party subaddress: ");
						break;
					case 0x74:
						sprintf((pbuf+strlen(pbuf)), "[redirecting number: ");
						break;
					case 0x76:
						sprintf((pbuf+strlen(pbuf)), "[redirection number: ");
						break;
					case 0x78:
						sprintf((pbuf+strlen(pbuf)), "[transit network selection: ");
						break;
					case 0x79:
						sprintf((pbuf+strlen(pbuf)), "[restart indicator: ");
						break;
					case 0x7c:
						sprintf((pbuf+strlen(pbuf)), "[low layer compatibility: ");
						break;
					case 0x7d:
						sprintf((pbuf+strlen(pbuf)), "[high layer compatibility: ");
						break;
					case 0x7e:
						sprintf((pbuf+strlen(pbuf)), "[user-user: ");
						break;
					case 0x7f:
						sprintf((pbuf+strlen(pbuf)), "[escape for extension: ");
						break;
					default:
						sprintf((pbuf+strlen(pbuf)), "[UNKNOWN INFO-ELEMENT-ID=0x%02x: ", buf[i]);
						break;
				}
			}
			else
			{
				sprintf((pbuf+strlen(pbuf)), "[UNKNOWN CODESET=%d, MSG=0x%02x", codeset, buf[i]);
			}

			i++;	/* index -> length */

			len = buf[i];

			i++;	/* index -> 1st param */

			for(j = 0; j < len; j++)
			{
				sprintf((pbuf+strlen(pbuf))," 0x%02x", buf[j+i]);
			}
	
			sprintf((pbuf+strlen(pbuf)),"]");

			i += len;

next:

			if(!codelock && (codeset != oldcodeset))
				codeset = oldcodeset;
		}
	}
	sprintf((pbuf+strlen(pbuf)),"\n");
}

/*---------------------------------------------------------------------------*
 *	decode and print the cause
 *---------------------------------------------------------------------------*/
static int
p_cause(char *pbuf, unsigned char buf[])
{
	int j;
	int len;
	int i = 0;
	int ls;
	int r = 0;
	int rflag = 0;
	
	i++;	/* index -> length */

	len = buf[i];

	i++;	/* coding/location */
	len--;

	ls = buf[i];
	
	i++;
	len--;
	
	if(!(buf[i] & 0x80))
	{
		r = buf[i];		
		rflag = 1;
		i++;
		len--;
	}

	sprintf((pbuf+strlen(pbuf)), "%s ", print_error(ISDNDSS1, buf[i]));

	sprintf((pbuf+strlen(pbuf)), "\n          (location=");
	
	switch(ls & 0x0f)
	{
		case 0x00:
			sprintf((pbuf+strlen(pbuf)), "user");
			break;
		case 0x01:
			sprintf((pbuf+strlen(pbuf)), "private network serving local user");
			break;
		case 0x02:
			sprintf((pbuf+strlen(pbuf)), "public network serving local user");
			break;
		case 0x03:
			sprintf((pbuf+strlen(pbuf)), "transit network");
			break;
		case 0x04:
			sprintf((pbuf+strlen(pbuf)), "public network serving remote user");
			break;
		case 0x05:
			sprintf((pbuf+strlen(pbuf)), "private network serving remote user");
			break;
		case 0x07:
			sprintf((pbuf+strlen(pbuf)), "international network");
			break;
		case 0x0a:
			sprintf((pbuf+strlen(pbuf)), "network beyond interworking point");
			break;
		default:
			sprintf((pbuf+strlen(pbuf)), "reserved (0x%02x)", ls & 0x0f);
			break;
	}

	sprintf((pbuf+strlen(pbuf)), ", std=");

	switch((ls & 0x60) >> 5)
	{
		case 0:
			sprintf((pbuf+strlen(pbuf)), "CCITT");
			break;
		case 1:
			sprintf((pbuf+strlen(pbuf)), "ISO/IEC");
			break;
		case 2:
			sprintf((pbuf+strlen(pbuf)), "National");
			break;
		case 3:
			sprintf((pbuf+strlen(pbuf)), "Local");
			break;
	}

	if(rflag)
	{
		sprintf((pbuf+strlen(pbuf)), ", rec=");

		switch(r & 0x7f)
		{
			case 0:
				sprintf((pbuf+strlen(pbuf)), "Q.931");
				break;
			case 3:
				sprintf((pbuf+strlen(pbuf)), "X.21");
				break;
			case 4:
				sprintf((pbuf+strlen(pbuf)), "X.25");
				break;
			case 5:
				sprintf((pbuf+strlen(pbuf)), "Q.1031/Q.1051");
				break;
			default:
				sprintf((pbuf+strlen(pbuf)), "Reserved");
				break;
		}
	}

	sprintf((pbuf+strlen(pbuf)),")");
	
	i++;
	len--;
	
	for(j = 0; j < len; j++)
		sprintf((pbuf+strlen(pbuf))," 0x%02x", buf[j+i]);
	
	sprintf((pbuf+strlen(pbuf)),"]");

	i += (len+1);
	
	return(i);
}

/*---------------------------------------------------------------------------*
 *	decode and print the bearer capability
 *---------------------------------------------------------------------------*/
static int
p_bc(char *pbuf, unsigned char buf[])
{
	int len;
	int i = 0;
	int mr = 0;
	
	i++;	/* index -> length */

	len = buf[i];

	i++;

	sprintf((pbuf+strlen(pbuf)), "\n          cap=");
	
	switch(buf[i] & 0x1f)
	{
		case 0x00:
			sprintf((pbuf+strlen(pbuf)), "speech");
			break;
		case 0x08:
			sprintf((pbuf+strlen(pbuf)), "unrestricted digital information");
			break;
		case 0x09:
			sprintf((pbuf+strlen(pbuf)), "restricted digital information");
			break;
		case 0x10:
			sprintf((pbuf+strlen(pbuf)), "3.1 kHz audio");
			break;
		case 0x11:
			sprintf((pbuf+strlen(pbuf)), "unrestricted digital information with tones");
			break;
		case 0x18:
			sprintf((pbuf+strlen(pbuf)), "video");
			break;
		default:
			sprintf((pbuf+strlen(pbuf)), "reserved (0x%02x)", buf[i] & 0x0f);
			break;
	}

	sprintf((pbuf+strlen(pbuf)), "\n          std=");

	switch((buf[i] & 0x60) >> 5)
	{
		case 0:
			sprintf((pbuf+strlen(pbuf)), "CCITT");
			break;
		case 1:
			sprintf((pbuf+strlen(pbuf)), "ISO/IEC");
			break;
		case 2:
			sprintf((pbuf+strlen(pbuf)), "National");
			break;
		case 3:
			sprintf((pbuf+strlen(pbuf)), "NSI Std");
			break;
	}

	i++;
	len--;
	
	sprintf((pbuf+strlen(pbuf)), "\n          rate=");
	
	switch(buf[i] & 0x1f)
	{
		case 0x00:
			sprintf((pbuf+strlen(pbuf)), "packet mode");
			break;
		case 0x10:
			sprintf((pbuf+strlen(pbuf)), "64 kbit/s");
			break;
		case 0x11:
			sprintf((pbuf+strlen(pbuf)), "2 x 64 kbit/s");
			break;
		case 0x13:
			sprintf((pbuf+strlen(pbuf)), "384 kbit/s");
			break;
		case 0x15:
			sprintf((pbuf+strlen(pbuf)), "1536 kbit/s");
			break;
		case 0x17:
			sprintf((pbuf+strlen(pbuf)), "1920 kbit/s");
			break;
		case 0x18:
			sprintf((pbuf+strlen(pbuf)), "Multirate");
			mr = 1;
			break;
		default:
			sprintf((pbuf+strlen(pbuf)), "reserved (0x%02x)", buf[i] & 0x0f);
			break;
	}

	sprintf((pbuf+strlen(pbuf)), "\n          mode=");

	switch((buf[i] & 0x60) >> 5)
	{
		case 0:
			sprintf((pbuf+strlen(pbuf)), "circuit");
			break;
		case 2:
			sprintf((pbuf+strlen(pbuf)), "packet");
			break;
		default:
			sprintf((pbuf+strlen(pbuf)), "reserved (0x%02x)", ((buf[i] & 0x60) >> 5));
			break;
	}

	i++;
	len--;

	if(!len)
		goto exit;
	
	if(mr)
	{
		sprintf((pbuf+strlen(pbuf)), "\n          rate multiplier=%d", buf[i] & 0x7f);
		i++;
		len--;
	}

	if(!len)
		goto exit;
	
	if(!(buf[i] & 0x80))
	{
		{
			sprintf((pbuf+strlen(pbuf)), "\n          layer1=");
	
			switch(buf[i] & 0x1f)
			{
				case 0x01:
					sprintf((pbuf+strlen(pbuf)), "V.110");
					break;
				case 0x02:
					sprintf((pbuf+strlen(pbuf)), "G.711 u-law");
					break;
				case 0x03:
					sprintf((pbuf+strlen(pbuf)), "G.711 A-law");
					break;
				case 0x04:
					sprintf((pbuf+strlen(pbuf)), "G.721");
					break;
				case 0x05:
					sprintf((pbuf+strlen(pbuf)), "H.221/H.242");
					break;
				case 0x07:
					sprintf((pbuf+strlen(pbuf)), "Non-Std");
					break;
				case 0x08:
					sprintf((pbuf+strlen(pbuf)), "V.120");
					break;
				case 0x09:
					sprintf((pbuf+strlen(pbuf)), "X.31");
					break;
				default:
					sprintf((pbuf+strlen(pbuf)), "reserved (0x%02x)", buf[i] & 0x0f);
					break;
			}
			i++;
			len--;
		}
	
		if(!len)
			goto exit;

		if(!(buf[i] & 0x80))
		{
			sprintf((pbuf+strlen(pbuf)), "\n          user rate=0x%02x ", buf[i] & 0x1f);
	
			if(buf[i] & 0x40)
				sprintf((pbuf+strlen(pbuf)), "(async,");
			else
				sprintf((pbuf+strlen(pbuf)), "(sync,");		
	
			if(buf[i] & 0x20)
				sprintf((pbuf+strlen(pbuf)), "in-band neg. possible)");
			else
				sprintf((pbuf+strlen(pbuf)), "in-band neg not possible)");
			
			i++;
			len--;
		}
	
		if(!len)
			goto exit;

		if(!(buf[i] & 0x80))
		{
			sprintf((pbuf+strlen(pbuf)), "\n          clk/flow=0x%02x", buf[i] & 0x1f);
	
			sprintf((pbuf+strlen(pbuf)), "\n          intermediate rate=");
			
			switch((buf[i] & 0x60) >> 5)
			{
				case 0:
					sprintf((pbuf+strlen(pbuf)), "not used");
					break;
				case 1:
					sprintf((pbuf+strlen(pbuf)), "8 kbit/s");
					break;
				case 2:
					sprintf((pbuf+strlen(pbuf)), "16 kbit/s");
					break;
				case 3:
					sprintf((pbuf+strlen(pbuf)), "32 kbit/s");
					break;
			}
			i++;
			len--;
		}
	
		if(!len)
			goto exit;

		if(!(buf[i] & 0x80))
		{
			sprintf((pbuf+strlen(pbuf)), "\n          hdr/mfrm/etc.=0x%02x", buf[i]);
			i++;
			len--;
		}
	
		if(!len)
			goto exit;

		if(!(buf[i] & 0x80))
		{
			sprintf((pbuf+strlen(pbuf)), "\n          stop/data/parity=0x%02x", buf[i]);
			i++;
			len--;
		}
	
		if(!len)
			goto exit;

		if(!(buf[i] & 0x80))
		{
			sprintf((pbuf+strlen(pbuf)), "\n          modemtype=0x%02x", buf[i]);
			i++;
			len--;
		}
	}

	if(!len)
		goto exit;

	if(len)
	{
		switch(buf[i] & 0x7f)
		{
			case 0x42:
				sprintf((pbuf+strlen(pbuf)), "\n          layer2=Q.921/I.441");
				break;
			case 0x46:
				sprintf((pbuf+strlen(pbuf)), "\n          layer2=X.25 link");
				break;
			default:
				sprintf((pbuf+strlen(pbuf)), "\n          layer2=0x%02x",(buf[i] & 0x7f));
				break;
		}
		i++;
		len--;
	}

	if(!len)
		goto exit;
	
	if(len)
	{
		switch(buf[i] & 0x7f)
		{
			case 0x62:
				sprintf((pbuf+strlen(pbuf)), "\n          layer3=Q.921/I.441");
				break;
			case 0x66:
				sprintf((pbuf+strlen(pbuf)), "\n          layer3=X.25 packet");
				break;
			default:
				sprintf((pbuf+strlen(pbuf)), "\n          layer3=0x%02x",(buf[i] & 0x7f));
				break;
		}
		i++;
		len--;
	}

exit:	
	sprintf((pbuf+strlen(pbuf)), "]");	
	return(i);
}

/*---------------------------------------------------------------------------*
 *	decode and print the ISDN (telephone) number
 *---------------------------------------------------------------------------*/
static int
p_address(char *pbuf, unsigned char buf[])
{
	int j;
	int len;
	int i = 0;
	int tp;
	int ind = 0;
	int indflag = 0;
	
	i++;	/* index -> length */
	len = buf[i];

	i++;	/* index -> type/plan */
	tp = buf[i];
	
	i++;
	len--;
	
	if(!(tp & 0x80))
	{
		ind = buf[i];		
		indflag = 1;
		i++;
		len--;
	}

	for(j = 0; j < len; j++)
	{
		sprintf((pbuf+strlen(pbuf)),"%c", buf[j+i]);
	}

	switch((tp & 0x70) >> 4)
	{
		case 0:
			sprintf((pbuf+strlen(pbuf)), " (type=unknown, ");
			break;
		case 1:
			sprintf((pbuf+strlen(pbuf)), " (type=international, ");
			break;
		case 2:
			sprintf((pbuf+strlen(pbuf)), " (type=national, ");
			break;
		case 3:
			sprintf((pbuf+strlen(pbuf)), " (type=network specific, ");
			break;
		case 4:
			sprintf((pbuf+strlen(pbuf)), " (type=subscriber, ");
			break;
		case 6:
			sprintf((pbuf+strlen(pbuf)), " (type=abbreviated, ");
			break;
		default:
			sprintf((pbuf+strlen(pbuf)), " (type=reserved (%d), ", ((tp & 0x70) >> 4));
			break;
	}

	switch(tp & 0x0f)
	{
		case 0:
			sprintf((pbuf+strlen(pbuf)), "plan=unknown");
			break;
		case 1:
			sprintf((pbuf+strlen(pbuf)), "plan=ISDN");
			break;
		case 3:
			sprintf((pbuf+strlen(pbuf)), "plan=Data");
			break;
		case 4:
			sprintf((pbuf+strlen(pbuf)), "plan=Telex");
			break;
		case 8:
			sprintf((pbuf+strlen(pbuf)), "plan=National");
			break;
		case 9:
			sprintf((pbuf+strlen(pbuf)), "plan=private");
			break;
		default:
			sprintf((pbuf+strlen(pbuf)), "plan=reserved (%d)", (tp & 0x0f));
			break;
	}

	if(indflag)
	{
		sprintf((pbuf+strlen(pbuf)), ",\n          ");
		switch((ind & 0x60) >> 5)
		{
			case 0:
				sprintf((pbuf+strlen(pbuf)), "presentation allowed, ");
				break;
			case 1:
				sprintf((pbuf+strlen(pbuf)), "presentation restricted, ");
				break;
			case 2:
				sprintf((pbuf+strlen(pbuf)), "number not available, ");
				break;
			case 3:
				sprintf((pbuf+strlen(pbuf)), "reserved, ");
				break;
		}

		switch(ind & 0x03)
		{
			case 0:
				sprintf((pbuf+strlen(pbuf)), "screening user prov: not screened");
				break;
			case 1:
				sprintf((pbuf+strlen(pbuf)), "screening user prov: verfyd & pass");
				break;
			case 2:
				sprintf((pbuf+strlen(pbuf)), "screening user prov: verfyd & fail");
				break;
			case 3:
				sprintf((pbuf+strlen(pbuf)), "screening network provided");
				break;
		}
	}		

	sprintf((pbuf+strlen(pbuf)),")]");

	i += j;

	return(i);
}

/* EOF */

