/*
 * Copyright (c) 1995 Dietmar Friede and Juergen Krause.  All rights reserved.
 *
 * Copyright (c) 1996 Arne Helme. All rights reserved.
 *
 * Copyright (c) 1996 Gary Jennejohn. All rights reserved. 
 *
 * Copyright (c) 1995, 1996 Hellmuth Michaelis. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Dietmar Friede,
 *	Juergen Krause, Arne Helme, Gary Jennejohn and Hellmuth Michaelis.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *---------------------------------------------------------------------------
 *
 *	ipi (ip over isdn) network driver for the bisdn package
 *	-------------------------------------------------------
 *
 *	last edit-date: [Sun Jul  8 08:12:21 1996]
 *
 *	-hm	totally rewrote the ii driver
 *	-hm	adding accounting support
 *	-hm	adding Van Jacobsen header compression
 *	-hm	watchdog for dyn. accounting report to iid
 *	-hm	adding Arne's NetBSD support
 *	-gj	unify for 2.1R, -stable, -cuurent && NetBSD
 *	-hm	adding Arne's unify patches
 *	-hm	small bpf fix for NetBSD from Arne
 *	-hm	applying watchdog/sl_compress patch from Gary
 *	-hm	merging in watchdog patch for NetBSD from Arne
 *	-hm	aaaand another patch from Gary for FreeBSD-current
 *	-hm	fixing interface lockup due to send queue fill
 *	-hm	fixing MTU setting (bugreport from Greg)
 *	-gj	include opt_bisdn.h for -current, delete mm from
 *			ipioutput (-Wunused complained), Joerg's comment
 *
 *---------------------------------------------------------------------------
 *
 *	Interface Configuration:
 *	------------------------
 *	ifconfig ipiX link0	- traffic on this interface will be
 *				  VJ compressed
 *	ifconfig ipiX link1	- check if we get VJ compressed packets,
 *				  if yes, automatically set link0
 *
 *	Interface Statistics:
 *	---------------------
 *	ifnet.ierrors		- counts dropped in packets (queue full)
 *	ifnet.oerrors		- counts dropped out packets (queue full)
 *	ifnet.collisions	- counts invalid packets after decompression
 *	ifnet.noproto		- counts invalid AF outbound packets
 *
 *	Compile time options:
 *	---------------------
 *
 * 	IPI_VJ			- do Van Jacobsen header compression.
 *	IPI_DIPA=n		- send IP accounting packets containing the
 *				  number of input bytes, output bytes, (if
 *				  IPI_VJ also compressed input bytes, compressed
 *				  output bytes,) input packets and ouput
 *				  packets every n seconds. N is checked to
 *				  be in the range 2 ... 60 seconds. Defining
 *				  this make _ONLY_ sense if you want to use
 *				  the isdn daemon in fullscreen mode, as this
 *				  packets are only used to update the tx/rx 
 *				  bytecount in fullscreen mode (A side effect
 *				  of this option is an increased loadaverage,
 *				  i have not yet found out why ...).
 *	IPI_DIALMAX=n		- set the number of successive unsuccessfull
 *				  dial attempts which we wait until the send
 *				  packet queue is flushed. defaults to 5.
 *	IPI_IFQ_MAXLEN=n	- the size of the outgoing mbuf queue,
 *				  defaults to IFQ_MAXLEN (=50) if unset.
 *	DEBUG_IPI_VJ		- enable printf statements for VJ debugging.
 *	DEBUG_IPI_DIAL		- enable printf statements for send queue
 *				  full recovery while no connection is active.
 *
 *---------------------------------------------------------------------------*/

#ifdef INET

#include "ipi.h"
#if NIPI > 0

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/kernel.h>
#include <sys/protosw.h>

#include <net/if.h>
#include <net/if_types.h>
#include <net/netisr.h>
#include <net/route.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>

#include "bpfilter.h"
#if NBPFILTER > 0
#include <sys/time.h>
#include <net/bpf.h>
#endif

#if BSD > 199306 || defined(__NetBSD__)
#include <i386/include/bisdn_ioctl.h>
#else
#include <machine/bisdn_ioctl.h>
#endif
#include <bisdn/b_isdn.h>

#if defined(__FreeBSD__) && BSD > 199306
#include "opt_bisdn.h"
#endif

#ifdef IPI_VJ
#include <net/slcompress.h>
#define IPI_COMPRESS IFF_LINK0	/* compress TCP traffic */
#define IPI_AUTOCOMP IFF_LINK1	/* auto-enable TCP compression */
#endif

#if IPI_DIPA
#if IPI_DIPA < 2
#warning "IPI_DIPA defined val < 2, redefining IPI_DIPA to use 2 !"
#undef IPI_DIPA
#define IPI_DIPA 2
#elif IPI_DIPA > 60
#warning "IPI_DIPA defined val > 60, redefining IPI_DIPA to use 60 !"
#undef IPI_DIPA
#define IPI_DIPA 60
#endif
#define IPI_WTIME	IPI_DIPA	/* watchdog timeout */
#endif

#define	IPI_IF_QFULL(ifq)	((ifq)->ifq_len >= ((ifq)->ifq_maxlen - 2))

#ifndef IPI_IFQ_MAXLEN
#define IPI_IFQ_MAXLEN	IFQ_MAXLEN	/* our packet send queue length */
#endif

#if IPI_DIALMAX
#if IPI_DIALMAX < 3 
#warning "IPI_DIALMAX defined val < 3, redefining IPI_DIALMAX to use 3 !"
#undef IPI_DIALMAX
#define IPI_DIALMAX 3
#elif IPI_DIALMAX > IPI_IFQ_MAXLEN
#warning "IPI_DIALMAX defined val > IPI_IFQ_MAXLEN, redefining IPI_DIALMAX to use IPI_IFQ_MAXLEN !"
#undef IPI_DIALMAX
#define IPI_DIALMAX IPI_IFQ_MAXLEN
#endif
#else
#define IPI_DIALMAX 5
#endif

#define	IPI_MTU		1500	/* MTU for IP via ISDN packets	*/

#ifdef __NetBSD__
#if !defined(ifr_mtu)
#define ifr_mtu ifr_ifru.ifru_metric
#endif
#endif /* __NetBSD__ */

struct ipi_softc {
	struct ifnet sc_if;	/* network visible interface	*/
#ifdef __NetBSD__
	int sc_unit;		/* Interface unit number        */
#endif /* __NetBSD__ */
	int sc_appl;		/* ISDN application number	*/
	int sc_isconn;		/* this i/f is connected flag	*/
	int sc_inb;		/* no of bytes came in		*/
	int sc_outb;		/* no of bytes gone out 	*/
	int sc_inpk;		/* no of packets came in	*/
	int sc_outpk;		/* no of packets gone out	*/

#if IPI_DIPA
	int sc_linb;		/* last no of bytes came in	*/
	int sc_loutb;		/* last no of bytes gone out 	*/
#endif	

#ifdef IPI_VJ
	int sc_cinb;		/* no of bytes in compressed	*/
	int sc_coutb;		/* no of bytes out compressed	*/
	struct slcompress sc_compr; /* tcp compression data	*/
	u_char *sc_cbuf;	/* tcp decompression buffer	*/
#endif

#if NBPFILTER > 0
#if BSD > 199306 || defined(__NetBSD__)
	/* don't need it */
#else
	caddr_t sc_bpf;		/* bpf support			*/
#endif
#endif /* NBPFILTER > 0 */

	int sc_dialnum;		/* number of dial req without connection */
	int sc_outstate;	/* state of the output interface side	 */
#define	STATE_IDLE	0	/* interface idle */
#define STATE_DIAL	1	/* interface dialing */
#define STATE_DIALDISC	2	/* interface got disconnect while dialing */
#define STATE_CONN	3	/* interface connected */

} ipi_softc[NIPI];

#ifdef __FreeBSD__
static int ipi_ioctl __P((struct ifnet *ifp, int cmd, caddr_t data));
#endif /* __FreeBSD__ */

#ifdef __NetBSD__
static int ipi_ioctl __P((struct ifnet *ifp, u_long cmd, caddr_t data));
#endif /* __NetBSD__ */

static int ipioutput __P((struct ifnet *ifp, struct mbuf *m,
                          struct sockaddr *dst, struct rtentry *rtp));

#ifdef __FreeBSD__
#define IFP2UNIT(ifp) (ifp)->if_unit
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
#define IFP2UNIT(ifp) ((struct ipi_softc *)ifp->if_softc)->sc_unit
#endif /* __NetBSD__ */

#ifdef __NetBSD__
/* Dummy attachment routine for NetBSD */
void
ipiattach(int unused)
{
}
#endif /* __NetBSD__ */
                         
#if IPI_DIPA
#if BSD > 199306 || defined(__NetBSD__)
void ipi_watchdog __P((struct ifnet *ifp));
#else
void ipi_watchdog __P((int));
#endif
#endif /* IPI_DIPA */

int
ipi_attach(int ifno)
{
	static int next_if = 0;	
	register struct ipi_softc *sc;

	if(next_if >= NIPI)
	{
		return(-1);
	}

	sc = &ipi_softc[next_if];

#ifdef __FreeBSD__
	sc->sc_if.if_unit = next_if;
	sc->sc_if.if_name = "ipi";
#if BSD > 199306
	sc->sc_if.if_softc = sc;
#endif
#endif /* __FreeBSD__ */

#ifdef __NetBSD__
	sprintf(sc->sc_if.if_xname, "ipi%d", next_if);
	sc->sc_if.if_softc = sc;
	sc->sc_unit = next_if;
#endif /* __NetBSD__ */

	sc->sc_if.if_mtu = IPI_MTU;

#ifdef IPI_VJ
	sc->sc_if.if_flags = IFF_SIMPLEX | IFF_POINTOPOINT | IPI_AUTOCOMP;
#else
	sc->sc_if.if_flags = IFF_SIMPLEX | IFF_POINTOPOINT ;
#endif

	sc->sc_if.if_ioctl = ipi_ioctl;
	sc->sc_if.if_output = ipioutput;
	sc->sc_if.if_type = IFT_ISDNBASIC;
	sc->sc_if.if_hdrlen = 0;
	sc->sc_if.if_addrlen = 0;

	sc->sc_if.if_snd.ifq_maxlen = IPI_IFQ_MAXLEN;	/* special length */

#if IPI_DIPA
	sc->sc_if.if_timer = 0;	
	sc->sc_if.if_watchdog = ipi_watchdog;	
	sc->sc_linb = 0;
	sc->sc_loutb = 0;
#endif

	sc->sc_appl = ifno;
	sc->sc_isconn = 0;
	sc->sc_inb = 0;
	sc->sc_outb = 0;
	sc->sc_inpk = 0;
	sc->sc_outpk = 0;

#ifdef IPI_VJ

#ifdef __FreeBSD__
#if BSD > 199306
	sl_compress_init(&sc->sc_compr, -1);
#else
	sl_compress_init(&sc->sc_compr);
#endif
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
	sl_compress_init(&sc->sc_compr, -1);
#endif /* __NetBSD__ */

	if((sc->sc_cbuf = (u_char *)malloc(IPI_MTU+128 , M_DEVBUF, M_WAITOK)) == NULL)
		panic("if_ipi.c, ipi_attach: malloc failed!");
	sc->sc_cinb = 0;
	sc->sc_coutb = 0;
#endif

	sc->sc_dialnum = 0;		/* number of dial attempts */
	sc->sc_outstate = STATE_IDLE;	/* output side state */

	if_attach(&sc->sc_if);

#if NBPFILTER > 0
#ifdef __FreeBSD__
#if BSD > 199306
	bpfattach(&sc->sc_if, DLT_NULL, sizeof(u_int));
#else
	bpfattach(&sc->sc_bpf, &sc->sc_if, DLT_NULL, sizeof(u_int));
#endif /* BSD > 199306 */
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
	bpfattach(&sc->sc_if.if_bpf, &sc->sc_if, DLT_NULL, sizeof(u_int));
#endif
#endif
	return next_if++;
}

/* discard all pending packets from the send queue */

static int
ipiflushoq(struct ifnet *ifp)
{
	register struct mbuf *mm; 
	register int mcnt = 0;
	
	for(;;)
	{
		IF_DEQUEUE(&ifp->if_snd, mm);
		if(mm)
		{
			mcnt++;
			IF_DROP(&ifp->if_snd);
			m_freem(mm);
			ifp->if_oerrors++;
		}
		else
		{
			break;
		}
	}
	return(mcnt);
}

int
ipioutput(struct ifnet *ifp, struct mbuf *m, struct sockaddr *dst,
		struct rtentry *rtp)
{
	register struct ipi_softc *sc = &ipi_softc[IFP2UNIT(ifp)];
	int x;

#ifdef IPI_VJ
	register struct ip *ip;
#if NBPFILTER > 0
	u_char bpfbuf[IPI_MTU];
        register int len = 0;
#endif /* NBPFILTER */
#endif /* IPI_VJ */

	if(dst->sa_family != AF_INET)
	{

#ifdef __FreeBSD__
		printf("ipi%d: af%d not supported\n", sc->sc_if.if_unit, dst->sa_family);
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
		printf("%s: af%d not supported\n", sc->sc_if.if_xname, dst->sa_family);
#endif /* __NetBSD__ */

		m_freem(m);
		sc->sc_if.if_noproto++;
		return EAFNOSUPPORT;
	}

	if(!(ifp->if_flags & IFF_UP))
	{
		m_freem(m);
		return(ENETDOWN);
        }

	sc->sc_outb += m->m_pkthdr.len;

#ifdef IPI_VJ

#if NBPFILTER > 0
#if BSD > 199306 || defined(__NetBSD__)
	if(ifp->if_bpf)
#else
	if(sc->sc_bpf)
#endif
	{
		/*
		 * We need to save the TCP/IP header before it's
		 * compressed.  To avoid complicated code, we just
		 * copy the entire packet into a stack buffer (since
		 * this is a serial line, packets should be short
		 * and/or the copy should be negligible cost compared
		 * to the packet transmission time).
		 */

		register struct mbuf *m1 = m;
		register u_char *cp = bpfbuf;
		u_int af = dst->sa_family;

		/* prepend the address family to bpf buffer */
		bcopy((u_char *)&af, cp, 4);	/* addr family -> buffer */
		cp += 4;	/* advance past address family */
		len = 4;	/* add size of address family */

		do
		{
			register int mlen = m1->m_len;

			bcopy(mtod(m1, caddr_t), cp, mlen);
			cp += mlen;
			len += mlen;
		} while((m1 = m1->m_next) != NULL);
	}
#endif	/* NBPFILTER */

	if((ip = mtod(m, struct ip *))->ip_p == IPPROTO_TCP)
	{
		if(ifp->if_flags & IPI_COMPRESS)
		{
			*mtod(m, u_char *) |= sl_compress_tcp(m, ip,
				&sc->sc_compr, 1);
		}
	}

#if NBPFILTER > 0
#ifdef __FreeBSD__
#if BSD > 199306
	if(ifp->if_bpf)
	{
		bpf_tap(ifp, bpfbuf, len);
	}
#else
	if(sc->sc_bpf)
  	{
		bpf_tap(sc->sc_bpf, bpfbuf, len);
  	}
#endif /* BSD > 199306 */
#endif /* __FreeBSD__ */

#ifdef __NetBSD__
	if(ifp->if_bpf)
	{
		bpf_tap(ifp->if_bpf, bpfbuf, len);
	}
#endif /* __NetBSD__ */
	
#endif /* NBPFILTER */

#else /* ! IPI_VJ */
	
#if NBPFILTER > 0
#if BSD > 199306 || defined(__NetBSD__)
	if(ifp->if_bpf)
#else
	if(sc->sc_bpf)
#endif
	{
		/*
		 * We need to prepend the address family as
		 * a four byte field.  Cons up a dummy header
		 * to pacify bpf.  This is safe because bpf
		 * will only read from the mbuf (i.e., it won't
		 * try to free it or keep a pointer to it).
		 */
		struct mbuf mm;
		u_int af = dst->sa_family;

		mm.m_next = m;
		mm.m_len = 4;
		mm.m_data = (char *)&af;

#ifdef __FreeBSD__
#if BSD > 199306
		bpf_mtap(ifp, &mm);
#else
		bpf_mtap(sc->sc_bpf, &mm);
#endif /* BSD > 199306 */
#endif /* __FreeBSD__ */

#ifdef __NetBSD__
		bpf_mtap(ifp->if_bpf, &mm);
#endif /* __NetBSD__ */
	}
#endif /* NBPFILTER */

#endif /* IPI_VJ */

	sc->sc_if.if_lastchange = time;
	
	x = splimp();

	if(!(ifp->if_flags & IFF_RUNNING)) /* already connected to remote ? */
	{
		isdn_dial(sc->sc_appl);    /* no, instruct daemon to dial */
		
		if(sc->sc_outstate == STATE_IDLE)	/* idle state ? */
		{
			sc->sc_outstate = STATE_DIAL;	/* yes, state=dialing */
		}
		else if((sc->sc_outstate == STATE_DIAL) ||
			(sc->sc_outstate == STATE_DIALDISC))
		{
			sc->sc_dialnum++;		/* no, incr attempts */
		}

#ifdef DEBUG_IPI_DIAL
printf("ipi%d: dial attempt %d\n", sc->sc_if.if_unit, sc->sc_dialnum);
#endif

		if( ((sc->sc_dialnum > IPI_DIALMAX) &&
		     (sc->sc_outstate == STATE_DIALDISC)) ||
		    (IPI_IF_QFULL(&ifp->if_snd)))
		{

			/* if:
			 *	((the max dial attempts are reached AND
			 *	 we go a disconnect while we were in
			 *	 the dialling state) OR (the output
			 *	 packet queue has nearly [!] filled up))
			 */

			int full;

			/* save the queue state */
			full = IPI_IF_QFULL(&ifp->if_snd);

			/* dischard the current packet */
			IF_DROP(&ifp->if_snd);
			m_freem(m);
			ifp->if_oerrors++;
			ifp->if_opackets++; /* keep pkts & errs in sync ! */

			/* discard all pending packets from the send queue */
#ifdef DEBUG_IPI_DIAL
			printf("ipi%d: discharded %d mbufs [disconnected]\n",
				sc->sc_if.if_unit,
				ipiflushoq(ifp));
#else
			ipiflushoq(ifp);
#endif
			sc->sc_dialnum = 0;		/* reset dial counter */
			sc->sc_outstate = STATE_IDLE;	/* reset state */
			splx(x);			/* ! */

			if(full)	/* return proper error code ! */
			{
#ifdef DEBUG_IPI_DIAL
printf("ipi%d: returning ENOBUFS\n", sc->sc_if.if_unit);
#endif
				return(ENOBUFS);
			}
			else
			{
#ifdef DEBUG_IPI_DIAL
printf("ipi%d: returning EHOSTUNREACH\n", sc->sc_if.if_unit);
#endif
				return(EHOSTUNREACH);
			}
		}
	}

	if(IF_QFULL(&ifp->if_snd))
	{

#ifdef DEBUG_IPI_DIAL
printf("ipi%d: queue is full while connected\n", sc->sc_if.if_unit);
#endif

		if(ifp->if_flags & IFF_RUNNING)
		{
			isdn_output(sc->sc_appl);
		}
		else
		{
#ifdef DEBUG_IPI_DIAL
			printf("ipi%d: discharded %d mbufs [connected]\n",
				sc->sc_if.if_unit,
				ipiflushoq(ifp));
#else
			ipiflushoq(ifp);
#endif
		}

		IF_DROP(&ifp->if_snd);
		m_freem(m);
		splx(x);
		ifp->if_oerrors++;
		return(ENOBUFS);
	}
	
	IF_ENQUEUE(&ifp->if_snd, m);

	ifp->if_opackets++;
	ifp->if_obytes += m->m_pkthdr.len;

#ifdef IPI_VJ
	sc->sc_coutb += m->m_pkthdr.len;
#endif

	sc->sc_outpk++;
	
	if(ifp->if_flags & IFF_RUNNING)
		isdn_output(sc->sc_appl);
		
	splx(x);
	return(0);
}

int
ipi_input(int no, struct mbuf *m, int dir)
{
	register struct ipi_softc *sc = &ipi_softc[no];
	int s;
	
#ifdef IPI_VJ
	register int c;
	register int len;
	u_char *cp = sc->sc_cbuf;
#endif

	if(m == 0)
		return(0);

	m->m_pkthdr.rcvif = &sc->sc_if;
	m->m_pkthdr.len = m->m_len;

	sc->sc_if.if_lastchange = time;
	
#ifdef IPI_VJ
	sc->sc_cinb += m->m_pkthdr.len;
	
	if((c = (*(mtod(m, u_char *)) & 0xf0)) != (IPVERSION << 4))
	{
		/* copy data to buffer */

		len = m->m_len;		

/* XXX */	m_copydata(m, 0, len, cp);
		
		if(c & 0x80)
		{
			c = TYPE_COMPRESSED_TCP;
		}
		else if(c == TYPE_UNCOMPRESSED_TCP)
		{
			*cp &= 0x4f;		/* XXX */
		}

		/*
		 * We've got something that's not an IP packet.
		 * If compression is enabled, try to decompress it.
		 * Otherwise, if `auto-enable' compression is on and
		 * it's a reasonable packet, decompress it and then
		 * enable compression.  Otherwise, drop it.
		 */
		if(sc->sc_if.if_flags & IPI_COMPRESS)
		{
			len = sl_uncompress_tcp(&cp,len,(u_int)c,&sc->sc_compr);

			if(len <= 0)
			{
#ifdef DEBUG_IPI_VJ
				printf("if_ipi, ipi_input: len <= 0 IPI_COMPRESS!\n");
#endif
				goto error;
			}
		}
		else if((sc->sc_if.if_flags & IPI_AUTOCOMP) &&
			(c == TYPE_UNCOMPRESSED_TCP) && (len >= 40))
		{
			len = sl_uncompress_tcp(&cp,len,(u_int)c,&sc->sc_compr);

			if(len <= 0)
			{
#ifdef DEBUG_IPI_VJ
				printf("if_ipi, ipi_input: len <= 0 IPI_AUTOCOMP!\n");
#endif
				goto error;
			}

			sc->sc_if.if_flags |= IPI_COMPRESS;
		}
		else
		{
#ifdef DEBUG_IPI_VJ
			printf("if_ipi, ipi_input: invalid ip packet!\n");
#endif

error:
			sc->sc_if.if_collisions++;
			return(0);
		}
/* XXX */	m_copyback(m, 0, len, cp);
	}

#endif /* IPI_VJ */

#if NBPFILTER > 0
#if BSD > 199306 || defined(__NetBSD__)
	if(sc->sc_if.if_bpf)
#else
	if(sc->sc_bpf)
#endif
	{
		/*
		 * We need to prepend the address family as
		 * a four byte field.  Cons up a dummy header
		 * to pacify bpf.  This is safe because bpf
		 * will only read from the mbuf (i.e., it won't
		 * try to free it or keep a pointer to it).
		 */
		struct mbuf mm;
		u_int af = AF_INET;

		mm.m_next = m;
		mm.m_len = 4;
		mm.m_data = (char *)&af;

#ifdef __FreeBSD__
#if BSD > 199306
		bpf_mtap(&sc->sc_if, &mm);
#else
		bpf_mtap(sc->sc_bpf, &mm);
#endif /* BSD > 199306 */
#endif /* __FreeBSD__ */

#ifdef __NetBSD__
		bpf_mtap(sc->sc_if.if_bpf, &mm);
#endif
	}
#endif /* NBPFILTER > 0 */

	s = splimp();

	if(IF_QFULL(&ipintrq))
	{
		IF_DROP(&ipintrq);
		splx(s);
		sc->sc_if.if_ierrors++;
		m_freem(m);
		return(0);
	}
	IF_ENQUEUE(&ipintrq, m);
	splx(s);

	sc->sc_inb += m->m_pkthdr.len;
	sc->sc_inpk += 1;

	sc->sc_if.if_ipackets++;
	sc->sc_if.if_ibytes += m->m_pkthdr.len;
	
	schednetisr(NETISR_IP);
	return(1);
}

void
ipi_connect(int no)
{
	register struct ipi_softc *sc = &ipi_softc[no];
	int s = splimp();

	sc->sc_if.if_flags |= IFF_RUNNING;

	sc->sc_isconn = 1;

	sc->sc_inb = 0;
	sc->sc_outb = 0;
	sc->sc_inpk = 0;
	sc->sc_outpk = 0;
#ifdef IPI_VJ
	sc->sc_cinb = 0;
	sc->sc_coutb = 0;
#endif
#if IPI_DIPA
	sc->sc_linb = 0;
	sc->sc_loutb = 0;
	sc->sc_if.if_timer = IPI_WTIME;	/* activate watchdog timer */
#endif

	sc->sc_dialnum = 0;
	sc->sc_outstate = STATE_CONN;
	splx(s);
}

void
ipi_disconnect(int no)
{
	register struct ipi_softc *sc = &ipi_softc[no];
	int s = splimp();
	
	sc->sc_if.if_flags &= ~IFF_RUNNING;

#if IPI_DIPA
	sc->sc_if.if_timer = 0;		/* de-activate watchdog timer */	
#endif

	if(sc->sc_isconn)
	{
		isdn_ipacct(sc->sc_appl, IPIACCT_FIN, sc->sc_inb, sc->sc_outb,
#ifdef IPI_VJ
				sc->sc_cinb, sc->sc_coutb,
#else
				sc->sc_inb, sc->sc_outb,
#endif
				sc->sc_inpk, sc->sc_outpk);
		sc->sc_isconn = 0;
	}

	if((sc->sc_outstate == STATE_DIAL) ||
	   (sc->sc_outstate == STATE_DIALDISC))
	{
		sc->sc_outstate = STATE_DIALDISC;
	}
	else
	{
		sc->sc_dialnum = 0;
		sc->sc_outstate = STATE_IDLE;
	}
	splx(s);
}

#if IPI_DIPA
void
#if BSD > 199306 || defined(__NetBSD__)
ipi_watchdog(struct ifnet *ifp)
#else
ipi_watchdog(int no)
#endif
{
#if BSD > 199306 || defined(__NetBSD__)
 	register struct ipi_softc *sc = (struct ipi_softc *)ifp->if_softc;	
#else
	register struct ipi_softc *sc = &ipi_softc[no];
#endif

	if((sc->sc_inb != sc->sc_linb) || (sc->sc_outb != sc->sc_loutb)) 
	{
		sc->sc_linb = sc->sc_inb;
		sc->sc_loutb = sc->sc_outb;

		isdn_ipacct(sc->sc_appl, IPIACCT_MID, sc->sc_inb, sc->sc_outb,
#ifdef IPI_VJ
				sc->sc_cinb, sc->sc_coutb,
#else
				sc->sc_inb, sc->sc_outb,
#endif
				sc->sc_inpk, sc->sc_outpk);
	}
	if(sc->sc_isconn)
	{
		sc->sc_if.if_timer = IPI_WTIME;
	}
}
#endif

struct mbuf *
ipi_out(int no)
{
	register struct ipi_softc *sc = &ipi_softc[no];
	struct mbuf    *m;

	if(sc->sc_if.if_flags & IFF_RUNNING)
	{
		IF_DEQUEUE(&sc->sc_if.if_snd, m);
		return(m);
	}
	return(NULL);
}

#ifdef __FreeBSD__
int
ipi_ioctl(register struct ifnet *ifp, int cmd, caddr_t data)
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
int
ipi_ioctl(register struct ifnet *ifp, u_long cmd, caddr_t data)
#endif /* __NetBSD__ */
{
	register struct ipi_softc *sc = &ipi_softc[IFP2UNIT(ifp)];
	struct ifaddr *ifa = (struct ifaddr *) data;
	struct ifreq *ifr = (struct ifreq *) data;
	int x;

	switch(cmd)
	{
	case SIOCSIFDSTADDR:
	case SIOCAIFADDR:
	case SIOCSIFADDR:
		if(ifa->ifa_addr->sa_family != AF_INET)
			return(EAFNOSUPPORT);
		sc->sc_if.if_flags |= IFF_UP;

	/* FALLTHROUGH */

	case SIOCSIFFLAGS:
		x= splimp();
		if((!(ifr->ifr_flags & IFF_UP)) &&
		   (sc->sc_if.if_flags & IFF_RUNNING))
		{
			isdn_disconnect(sc->sc_appl, 0);
			sc->sc_if.if_flags &= ~IFF_RUNNING;
		}
		splx(x);
		break;

	case SIOCGIFMTU:
		ifr->ifr_mtu = sc->sc_if.if_mtu;
		break;

	case SIOCSIFMTU:
		if(ifr->ifr_mtu > 2048)
			return(EINVAL);
		sc->sc_if.if_mtu = ifr->ifr_mtu;
		break;

	default:
		return(EINVAL);
	}
	return(0);
}

#endif /* NIPI > 0 */
#endif /* INET */
