/*
 * Copyright (c) 1995 Dietmar Friede and Juergen Krause.  All rights reserved.
 *
 * Copyright (c) 1996 Arne Helme. All rights reserved.
 *
 * Copyright (c) 1996 Gary Jennejohn. All rights reserved. 
 *
 * Copyright (c) 1995, 1996 Hellmuth Michaelis. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Dietmar Friede,
 *	Juergen Krause, Arne Helme, Gary Jennejohn and Hellmuth Michaelis.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *---------------------------------------------------------------------------
 *
 * A low level driver for the Teles passive ISDN cards and compatibles.
 *
 * It works with the following boards:
 * 
 *	- Teles S0    (8bit)
 *	- Teles S0/16 (16 bit)
 *	- Teles S0/16.3
 *	- Niccy 1008
 *	- Niccy 1016
 *	- Creatix S0  (8bit)
 *	- Creatix S0  (16 bit)
 *
 *---------------------------------------------------------------------------
 *
 *	last edit-date: [Sun Jul  7 18:04:31 1996]
 *
 *	-gj	added code to turn on the clocks in tel_d_stat
 *	-hm	DELAY(45000) -> DELAY(4500)
 *	-hm	fixing splx in tel_out()
 *	-hm	integrating Arne's/Gary's teles.c into plan b isdn
 *	-hm	bugfix for Teles S0/16.3 probe on FreeBSD
 *	-hm	debugging SETUP > 32 bytes
 *	-gj	add handling of edge-triggered interrupts, put in
 *		watchdog for D-channel transmits, delete un-needed trace
 *	-hm	adding Arne's unify patches, adding FreeBSD board type msgs
 *	-gj	added Jan-Hinrich Fessel's patches to support the S0/8 and
 *			S0/16 under NetBSD
 *	-gj	handle Joerg's comments
 *
 *---------------------------------------------------------------------------*/

#include "tel.h"
#if NTEL > 0

#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#ifdef __NetBSD__
#include <sys/device.h>
#endif /* __NetBSD__ */
#include <sys/mbuf.h>

#include <machine/stdarg.h>

#ifdef __FreeBSD__
#include <machine/clock.h>
#include <i386/isa/isa_device.h>
#endif /* __FreeBSD__ */

#ifdef __NetBSD__
#include <machine/cpu.h>
#include <machine/pio.h>
#include <dev/isa/isareg.h>
#include <dev/isa/isavar.h>
#endif /* __NetBSD__ */

#if BSD > 199306 || defined (__NetBSD__)
#include <i386/include/bisdn_ioctl.h>
#else
#include <machine/bisdn_ioctl.h>
#endif

#include <bisdn/b_isdn.h>
#include <bisdn/b_isdnlap.h>
#include <bisdn/b_disdn.h>

#define	CMDR	0x21
#define	EXIR	0x24
#define	ISTA	0x20
#define	MASK	0x20
#define	RBCL	0x25
#define	STAR	0x21
#define RSTA	0x27
#define CIXR	0x31
#define CIX0	0x31

#define	IRBCH	0x2a
#define	HRBCH	0x2d
#define SEC_DELAY	1000000

#define FIFOEMPTY	1
#define STARTFRAME	2
#define BADIFRAME	0x10

#define min(a,b)        ((a)<(b)?(a):(b))
#define C_CHAN(x)       ((x)&1)

#define L1_RUNNING	1
#define WANT_L1		2
#define TEL_OUT		4
#define WANT_OUT	8

#ifdef __FreeBSD__
#define TIMEOUT_FUNC_T	timeout_func_t
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
#define TIMEOUT_FUNC_T	void *
#endif /* __NetBSD__ */

extern int      ispy_applnr;

#ifdef __FreeBSD__
void            telintr __P((int));
int             telprobe __P((struct isa_device*));
int		telattach __P((struct isa_device*));
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
int		telintr __P((void *));
int		telprobe __P((struct device *, void *, void *));
void		telattach __P((struct device *, struct device *, void *));
#endif /* __NetBSD__ */
u_char		telget __P((u_char*, u_int));
int		inSTAR __P((u_char*, u_int));
static void	telput __P((u_char*, u_int, u_int));
static void	telput_pio __P((u_char *, u_int, u_int));
static u_char	telget_pio __P((u_char *, u_int));
void		telgetfifo_pio __P((void *, const void *, size_t));
void		telputfifo_pio __P((void *, const void *, size_t));
static void	outCMD_pio __P((u_char *, u_char));
static int	inSTAR_pio __P((u_char *, u_int));
void  memcpyb __P((void *to, const void *from, size_t len));

typedef struct
{
	u_short		c_ctrl;
	caddr_t		hscx;
	u_char		hscx_mask;
	u_short		service;
	u_short		c_state;
	u_char		crc_byte;
	u_char		cmd;
	struct	mbuf	*ibuf;
	u_short		ilen;
	u_char 		*ib;
	struct	mbuf	*obuf, *oact;
} chan_t;

#ifdef __FreeBSD__
static int tel_d_out __P((int, struct mbuf *));
static int tel_d_stat __P((int));
static int tel_init __P((int, int, int, int));
static int tel_out __P((int, int, struct mbuf *));
static int tel_setstate __P((int, int, u_char*));
static int tel_getstate __P((int));

struct disdn_ll_func sc_llif=
{
	tel_d_out, tel_d_stat, tel_init, tel_out, tel_setstate, tel_getstate
};
#endif /* __FreeBSD__ */

struct tel_softc
{
#ifdef __FreeBSD__
	u_short		sc_unit;
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
	struct device 	sc_dev;		/* base device */
	void   		*sc_ih;		/* interrupt vectoring */
#endif /* __NetBSD__ */
	u_short		sc_ctrl;
	u_short         sc_vector;	/* interrupt vector 		 */
	u_short         sc_port;
	u_char		sc_typ;
#define TEL_TYPE_UNK	0x00
#define TEL_TYPE_16	0x16
#define TEL_TYPE_16_3	0x19
#define TEL_TYPE_1000	0x10
	u_char		sc_spy;
	u_short		sc_state;
	caddr_t  sc_vmem_addr;	/* card RAM virtual memory base */
	caddr_t  Isac;
	u_char	Isac_mask;
	chan_t          sc_chan[2];
	struct	mbuf	*sc_ibuf;
	u_short		sc_ilen;
	u_char		*sc_ib;
	struct	mbuf	*sc_obuf;
	u_char		*sc_op;
	u_short		sc_ol;

	u_char		(*get)(u_char *, u_int);
	void		(*put)(u_char *, u_int, u_int);
	void		(*getfifo)(void *, const void *, size_t);
	void		(*putfifo)(void *, const void *, size_t);
	void		(*outCMD)(u_char *, u_char);
	int		(*inSTAR)(u_char *, u_int);
#ifdef __FreeBSD__
}               tel_sc[NTEL];
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
};
#endif /* __NetBSD__ */

/* watchdog for XMIT */
#ifdef __FreeBSD__
void  tel_d_wd __P((int));
#endif
#ifdef __NetBSD__
void  tel_d_wd __P((struct tel_softc*));
#endif

#define isac	sc->Isac
#define isac_mask	sc->Isac_mask
#define hscx_a	sc->sc_chan[0].hscx
#define hscx_a_mask	sc->sc_chan[0].hscx_mask
#define hscx_b	sc->sc_chan[1].hscx
#define hscx_b_mask	sc->sc_chan[1].hscx_mask
#ifdef __NetBSD__
#define sc_unit sc_dev.dv_unit
#endif /* __NetBSD__ */

#ifdef __NetBSD__
static int tel_d_out __P((struct tel_softc *, struct mbuf *));
static int tel_d_stat __P((struct tel_softc *));
static int tel_init __P((struct tel_softc *, int, int, int));
static int tel_out __P((struct tel_softc *, int, struct mbuf *));
static int tel_setstate __P((struct tel_softc *, int, u_char *));
static int tel_getstate __P((struct tel_softc *));

struct disdn_ll_func sc_llif=
{
	(int (*)(struct device *, struct mbuf *))	tel_d_out, 
	(int (*)(struct device *))			tel_d_stat, 
	(int (*)(struct device *, int, int, int))	tel_init, 
	(int (*)(struct device *, int, struct mbuf *))	tel_out, 
	(int (*)(struct device *, int, u_char *))	tel_setstate, 
	(int (*)(struct device *))			tel_getstate
};
#endif /* __NetBSD */

#ifdef __FreeBSD__
struct isa_driver teldriver = {
	telprobe,
	telattach,
	"tel",
};
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
struct cfattach tel_ca = {
	sizeof(struct tel_softc), telprobe, telattach,
};

struct cfdriver tel_cd = {
	NULL, "tel", DV_IFNET
};
#endif /* __NetBSD__ */

/*
 * S0/16 specific functions:
 */
static void
telput(u_char * a, u_int i, u_int v)
{
	if (i & 1)
		i |= 0x200;
	a[i] = v;
}

u_char
telget(u_char * a, u_int i)
{
	if (i & 1)
		i |= 0x200;
	return (a[i]);
}

static void
outCMD(u_char * s, u_char v)
{
	u_short i = 10;

	if(v == 0)
		return;

	while ((s[0x220] & 4) && i) /* STAR */
	{
		DELAY(10);
		i--;		/* CEC */
	}

	if (i)
	{
		s[0x220]= v;	/* CMDR */
		return;
	}
#ifdef DEBUG
	printf("CMDCEC%x", telget(s, STAR));
#endif
}

int
inSTAR(u_char * s, u_int w)
{
	u_char r;
	if (((r=s[0x220])&4)==0)
		return(r);
	return(4);
}

/*
 * S0/16.3 specific functions for polled I/O access:
 */
static void
telput_pio(u_char *base, u_int offset, u_int v)
{
	outb((int)base + offset, (u_char)v);
}

static u_char
telget_pio(u_char *base, u_int offset)
{
	return (inb((int)base + offset));
}

void		
telgetfifo_pio(void *buf, const void *base, size_t len)
{
	insb((int)base + 0x3e, (u_char *)buf, (u_int)len);
}

void
telputfifo_pio(void *base, const void *buf, size_t len)
{
	outsb((int)base + 0x3e, (u_char *)buf, (u_int)len);
}

static void
outCMD_pio(u_char *base, u_char v)
{
	u_short i = 10;

	if (v == 0)
		return;

	while ((inb((int)base + STAR) & 4) && i) 
	{
	        DELAY(10);
		i--;	/* CEC */
        }
	if (i) 
	{
	        outb((int)base + CMDR, v);
		return;
	}	
#ifdef DEBUG
	printf("CMDCEC%x", telget_pio(base, STAR));
#endif
}

static int
inSTAR_pio(u_char *base, u_int w)
{
	u_char  r;

	if ((r = inb((int)base + STAR) & 4) == 0)	/* STAR */
		return (r);
	return (4);
}

static int
#ifdef __FreeBSD__
tel_d_stat(int unit)
{
	struct tel_softc *sc = &tel_sc[unit];
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
tel_d_stat(struct tel_softc *sc)
{
#endif /* __NetBSD__ */

	u_char xfw = 0, xac;

	/* Transmit FIFO Write Enable - can write bytes */
	xfw = ((*sc->get)(isac, STAR));
	/* Transmitter Active */
	xac = ((*sc->get)(isac, IRBCH));

	sc->sc_state |= WANT_OUT;
	if(!(sc->sc_state & L1_RUNNING))
	{
		if(sc->sc_state&WANT_L1)
		{
			return(1);
		}
		sc->sc_state|= WANT_L1;
		/*
		 * make sure the clocks are running ! This is
		 * very important because the ISAC disables its
		 * clocks after receiving a DISC from the network,
		 * even though it's set up so that the clocks should never
		 * be disabled.
		 */
		if (sc->sc_typ != TEL_TYPE_16_3)
		{
			(*sc->put)(isac,CIXR,0x00);
			/* the book says it can take up to 4ms to
                           start the clocks */
			DELAY(4500); /* 4.5 ms delay */
			(*sc->put)(isac,CIXR,0x27);
		}
		else
		{
			/* In IOM-2 mode we do it like this: */
			(*sc->put)(isac, CIX0, 0x03);  /* TIM for IOM-2 */
			/* the book says it can take up to 4ms to
                           start the clocks */
			DELAY(4500); /* 4.5 ms delay */

		}
		return(1);
	}
	if (sc->sc_state & TEL_OUT)
	{
		/* this says:
		 *	a) the XMIT buffer can accept more bytes
		 *	b) the XMITter is not active
		 */
		if ((xfw & 0x4a) && !(xac & 0x80))
		{
			/* it appears that the command sometimes */
			/* doesn't "take". Send it again */
			if (sc->sc_op)
				(*sc->put)(isac, CMDR, 0x08); /* more later */
			else
				(*sc->put)(isac, CMDR, 0x0a); /* that's all */

			/* since 32 bytes should be xmitted in ~ 4 ms, wait */
			/* for 1 tick */
#ifdef __FreeBSD__
			untimeout((TIMEOUT_FUNC_T) tel_d_wd, (caddr_t) unit);
			timeout((TIMEOUT_FUNC_T) tel_d_wd, (caddr_t) unit, 1);
#endif
#ifdef __NetBSD__
			untimeout((TIMEOUT_FUNC_T) tel_d_wd, (caddr_t) sc);
			timeout((TIMEOUT_FUNC_T) tel_d_wd, (caddr_t) sc, 1);
#endif
		}
		return(2);
	}
	if (!(xfw & 0x40))
	{
		return (4);
	}
	sc->sc_state &= ~WANT_OUT;
	return (0);
}

static int
#ifdef __FreeBSD__
tel_d_out(int unit, struct mbuf * m)
{
	struct tel_softc *sc = &tel_sc[unit];
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
tel_d_out(struct tel_softc *sc, struct mbuf * m)
{
#endif /* __NetBSD__ */

	int             l;
	u_char         *b = m->m_data;

#ifdef __FreeBSD__
	if ((l=tel_d_stat(unit)) || (m == NULL))
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
	if ((l=tel_d_stat(sc)) || (m == NULL))
#endif /* __NetBSD__ */
		return (0);

	sc->sc_state|= TEL_OUT;
	sc->sc_obuf= NULL;

	/* if tracing, output packet to trace device */

	if(sc->sc_spy & 1)
	{
#ifdef __FreeBSD__
		isdn_binput(ispy_applnr,m->m_len,b,(unit << 2) | 1);
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
		isdn_binput(ispy_applnr,m->m_len,b,(sc->sc_unit << 2) | 1);
#endif /* __NetBSD__ */
	}
	
	l = min(m->m_len, 32);	/* set length, maximum size = 32 bytes */

#ifdef DEBUG
printf("tel_d_out: data %d at %x...\n", l, b);
#endif

	(*sc->putfifo)(isac,b,l);		/* output to tx fifo */

	if (m->m_len > 32)			/* message > 32 bytes ? */
	{
		sc->sc_op= b+32;		/* ptr + 32 for irq routine */
		sc->sc_ol= m->m_len-32;		/* length - 32 for irq rout */
#ifdef DEBUG
printf("tel_d_out: more data %d at %x...\n", sc->sc_ol, sc->sc_op);
#endif
		(*sc->put)(isac, CMDR, 0x08);	/* set XTF */
		sc->sc_obuf= m;
		return (32);
	}

	sc->sc_op= NULL;
	(*sc->put)(isac, CMDR, 0x0a);	/* set XTF and XME (end of frame) */

	/* since 32 bytes should be xmitted in ~ 4 ms, wait for 1 tick */

#ifdef __FreeBSD__
	timeout((TIMEOUT_FUNC_T) tel_d_wd, (caddr_t) unit, 1);
#endif
#ifdef __NetBSD__
	timeout((TIMEOUT_FUNC_T) tel_d_wd, (caddr_t) sc, 1);
#endif
  	return (l);
}

static int
tel_init_16_3(struct tel_softc *sc, int c, int ichan, int service)
{
	chan_t *chan = &sc->sc_chan[c];
	caddr_t hscx = chan->hscx;

	if (chan->obuf)
		m_freem(chan->obuf);
	if (chan->ibuf)
		m_freem(chan->ibuf);
	chan->ib = NULL;
	chan->ilen = 0;
	chan->oact = chan->obuf = chan->ibuf = NULL;
	chan->service = service;

	/* Power Up, Clock Mode 5 */
	(*sc->put)(hscx, 0x2f, 0x85);	/* CCR1 */
	 /* Transmit Address Byte 1 */
	(*sc->put)(hscx, 0x24, 0xff);	/* XAD1 */
	 /* Transmit Address Byte 2 */
	(*sc->put)(hscx, 0x25, 0xff);	/* XAD2 */
	 /* Receive Address Byte High Reg. 2 */
	(*sc->put)(hscx, 0x27, 0xff);	/* RAH2 */
	 /* Transmit Byte Count High */
	(*sc->put)(hscx, 0x2d, 0x00);	/* XBCH */
	 /* Receive Length Check Register */
	(*sc->put)(hscx, 0x2e, 0x00);	/* RLCR */

	/*
	 * B channel setup.  In IOM-2 mode it is asymmetric.
	 */
	switch (ichan) 
	{
	case 0:
		/* Prepare HSCX for B channel 1 */
		(*sc->put)(hscx, 0x2c, 0x30);	/* CCR2 */
	        (*sc->put)(hscx, 0x30, 0x2f);	/* TSAX */
		(*sc->put)(hscx, 0x31, 0x2f);	/* TSAR */
		(*sc->put)(hscx, 0x32, 0x07);	/* XCCR */
		(*sc->put)(hscx, 0x33, 0x07);	/* RCCR */
		break;
	case 1:
		/* Prepare HSCX for B channel 2 */
		(*sc->put)(hscx, 0x2c, 0x30); 	/* CCR2 */
		(*sc->put)(hscx, 0x30, 0x03); 	/* TSAX */
		(*sc->put)(hscx, 0x31, 0x03);	/* TSAR */ 
		(*sc->put)(hscx, 0x32, 0x07);	/* XCCR */
		(*sc->put)(hscx, 0x33, 0x07);	/* RCCR */
		break;
	case -1:
	default:
		/* Disable B channel for HSCX */
		(*sc->put)(hscx, 0x2c, 0x30);	/* CCR2 */
		(*sc->put)(hscx, 0x30, 0xff);	/* TSAX */
		(*sc->put)(hscx, 0x31, 0xff);	/* TSAR */
		(*sc->put)(hscx, 0x32, 0x07);	/* XCCR */
		(*sc->put)(hscx, 0x33, 0x07);	/* RCCR */
	}
	/*
	 * Application specific setup
	 */
	chan->crc_byte = 0;
	switch (service & 0xf) 
	{
	case 1:		/* Raw Telephony */
		(*sc->put)(hscx, 0x22, 0xe4);	/* MODE */
		break;
	case 2:		/* HDLC Frames */
		chan->crc_byte = 1;
		(*sc->put)(hscx, 0x22, 0x8c);	/* MODE */
		break;
	case 7:		/* HDLC Spying */
		chan->crc_byte = 1;
		(*sc->put)(hscx, 0x22, 0x8c);	/* MODE */
		chan->hscx_mask = 0x18;
		(*sc->put)(hscx, 0x20, 0x18);	/* MASK */
		(*sc->outCMD)(hscx, 0x41);
		return 0;
	default: /* Disable */
		(*sc->put)(hscx, 0x22, 0x84);	/* MODE */
		return (0);
	}
	chan->hscx_mask = 0x00;
	(*sc->put)(hscx, 0x20, 0);	/* MASK Allow interrupts */
	(*sc->outCMD)(hscx, 0x41);
	(*sc->put)(hscx, 0x20, 0x00);	/* ISTA */
	return (0);
}

static int
#ifdef __FreeBSD__
tel_init(int unit, int c, int ichan, int service)
{
	struct tel_softc *sc = &tel_sc[unit];
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
tel_init(struct tel_softc *sc, int c, int ichan, int service)
{
#endif /* __NetBSD__ */
	chan_t         *chan = &sc->sc_chan[c];
	caddr_t         hscx= chan->hscx;

	if (sc->sc_typ == TEL_TYPE_16_3)
	{
		return (tel_init_16_3(sc, c, ichan, service));
	}

	if(chan->obuf)
		m_freem(chan->obuf);
	if(chan->ibuf)
		m_freem(chan->ibuf);
	chan->ib= NULL;
	chan->ilen= 0;
	chan->oact= chan->obuf= chan->ibuf= NULL;
	chan->service= service;
	
	(*sc->put)(hscx, 0x20, 0xff);	/* MASK */
	(*sc->put)(hscx, 0x2f, 0x81);	/* CCR1 */
	(*sc->put)(hscx, 0x2c, 0);		/* CCR2 */
	(*sc->put)(hscx, 0x22, 0x84);	/* MODE */
	(*sc->put)(hscx, 0x2d, 0);		/* XBCH */
	(*sc->put)(hscx, 0x2e, 0);		/* RLCR */
	(*sc->put)(hscx, 0x22, 0x8c);	/* MODE */
	
	chan->crc_byte= 0;

	switch(service&0xf)
	{
	case 1: /* Raw Telephony */
		(*sc->put)(hscx, 0x24, 0xff);	/* XAD1 */
		(*sc->put)(hscx, 0x25, 0xff);	/* XAD2 */
		(*sc->put)(hscx, 0x27, 0xff);	/* RAH2 */
		(*sc->put)(hscx, 0x22, 0xe4);	/* MODE */
		(*sc->put)(hscx, 0x2f, 0x81);	/* CCR1 */
		break;
	case 2: /* HDLC Frames */
		chan->crc_byte= 1;
		(*sc->put)(hscx, 0x22, 0x8c);	/* MODE */
		(*sc->put)(hscx, 0x2f, 0x89);	/* CCR1 */
		break;
	case 7: /* HDLC Spying */
		chan->crc_byte= 1;
		(*sc->put)(hscx, 0x22, 0x8c);	/* MODE */
		(*sc->put)(hscx, 0x2f, 0x81);	/* CCR1 */
		chan->hscx_mask = 0x18;
		(*sc->put)(hscx, 0x20, 0x18);	/* MASK */
		(*sc->outCMD)(hscx, 0x41);
		return(0);
	default: /* disable */
		(*sc->put)(hscx, 0x22, 0x84);	/* MODE */
		(*sc->put)(hscx, 0x2f, 0x81);	/* CCR1 */
		chan->hscx_mask = 0xf8;
		(*sc->put)(hscx, 0x20, 0xf8);	/* MASK */
		(*sc->outCMD)(hscx, 0xe3);
		return(0);
	}
	chan->hscx_mask = 0;
	(*sc->put)(hscx, 0x20, 0);	/* MASK */
	(*sc->outCMD)(hscx, 0x41);
	return(0);
}

static void
#ifdef __FreeBSD__
tel_start(int unit, int c)
{
	struct tel_softc *sc = &tel_sc[unit];
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
tel_start(struct tel_softc *sc, int c)
{
#endif /* __NetBSD__ */
	chan_t         *chan = &sc->sc_chan[c];
	caddr_t         hscx= chan->hscx;
	int             l= 0;
	struct	mbuf	*m = chan->oact;

	if(m == NULL)
	{
		chan->c_state |= FIFOEMPTY | STARTFRAME;
#ifdef __FreeBSD__
		disdn_startout(unit,c);
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
		disdn_startout(sc->sc_unit,c);
#endif /* __NetBSD__ */
		return;
	}

	chan->c_state &= ~FIFOEMPTY;

	while((l < 32) && m )
	{
		for (; (l < 32) && (m->m_len > 0); l++, m->m_data++, m->m_len--)
		{
			if (sc->sc_typ != TEL_TYPE_16_3)
			{
				*hscx = *m->m_data;
			}
			else
			{
				outb((int)hscx + 0x3e, *m->m_data);
			}
		}
		if(m->m_len == 0)
			m= m->m_next;
	}
	chan->oact= m;

	if(m)
	{
		chan->c_state &= ~STARTFRAME;
		(*sc->outCMD)(hscx, chan->cmd | 8);
		chan->cmd= 0;
		return;
	}

	chan->c_state |= STARTFRAME;
	if(chan->service == TELEFON)
		(*sc->outCMD)(hscx, chan->cmd | 8);
	else
		(*sc->outCMD)(hscx, chan->cmd | 0xa);
	chan->cmd= 0;
	if(!(chan->obuf->m_flags & 0x10))
		m_freem(chan->obuf);
	chan->obuf= NULL;
#ifdef __FreeBSD__
	disdn_startout(unit,c);
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
	disdn_startout(sc->sc_unit,c);
#endif /* __NetBSD__ */
	return;
}

static int
#ifdef __FreeBSD__
tel_out(int unit, int c, struct mbuf * m)
{
	struct tel_softc *sc = &tel_sc[unit];
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
tel_out(struct tel_softc *sc, int c, struct mbuf * m)
{
#endif /* __NetBSD__ */
	chan_t         *chan = &sc->sc_chan[c];
	int x;

	x= splimp();
	if(chan->obuf)
	{
		splx(x);
		return(0);
	}
	if(sc->sc_spy & 2)
	{
#ifdef __FreeBSD__
		isdn_binput(ispy_applnr,m->m_len,m->m_data,(c << 6) | (unit << 2) | 3);
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
		isdn_binput(ispy_applnr,m->m_len,m->m_data,(c << 6) | (sc->sc_unit << 2) | 3);
#endif /* __NetBSD__ */
	}
	chan->oact= chan->obuf= m;
	chan->c_state |= STARTFRAME;
	if((chan->c_state & FIFOEMPTY) || ((*sc->inSTAR)(chan->hscx,1) & 0x40))
	{
#ifdef __FreeBSD__
		tel_start(unit,c);
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
		tel_start(sc,c);
#endif /* __NetBSD__ */
	}
	splx(x);
	return(1);
}

static void
initISAC(struct tel_softc *sc, u_char * is)
{
	if (sc->sc_typ != TEL_TYPE_16_3)
	{
		/* This is more or less magic */
		/* To understand read the ISAC manuals from Siemens */
		/* The code sets the ISAC to IOM-1 ... */
		/* Teles implements some sort of IOM-1 Bus on their cards */
		sc->Isac_mask = 0x2e;
		(*sc->put)(is, 0x20, 0x2e);	/* MASK */
		(*sc->put)(is, 0x22, 0xc1);	/* MODE */
		(*sc->put)(is, 0x31, 0xbc);	/* CIXR */
		(*sc->put)(is, 0x37, 0x70);	/* STCR */
		(*sc->put)(is, 0x38, 4);	/* ADF1 */
		(*sc->put)(is, 0x21, 0x41);	/* CMDR */
		(*sc->put)(is, 0x22, 0xc9);	/* MODE */
		(*sc->put)(is, 0x30, 0xa);	/* SPCR */
	}
	else
	{
		/* This is hopefully not that magic.
		 * The S0/16.3 card uses the IOM-2 bus to connect the
		 * ISAC and HSCX.  The following configures IOM-2:
		 */
		 /* Disable all interrupts */
		sc->Isac_mask = 0xff;
		(*sc->put)(is, 0x20, 0xff);	/* MASK */
		/* Select IOM-2 */		
		(*sc->put)(is, 0x39, 0x80); /* ADF2 */
		/* CI1E:CIC1 enabled, set S/Q bits */
		(*sc->put)(is, 0x3b, 0x2f); /* SQXR */
		/* Terminal mode, B1 + C1 and B2 + IC2 monitoring */
		(*sc->put)(is, 0x30, 0x00); /* SPCR */
		/* Select IOM channel 1 */
		(*sc->put)(is, 0x38, 0x02); /* ADF1 */
		/* Select TIC Bus Address 7 */
		(*sc->put)(is, 0x37, 0x70); /* STCR */
		/* Transparent Mode 2, Receiver Active, DIM= 0x1 */
		(*sc->put)(is, 0x22, 0xc9); /* MODE */
		 /* Timer "reset" */
		(*sc->put)(is, 0x23, 0x00); /* TIMR */
		/* Select IOM channel 0 */		
		(*sc->put)(is, 0x38, 0x00); /* ADF1 */
		/* RX & TX reset */
		(*sc->outCMD)(is, 0x41);
		 /* CIX res, CODX0= 0x1 */
		(*sc->put)(is, 0x31, (1 << 2) | 3); /* CIX0 */
	}
}

static int
#ifdef __FreeBSD__
telinit(int unit)
{
	struct tel_softc *sc = &tel_sc[unit];
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
telinit(struct tel_softc *sc)
{
#endif /* __NetBSD__ */
	int             i;

	if (sc->sc_typ != TEL_TYPE_16_3)
	{
		sc->sc_vmem_addr[0x80] = 1;
	}
	DELAY(SEC_DELAY / 5);
	initISAC(sc, isac);
	for (i = 0; i < 10; i++)
	{
		u_char          x, c, e;
		c = 0;
		e = 0;
		DELAY(SEC_DELAY / 5);
		if((x = (*sc->get)(isac, 0x20)) != 0)
		{
			if (x & 0xa0)
				c |= 0x80;
			if (x & 4)
			{
				u_char ci;
				ci= (*sc->get)(isac, 0x31 /* CIRR */ );
				(*sc->get)(isac, 0x3b /* SQRR */ );

				if(ci &2)
				{
					ci= (ci >> 2) & 0xf;
					sc->sc_state= 0;
					switch(ci)
					{
					case 0xc:
					case 0xd:
						sc->sc_state= L1_RUNNING;
					}
				}

			}
			if (x & 1)
			{
				e = (*sc->get)(isac, 0x24);
				if (e & 2)
					return (1);
				if (e & 0x10)
				{
					c |= 0xc0;
				}
			}
			if (c)
				(*sc->put)(isac, 0x21, c);
			sc->Isac_mask = 0;
			(*sc->put)(isac, 0x20, 0);
		}
		if (sc->sc_typ == TEL_TYPE_16_3)
		{
			/*
			 * Issue an explicit "Activate Request" command to 
			 * transfer the 2B+D channels to the IOM-2 interface.
			 */
			(*sc->put)(isac, CIX0, 0x27);
		}
	}

#ifdef __FreeBSD__
	tel_init(unit,0,-1,0);
	tel_init(unit,1,-1,0);
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
	tel_init(sc,0,-1,0);
	tel_init(sc,1,-1,0);
#endif /* __NetBSD__ */
	return (0);
}


#ifdef __FreeBSD__
int
telprobe(struct isa_device * is)
{
	struct tel_softc *sc = &tel_sc[is->id_unit];
	u_char intr_no[16] = {1, 1, 0, 2, 4, 6, 1, 1, 1, 0, 8, 10, 12, 1, 1, 14};
	u_char	cfval;

	/* Set up the softc structure with card specific info. */
	sc->sc_vector = is->id_irq;
	sc->sc_port = is->id_iobase;
	sc->sc_unit = is->id_unit;
	sc->sc_vmem_addr = (caddr_t) is->id_maddr;

	/* 
	 * Assumption: The value of sc->sc_vmem_addr is zero if a
	 * memory address * has not bee defined in the kernel
	 * configuration file.  This must * be the case for Teles
	 * S0/16.3 cards.  XXX Arne
	 */
	if (sc->sc_vmem_addr)
	{
		/* Initially we set up the standard S0/16 functions here */
		sc->get = telget;
		sc->put = telput;
		sc->getfifo = memcpyb;
		sc->putfifo = memcpyb;
		sc->outCMD = outCMD;
		sc->inSTAR = inSTAR;
		sc->sc_typ= TEL_TYPE_1000;
		isac = (caddr_t) ((is->id_maddr) + 0x100);
		hscx_a = (caddr_t) ((is->id_maddr) + 0x180);
		hscx_b = (caddr_t) ((is->id_maddr) + 0x1c0);

		if (is->id_iobase)
		{
			unsigned long   val;
			val = inb(sc->sc_port) << 16;
			val |= inb(sc->sc_port + 1) << 8;
			val |= inb(sc->sc_port + 2);
			if (val == 0x51931e)
			{
				val = intr_no[ffs(sc->sc_vector) - 1];
				if (val == 1)
					return (0);
				val |= ((u_long) sc->sc_vmem_addr) >> 9;
				DELAY(SEC_DELAY / 10);
				outb(sc->sc_port + 4, val);
				DELAY(SEC_DELAY / 10);
				outb(sc->sc_port + 4, val | 1);
				DELAY(SEC_DELAY / 5);
				sc->sc_typ= TEL_TYPE_16;
			}
		}

		/* Toggle card bit off */
		sc->sc_vmem_addr[0x80] = 0;
		DELAY(SEC_DELAY / 5);
		sc->sc_vmem_addr[0x80] = 1;
		DELAY(SEC_DELAY / 5);

		if ((0x48 == ((*sc->get)(isac, 0x21) & 0xfd)) || 
		    (0x48 == ((*sc->get)(hscx_a, 0x21) & 0xfd)) || 
		    (0x48 == ((*sc->get)(hscx_b, 0x21) & 0xfd)))
		{
			is->id_msize = 0x1000;
			sc->sc_vmem_addr[0x80] = 0;
			return (1);
		}
	}
	else
	{
		/* Teles 16.3 card initialization */
	        switch (is->id_iobase) 
		{
			case 0xd80:
		        	isac = (caddr_t) 0x960;
				hscx_a = (caddr_t) 0x160;
				hscx_b = (caddr_t) 0x560;
				break;
			case 0xe80:
			case 0xf80:
			default:
				/* Don't know these values yet */
				sc->sc_typ = TEL_TYPE_UNK;
				printf("Teles S0/16.3 no I/O ports?!\n");
				return (0);
	        }
		/* Initially we set up the standard S0/16.3 functions here */
		sc->get = telget_pio;
		sc->put = telput_pio;
		sc->getfifo = telgetfifo_pio;
		sc->putfifo = telputfifo_pio;
		sc->outCMD = outCMD_pio;
		sc->inSTAR = inSTAR_pio;
		sc->sc_typ = TEL_TYPE_16_3; 

		/* 
		 * Read HSCX A/B VSTR.  Expected value for the S0/16.3 card is
		 * 0x05 in the least significant bits.
		 */
		if (!((*sc->get)(hscx_a, 0x2e) & 0xf == 0x5) && 
		    ((*sc->get)(hscx_b, 0x2e) & 0xf == 0x5))
			return (0);

		if( (cfval = intr_no[ffs(sc->sc_vector) - 1]) != 0x01 )
		{
			DELAY(SEC_DELAY / 10);	
			outb(is->id_iobase + 4, cfval);
			DELAY(SEC_DELAY / 10);	
			outb(is->id_iobase + 4, cfval | 1);
			DELAY(SEC_DELAY / 10);	
		} 
		else 
		{
			printf("Teles S0/16.3: IRQ %d not supported!\n", 
			       is->id_irq);
			return (0);
		}
		is->id_msize = 0;
		return (1);
	}
	sc->sc_typ= TEL_TYPE_UNK;
	return (0);
}

int
telattach(struct isa_device * is)
{
	struct tel_softc *sc = &tel_sc[is->id_unit];
	int cn;

	cn = 10;
	do
	{
		if (sc->sc_typ != TEL_TYPE_16_3)
		{
			sc->sc_vmem_addr[0x80] = 0;
		}
		DELAY(SEC_DELAY / 5);
	} while (telinit(sc->sc_unit) && cn--);

	sc->sc_ctrl = -1;
	if ((cn = disdn_attach(sc->sc_unit,&sc_llif,sc->sc_typ,2)) == -1)
	{
		return (0);
	}
	sc->sc_ctrl = (cn >> 8) & 0xff;
	cn &= 0xff;
	sc->sc_chan[0].c_ctrl= cn;
	sc->sc_chan[1].c_ctrl= cn+1;
	sc->sc_spy= 0;

	printf("tel%d: card type ", is->id_unit);
	
	if (sc->sc_typ == TEL_TYPE_16_3)
	{
		printf("Teles S0/16.3\n");
	}
	else
	{
		printf("Teles S0 or Creatix S0\n");
	}

	return (1);
}
#endif /* __FreeBSD__ */

#ifdef __NetBSD__
int
telprobe(parent, match, aux)
	struct device *parent;
	void   *match, *aux;
{
	struct tel_softc *sc = match;
	struct isa_attach_args *ia = aux;
	u_char intr_no[16] = {0x01, 0x01, 0x00, 0x02, 0x04, 0x06, 0x01, 0x01,
			      0x01, 0x01, 0x08, 0x0a, 0x0c, 0x01, 0x01, 0x0e};
	u_char cfval;

	/* Set up the softc structure with card specific info. */
	sc->sc_vector = ia->ia_irq;
	sc->sc_port = ia->ia_iobase;
	sc->sc_vmem_addr = (caddr_t) ISA_HOLE_VADDR(ia->ia_maddr);

	if (ia->ia_maddr == MADDRUNK) 
	{ /* Assume Teles S0/16.3 card */
	        switch (ia->ia_iobase) 
		{
			case 0xd80:
		        	isac = (caddr_t) 0x960;
				hscx_a = (caddr_t) 0x160;
				hscx_b = (caddr_t) 0x560;
				break;
			case 0xe80:
		        	isac = (caddr_t) 0xa60;
				hscx_a = (caddr_t) 0x260;
				hscx_b = (caddr_t) 0x660;
				printf("Teles S0/16.3 I/O-Ports in unconfirmed adress-space!\n");
				break;
			case 0xf80:
		        	isac = (caddr_t) 0xb60;
				hscx_a = (caddr_t) 0x360;
				hscx_b = (caddr_t) 0x760;
				printf("Teles S0/16.3 I/O-Ports in unconfirmed adress-space!\n");
				break;
			default:
				/* Don't know these values yet */
				sc->sc_typ = TEL_TYPE_UNK;
				printf("Teles S0/16.3 no I/O ports?!\n");
				return 0;
	        }
		sc->sc_typ = TEL_TYPE_16_3;
		sc->get = telget_pio;
		sc->put = telput_pio;
		sc->getfifo = telgetfifo_pio;
		sc->putfifo = telputfifo_pio;
		sc->outCMD = outCMD_pio;
		sc->inSTAR = inSTAR_pio;
		/* 
		 * Read HSCX A/B VSTR.  Expected value for the S0/16.3 card is
		 * 0x05 in the least significant bits.
		 */
		if (!((*sc->get)(hscx_a, 0x2e) & 0xf == 0x5) && 
		    ((*sc->get)(hscx_b, 0x2e) & 0xf == 0x5))
		    return 0;

		/* 
		 * Set SW programmed IRQ level for S0/16.3
		 */
		if (!((ia->ia_irq > 0 && ia->ia_irq < 16) &&
		    ((cfval = intr_no[ia->ia_irq]) != 0x01)))
		{
			printf("Teles S0/16.3: IRQ %d not supported!\n", ia->ia_irq);
			return 0;
		}
	}
	else 
	{
		/* S0/16 and S0/8 */
		sc->get = telget;
		sc->put = telput;
		sc->getfifo = memcpyb;
		sc->putfifo = memcpyb;
		sc->outCMD = outCMD;
		sc->inSTAR = inSTAR;
		sc->sc_typ= TEL_TYPE_1000;
		isac = (sc->sc_vmem_addr) + 0x100;
		hscx_a = (sc->sc_vmem_addr) + 0x180;
		hscx_b = (sc->sc_vmem_addr) + 0x1c0;
		/* check for S0/16 */
		if (ia->ia_iobase)
		{
			unsigned long   val;
			val = inb(ia->ia_iobase) << 16;
			val |= inb(ia->ia_iobase + 1) << 8;
			val |= inb(ia->ia_iobase + 2);
			if (val == 0x51931e)
			{
				sc->sc_typ = TEL_TYPE_16;
				val = intr_no[ia->ia_irq];
				if (val == 1)
				{
					printf("Teles S0/16: IRQ %d not supported!\n", ia->ia_irq);
					return 0;
				}
				val |= ((u_long) ia->ia_maddr) >> 9;
				if( (val & 0xff) != val)
				{
					printf("Teles S0/16: configuration values do not fit into configuration register!\n")
					return 0;
				}
				cfval = val & 0xff;
			}
		}
	}
	/*
	 * Write the configuration to card.
	 * S0/16: IRQ and memory base combined
	 * S0/16.3: IRQ
	 * S0/8: don't even try to overwrite the jumper settings
	 */
	if(ia->ia_iobase)
	{
		delay(SEC_DELAY / 10);	
		outb(ia->ia_iobase + 4, cfval);
		delay(SEC_DELAY / 10);	
		outb(ia->ia_iobase + 4, cfval | 1);
		delay(SEC_DELAY / 10);	
	}
	if (ia->ia_maddr == MADDRUNK)
	{
		ia->ia_iosize = 8;
		ia->ia_msize = 0;
		return 1;
	}
	else
	{
		/* Toggle card bit off */
		sc->sc_vmem_addr[0x80] = 0;
		DELAY(SEC_DELAY / 5);
		sc->sc_vmem_addr[0x80] = 1;
		DELAY(SEC_DELAY / 5);

		if ((0x48 == ((*sc->get)(isac, 0x21) & 0xfd)) ||
		    (0x48 == ((*sc->get)(hscx_a, 0x21) & 0xfd)) ||
		    (0x48 == ((*sc->get)(hscx_b, 0x21) & 0xfd)))
		{
			ia->ia_msize = 0x1000;
			sc->sc_vmem_addr[0x80]  = 0;
			return 1;
		}
	}
	return 0;
}

void
telattach(parent, self, aux)
	struct device *parent, *self;
	void   *aux;
{
	struct tel_softc *sc = (void *) self;
	struct isa_attach_args *ia = aux;
	int     cn;

	cn = 10;
	do {
		delay(SEC_DELAY / 5);
	} while (telinit(sc) && cn--);

	sc->sc_ctrl = -1;
	if ((cn = disdn_attach((struct device *)sc, &sc_llif, 
			       sc->sc_typ, 2)) == -1) 
	{
		printf("telattach: COULD NOT do disdnattach()!!!\n");
		return;
	}
	sc->sc_ctrl = (cn >> 8) & 0xff;
	cn &= 0xff;
	sc->sc_chan[0].c_ctrl = cn;
	sc->sc_chan[1].c_ctrl = cn + 1;
	sc->sc_spy = 0;

	if (sc->sc_typ == TEL_TYPE_16_3)
	{
		printf(": S0/16.3\n");
	}
	else
	{
		printf("\n");
	}
	sc->sc_ih = isa_intr_establish(ia->ia_ic, ia->ia_irq, IST_EDGE,
				       IPL_NET, telintr, sc);
}
#endif /* __NetBSD__ */

static struct mbuf*
tel_ibuf(chan_t *c, int len)
{
	if(c->ibuf)
		return(c->ibuf);

	if((c->ibuf= isdn_getbuf(len)) != NULL)
		c->ib= c->ibuf->m_data;

	c->ilen = 0;
	return(c->ibuf);
}

static void
#ifdef __FreeBSD__
intr_HSCX(int unit, u_char r, int c, u_char hasEX)
{
	struct tel_softc *sc = &tel_sc[unit];
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
intr_HSCX(struct tel_softc *sc, u_char r, int c, u_char hasEX)
{
#endif /* __NetBSD__ */
	chan_t         *chan = &sc->sc_chan[c];
	caddr_t         hscx= chan->hscx;
	if (hasEX)
	{
		u_char          exir;

		exir = (*sc->get)(hscx, EXIR);
		if (exir & 0xc0)
			chan->cmd |= 1;
		if (exir & 0x10)
		{
			if (r & 0x80)
			{
				chan->ilen = 0;
				if(chan->ibuf)
					m_freem(chan->ibuf);
				chan->ibuf= NULL;
				chan->ib= NULL;
				chan->c_state &= ~BADIFRAME;
				chan->cmd|= 0xc0;
			} else
			{
				chan->c_state |= BADIFRAME;
				chan->cmd|= 0x80;
			}
			r &= ~0xC0;
		}
		if (exir & 0x40)
		{
			chan->cmd |= 1;
			chan->oact= chan->obuf;
		}
	}

	if (r & 0x80)
	{
		u_short           len,  rest;
		struct mbuf	*m;
		u_char *b;

		len = (((*sc->get)(hscx, HRBCH) & 0x0F) << 8) + (*sc->get)(hscx, RBCL);
		rest = (len & 31);
		if(rest == 0)
			rest= 32;

		m= tel_ibuf(chan,32);
		chan->ilen += rest - chan->crc_byte;
		if (m && (chan->ilen <= 2048))
		{
			b= chan->ib;
			(*sc->getfifo)(b,hscx,rest);
			m->m_pkthdr.len = m->m_len = chan->ilen;
			chan->cmd |= 0x80;
			if(sc->sc_spy & 6)
			{
#ifdef __FreeBSD__
				isdn_binput(ispy_applnr,m->m_len,m->m_data,(c << 6) | (unit << 2) | 2);
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
				isdn_binput(ispy_applnr,m->m_len,m->m_data,(c << 6) | (sc->sc_unit << 2) | 2);
#endif /* __NetBSD__ */
			}
			if(!(chan->c_state & BADIFRAME))
				disdn_input(chan->c_ctrl,m);
			else m_freem(m);
			m= NULL;
		}
		else
			chan->cmd |= 0xc0;
		if(m) m_freem(m);
		chan->ilen = 0;
		chan->ibuf= NULL;
		chan->ib= NULL;
		chan->c_state &= ~BADIFRAME;
	}

	if (r & 0x40)	/* receive fifo full */
	{
		struct mbuf	*m;
		u_char *b;

		m= tel_ibuf(chan,2048);
		chan->ilen += 32;
		if (m && (chan->ilen <= 2048))
		{
			b= chan->ib;
			(*sc->getfifo)(b,hscx,32);
			chan->ib+= 32;
		}
		chan->cmd |= 0x80;
		if((chan->ilen >= 2048) && (chan->service == TELEFON))
		{
			m->m_pkthdr.len = m->m_len = chan->ilen;
			if(sc->sc_spy & 6)
			{
#ifdef __FreeBSD__
				isdn_binput(ispy_applnr,m->m_len,m->m_data,(c << 6) | (unit << 2) | 2);
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
				isdn_binput(ispy_applnr,m->m_len,m->m_data,(c << 6) | (sc->sc_unit << 2) | 2);
#endif /* __NetBSD__ */
			}
			disdn_input(chan->c_ctrl,m);
			chan->ilen = 0;
			chan->ibuf= NULL;
			chan->ib= NULL;
			chan->c_state &= ~BADIFRAME;
		}
	}

	if(chan->cmd)
	{
		(*sc->outCMD)(hscx, chan->cmd);
		chan->cmd= 0;
	}

	if(r & 0x10)
	{
#ifdef __FreeBSD__
		tel_start(unit,c);
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
		tel_start(sc,c);
#endif /* __NetBSD__ */
	}
}

static void
#ifdef __FreeBSD__
intr_ISAC(int unit, int r)
{
	register struct tel_softc *sc = &tel_sc[unit];
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
intr_ISAC(register struct tel_softc *sc, int r)
{
#endif /* __NetBSD__ */

	u_char c;
	c = 0;

	if (r & 0x80)	/* receive message end */
	{
		u_short         len, rest;
		struct mbuf	*m;
		u_char         *b;

		len = (((*sc->get)(isac, IRBCH) & 0x0F) << 8) + (*sc->get)(isac, RBCL);
		rest = (len & 31);
		if (rest == 0)
			rest = 32;

		if(sc->sc_ibuf == NULL)
		{
			if((sc->sc_ibuf= isdn_getbuf(rest)) != NULL)
				sc->sc_ib= sc->sc_ibuf->m_data;
			sc->sc_ilen = 0;
		}
		sc->sc_ilen += rest;
		if((m= sc->sc_ibuf) && (sc->sc_ilen < 260))
		{
			b = sc->sc_ib;
			(*sc->getfifo)(b,isac,rest);
			m->m_pkthdr.len = m->m_len = sc->sc_ilen;
			if(sc->sc_spy & 1)
			{
#ifdef __FreeBSD__
				isdn_binput(ispy_applnr,m->m_len,m->m_data,unit << 2);
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
				isdn_binput(ispy_applnr,m->m_len,m->m_data,sc->sc_unit << 2);
#endif /* __NetBSD__ */
			}
			c |= 0x80;	/* RMC */
			sc->sc_ibuf= NULL;
			disdn_d_2_intr(sc->sc_ctrl,m);
			m= NULL;
		} else
			c |= 0xc0;
		if(m) m_freem(m);
		sc->sc_ibuf= NULL;
	}

	if (r & 0x40)	/* receive fifo full */
	{
		struct mbuf *m;
		u_char *b;
		if(sc->sc_ibuf == NULL)
		{
			if((sc->sc_ibuf= isdn_getbuf(260)) != NULL)
				sc->sc_ib= sc->sc_ibuf->m_data;
			sc->sc_ilen = 0;
		}
		sc->sc_ilen += 32;
		if ((m= sc->sc_ibuf) && (sc->sc_ilen <= 256))
		{
			b = sc->sc_ib;
			(*sc->getfifo)(b,isac,32);
			sc->sc_ib+=32;
		}
		c |= 0x80;
	}

	if (r & 4)	/* channel status change */
	{			/* CISQ */
		u_char          ci;

		ci = (*sc->get)(isac, 0x31 /* CIRR */ );
		(*sc->get)(isac, 0x3b /* SQRR */ );

		if(ci & 2)
		{
			ci= (ci >> 2) & 0xf;
			switch(ci)
			{
			case 0xc:
			case 0xd:
				sc->sc_state&= ~WANT_L1;
				sc->sc_state|= L1_RUNNING;
				if (sc->sc_typ == TEL_TYPE_16_3)
				{
					(*sc->put)(isac, CIX0, 0x27);
				}
				break;
			case 4:
				break;
			case 7:
				if (sc->sc_typ == TEL_TYPE_16_3)
				{
					(*sc->put)(isac, CIX0, 0x27);
				}
				break;
			case 0:
			case 0xf:
				if(sc->sc_state)
				{
					sc->sc_state= 0;
					if(!(sc->sc_spy & 4))
					{
#ifdef __FreeBSD__
						tel_init(unit,0,-1,0);
						tel_init(unit,1,-1,0);
#endif /* __FreeBSD__ */
#ifdef __NetBSD
						tel_init(sc,0,-1,0);
						tel_init(sc,1,-1,0);
#endif /* __NetBSD__ */
					}
					disdn_release(sc->sc_ctrl);
				}
			default:
			}
		}
	}

	if (r & 1)	/* extended interrupt */
	{
		u_char          e;

		e = (*sc->get)(isac, EXIR);
		if (e & 2)
		{
			(*sc->put)(isac, MASK, 0xff);
			if (sc->sc_typ != TEL_TYPE_16_3)
			{
				sc->sc_vmem_addr[0x80] = 0;
			}
#ifdef __FreeBSD__
			timeout((TIMEOUT_FUNC_T) telinit, (caddr_t) unit, hz / 10);
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
			timeout((TIMEOUT_FUNC_T) telinit, (caddr_t) sc, hz / 10);
#endif /* __NetBSD__ */
			return;
		}
		if (e & 0xc0)
			c |= 1;
		if (e & 0x10)
			c |= 0xc0;
	}

	if(r & 0x10)	/* transmit fifo empty (XPR bit set) */
	{
		u_char *b= sc->sc_op;		/* get buffer ptr */

		/* clear the timeout */
#ifdef __FreeBSD__
		untimeout((TIMEOUT_FUNC_T) tel_d_wd, (caddr_t) unit);
#endif
#ifdef __NetBSD__
		untimeout((TIMEOUT_FUNC_T) tel_d_wd, (caddr_t) sc);
#endif

		if(b)
		{			
			u_short l;

#ifdef DEBUG
printf("XPR more data %d at %x...\n", sc->sc_ol, b);
#endif
			l = min(sc->sc_ol, 32);	/* get length, max = 32 bytes */
			(*sc->putfifo)(isac,b,l); /* send segment out to tx fifo */

			if (sc->sc_ol > 32)	/* length > 32 bytes ? */
			{
				sc->sc_ol -= 32;  /* length - 32 */
				sc->sc_op = b+32; /* buffer ptr + 32 bytes */
				c|= 0x08;         /* set XTF bit */
			}
			else
			{
				sc->sc_op= NULL;  /* last message segment */
				c|= 0x0a;         /* set XTF and XME bits */
			}
			
			(*sc->put)(isac, CMDR, c); /* feed command register */

			/* since 32 bytes should be xmitted in ~ 4 ms, wait */
			/* for 1 tick */
#ifdef __FreeBSD__
			timeout((TIMEOUT_FUNC_T) tel_d_wd, (caddr_t) unit, 1);
#endif
#ifdef __NetBSD__
			timeout((TIMEOUT_FUNC_T) tel_d_wd, (caddr_t) sc, 1);
#endif
			return;
		}

		if (c)
			(*sc->put)(isac, CMDR, c);

		sc->sc_state &= ~TEL_OUT;
		sc->sc_obuf= NULL;
		disdn_d_startout(sc->sc_ctrl);
		return;
	}
	if (c)
		(*sc->put)(isac, CMDR, c);
}

#ifdef __FreeBSD__
void
telintr(int unit)
{
	struct tel_softc *sc = &tel_sc[unit];
	u_char          r, s;

	r = (*sc->get)(hscx_b, ISTA);
	s = (*sc->get)(isac, ISTA);
	while(r||s)
	{
		if (r & 0x06)
			intr_HSCX(unit, (*sc->get)(hscx_a, ISTA), 0, r & 2);
		if (r & 0xF9)
			intr_HSCX(unit, r, 1, r & 1);
		if (s)
			intr_ISAC(unit,s);
		r = (*sc->get)(hscx_b, ISTA);
		s = (*sc->get)(isac, ISTA);
	}

	/* the Siemens books say that this is the correct way to */
	/* handle edge-triggered interrupts. Urlichs does this too. */
	(*sc->put)(isac, MASK, 0xff);
	(*sc->put)(hscx_a, MASK, 0xff);
	(*sc->put)(hscx_b, MASK, 0xff);
	/* put in a small delay */
	DELAY (100);
	(*sc->put)(hscx_a, MASK, hscx_a_mask);
	(*sc->put)(hscx_b, MASK, hscx_b_mask);
	(*sc->put)(isac, MASK, isac_mask);
}
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
int
telintr(void *arg)
{
	struct tel_softc *sc = arg;
	u_char          r, s;

	r = (*sc->get)(hscx_b, ISTA);
	s = (*sc->get)(isac, ISTA);
	while(r||s)
	{
		if (r & 0x06)
			intr_HSCX(sc, (*sc->get)(hscx_a, ISTA), 0, r & 2);
		if (r & 0xF9)
			intr_HSCX(sc, r, 1, r & 1);
		if (s)
			intr_ISAC(sc,s);
		r = (*sc->get)(hscx_b, ISTA);
		s = (*sc->get)(isac, ISTA);
	}

	/* the Siemens books say that this is the correct way to */
	/* handle edge-triggered interrupts. Urlichs does this too. */
	(*sc->put)(isac, MASK, 0xff);
	(*sc->put)(hscx_a, MASK, 0xff);
	(*sc->put)(hscx_b, MASK, 0xff);
	/* put in a small delay */
	DELAY (100);
	(*sc->put)(hscx_a, MASK, hscx_a_mask);
	(*sc->put)(hscx_b, MASK, hscx_b_mask);
	(*sc->put)(isac, MASK, isac_mask);

	return (1);
}
#endif /* __NetBSD__ */

static int
#ifdef __FreeBSD__
tel_setstate(int unit, int cmd, u_char *data)
{
	struct tel_softc *sc = &tel_sc[unit];
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
tel_setstate(struct tel_softc *sc, int cmd, u_char *data)
{
#endif /* __NetBSD__ */
	switch(cmd)
	{
	case ISDN_SPY:
		sc->sc_spy= data[0];
		break;
	}
	return(0);
}

#ifdef __FreeBSD__
static int
tel_getstate(int unit)
{
	return(tel_sc[unit].sc_state);
}
#endif /* __FreeBSD__ */
#ifdef __NetBSD__
int
tel_getstate(struct tel_softc *sc)
{
	return sc->sc_state;
}
#endif /* __NetBSD__ */

/* for FreeBSD this is in support.s */
#if defined (__NetBSD__) || defined (TELES_HAS_MEMCPYB)
void
memcpyb (void *to, const void *from, size_t len)
{
   /* %edi=to    %esi=from    %ecx=len  */
   asm volatile ("
       .align 4,0x90
       cld
       rep
       movsb"
       : "=D" (to), "=S" (from), "=c" (len)
       : "0" (to), "1" (from), "2" (len)
       : "memory");
}
#endif

/* watchdog function for XMIT */
void
#ifdef __FreeBSD__
tel_d_wd (int unit)
#endif
#ifdef __NetBSD__
tel_d_wd (struct tel_softc *sc)
#endif
{
#ifdef __FreeBSD__
	struct tel_softc *sc = &tel_sc[unit];
#endif
	u_char xfw = 0, xac;

	/* Transmit FIFO Write Enable - can write bytes */
	xfw = ((*sc->get)(isac, STAR));
	/* Transmitter Active */
	xac = ((*sc->get)(isac, IRBCH));

	if (sc->sc_state & TEL_OUT)
	{
		/* this says:
		 *	a) the XMIT buffer can accept more bytes
		 *	b) the XMITter is not active
		 */
		if ((xfw & 0x4a) && !(xac & 0x80))
		{
			/* it appears that the command sometimes */
			/* doesn't "take". Send it again ! */
			if (sc->sc_op)
				(*sc->put)(isac, CMDR, 0x08); /* more later */
			else
				(*sc->put)(isac, CMDR, 0x0a); /* that's all */

		}
		/* since 32 bytes should be xmitted in ~ 4 ms, wait */
		/* for 1 tick */
#ifdef __FreeBSD__
	timeout((TIMEOUT_FUNC_T) tel_d_wd, (caddr_t) unit, 1);
#endif
#ifdef __NetBSD__
	timeout((TIMEOUT_FUNC_T) tel_d_wd, (caddr_t) sc, 1);
#endif
	}
}
#endif /* NTEL > 0 */
